/* Generated by Cython 3.1.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "cyares/inc/cares_flag_check.h",
            "cyares/inc/cares_headers.h",
            "cyares/inc/cyares_utils.h"
        ],
        "extra_compile_args": [
            "-O2"
        ],
        "include_dirs": [
            "cyares",
            "cyares/inc"
        ],
        "name": "cyares.channel",
        "sources": [
            "cyares/channel.pyx"
        ]
    },
    "module_name": "cyares.channel"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API) && !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_1_2"
#define CYTHON_HEX_VERSION 0x030102F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x03090000
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cyares__channel
#define __PYX_HAVE_API__cyares__channel
/* Early includes */
#include <stdint.h>
#include <stddef.h>
#include <time.h>
#include "ares_private.h"
#include "inc/cares_headers.h"
#include <string.h>
#include <stdio.h>
#include "cares_flag_check.h"
#include "cyares_utils.h"
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "cyares/channel.pyx",
  "cyares/resulttypes.pxd",
  "cyares/channel.pxd",
  "<stringsource>",
  "cyares/exception.pxd",
  "cpython/type.pxd",
  "cyares/socket_handle.pxd",
  "cyares/handles.pxi",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_relaxed(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_acq_rel(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* CriticalSections.proto */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_Begin1(cs, arg) (void)cs
#define __Pyx_PyCriticalSection_Begin2(cs, arg1, arg2) (void)cs
#define __Pyx_PyCriticalSection_End1(cs)
#define __Pyx_PyCriticalSection_End2(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_Begin1 PyCriticalSection_Begin
#define __Pyx_PyCriticalSection_Begin2 PyCriticalSection2_Begin
#define __Pyx_PyCriticalSection_End1 PyCriticalSection_End
#define __Pyx_PyCriticalSection_End2 PyCriticalSection2_End
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6cyares_9exception_AresError;
struct __pyx_obj_6cyares_11resulttypes_AresResult;
struct __pyx_obj_6cyares_11resulttypes_ares_optval_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result;
struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result;
struct __pyx_obj_6cyares_11resulttypes_ares_host_result;
struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result;
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result;
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result;
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result;
struct __pyx_obj_6cyares_13socket_handle_SocketHandle;
struct __pyx_obj_6cyares_7channel_Channel;

/* "exception.pxd":1
 * cdef class AresError(Exception):             # <<<<<<<<<<<<<<
 *     cdef:
 *         # Allow python access but not-alterable access
*/
struct __pyx_obj_6cyares_9exception_AresError {
  PyBaseExceptionObject __pyx_base;
  int status;
  PyObject *strerror;
  PyObject *name;
};


/* "resulttypes.pxd":9
 * 
 * 
 * cdef class AresResult:             # <<<<<<<<<<<<<<
 *     cdef tuple _attrs
 * 
*/
struct __pyx_obj_6cyares_11resulttypes_AresResult {
  PyObject_HEAD
  PyObject *_attrs;
};


/* "resulttypes.pxd":29
 * 
 * 
 * cdef class ares_optval_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly str val
*/
struct __pyx_obj_6cyares_11resulttypes_ares_optval_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_optval_result *__pyx_vtab;
  PyObject *val;
  uint16_t id;
};


/* "resulttypes.pxd":58
 * 
 * 
 * cdef class ares_query_opt_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t udp_size
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_opt_result *__pyx_vtab;
  uint16_t udp_size;
  uint8_t version;
  uint16_t flags;
  PyObject *options;
  PyObject *type;
};


/* "resulttypes.pxd":79
 *         #
 * 
 * cdef class ares_query_hinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes cpu
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_vtab;
  PyObject *cpu;
  PyObject *os;
  PyObject *type;
};


/* "resulttypes.pxd":97
 * 
 * 
 * cdef class ares_query_sig_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t type_covered
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_sig_result *__pyx_vtab;
  uint16_t type_covered;
  uint8_t algorithm;
  uint8_t labels;
  unsigned int original_ttl;
  unsigned int expiration;
  unsigned int inception;
  uint16_t key_tag;
  PyObject *signature;
  PyObject *signers_name;
  PyObject *type;
};


/* "resulttypes.pxd":132
 * 
 * 
 * cdef class ares_query_tlsa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint8_t  cert_usage
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_vtab;
  uint8_t cert_usage;
  uint8_t selector;
  uint8_t match;
  PyObject *data;
  PyObject *type;
};


/* "resulttypes.pxd":158
 * 
 * 
 * cdef class ares_query_svcb_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_svcb_result *__pyx_vtab;
  uint16_t priority;
  PyObject *target;
  PyObject *params;
  PyObject *type;
};


/* "resulttypes.pxd":180
 * 
 * # same as svcb but don't let the https thingy fool you...
 * cdef class ares_query_https_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_https_result *__pyx_vtab;
  uint16_t priority;
  PyObject *target;
  PyObject *params;
  PyObject *type;
};


/* "resulttypes.pxd":200
 * 
 * 
 * cdef class ares_query_uri_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_uri_result *__pyx_vtab;
  uint16_t priority;
  uint16_t weight;
  PyObject *target;
  PyObject *type;
};


/* "resulttypes.pxd":221
 * 
 * 
 * cdef class ares_query_raw_rr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t ty
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_raw_rr_result *__pyx_vtab;
  uint16_t ty;
  PyObject *data;
};


/* "resulttypes.pxd":232
 * 
 * 
 * cdef class ares_query_a_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_a_result *__pyx_vtab;
  PyObject *host;
  int ttl;
};


/* "resulttypes.pxd":246
 * 
 * 
 * cdef class ares_query_aaaa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_aaaa_result *__pyx_vtab;
  PyObject *host;
  int ttl;
};


/* "resulttypes.pxd":258
 * 
 * 
 * cdef class  ares_query_caa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int critical
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_caa_result *__pyx_vtab;
  int critical;
  PyObject *property;
  PyObject *value;
  int ttl;
};


/* "resulttypes.pxd":272
 * 
 * 
 * cdef class ares_query_cname_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes cname
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_cname_result *__pyx_vtab;
  PyObject *cname;
  int ttl;
};


/* "resulttypes.pxd":283
 * 
 * 
 * cdef class ares_query_mx_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_mx_result *__pyx_vtab;
  PyObject *host;
  unsigned short priority;
  int ttl;
};


/* "resulttypes.pxd":296
 * 
 * 
 * cdef class ares_query_naptr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes flags
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_naptr_result *__pyx_vtab;
  PyObject *flags;
  PyObject *service;
  PyObject *regex;
  PyObject *replacement;
  unsigned short order;
  unsigned short preference;
  int ttl;
};


/* "resulttypes.pxd":313
 * 
 * 
 * cdef class ares_query_ns_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef readonly bytes host
 *     cdef readonly int ttl
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ns_result *__pyx_vtab;
  PyObject *host;
  int ttl;
};


/* "resulttypes.pxd":323
 * 
 * 
 * cdef class ares_query_ptr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes name
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ptr_result *__pyx_vtab;
  PyObject *name;
  PyObject *aliases;
  int ttl;
};


/* "resulttypes.pxd":341
 * 
 * 
 * cdef class ares_query_soa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes nsname
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_soa_result *__pyx_vtab;
  PyObject *nsname;
  PyObject *hostmaster;
  unsigned int serial;
  unsigned int refresh;
  unsigned int retry;
  unsigned int expire;
  unsigned int minttl;
  unsigned int ttl;
};


/* "resulttypes.pxd":361
 * 
 * 
 * cdef class ares_query_srv_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_srv_result *__pyx_vtab;
  PyObject *host;
  unsigned short port;
  unsigned short priority;
  int weight;
  int ttl;
};


/* "resulttypes.pxd":376
 *     )
 * 
 * cdef class ares_query_txt_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes text
*/
struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_txt_result *__pyx_vtab;
  PyObject *text;
  int ttl;
};


/* "resulttypes.pxd":396
 * #
 * 
 * cdef class ares_host_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes name
*/
struct __pyx_obj_6cyares_11resulttypes_ares_host_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_host_result *__pyx_vtab;
  PyObject *name;
  PyObject *aliases;
  PyObject *addresses;
};


/* "resulttypes.pxd":406
 * 
 * 
 * cdef class ares_nameinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes node
*/
struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_nameinfo_result *__pyx_vtab;
  PyObject *node;
  PyObject *service;
};


/* "resulttypes.pxd":415
 * 
 * 
 * cdef class ares_addrinfo_node_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int ttl
*/
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_node_result *__pyx_vtab;
  int ttl;
  int flags;
  int family;
  int socktype;
  int protocol;
  PyObject *addr;
};


/* "resulttypes.pxd":427
 *     cdef ares_addrinfo_node_result new(ares_addrinfo_node* ares_node)
 * 
 * cdef class ares_addrinfo_cname_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int ttl
*/
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_cname_result *__pyx_vtab;
  int ttl;
  PyObject *alias;
  PyObject *name;
};


/* "resulttypes.pxd":437
 * 
 * 
 * cdef class ares_addrinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly list cnames
*/
struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_AresResult __pyx_base;
  struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_result *__pyx_vtab;
  PyObject *cnames;
  PyObject *nodes;
};


/* "socket_handle.pxd":4
 * 
 * 
 * cdef class SocketHandle:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
*/
struct __pyx_obj_6cyares_13socket_handle_SocketHandle {
  PyObject_HEAD
  struct __pyx_vtabstruct_6cyares_13socket_handle_SocketHandle *__pyx_vtab;
  PyObject *callback;
};


/* "cyares/channel.pxd":11
 * 
 * 
 * cdef class Channel:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # ares will get to decide the fate of the channel pointer...
*/
struct __pyx_obj_6cyares_7channel_Channel {
  PyObject_HEAD
  struct __pyx_vtabstruct_6cyares_7channel_Channel *__pyx_vtab;
  ares_channel_t *channel;
  struct ares_options options;
  uint64_t _running;
  uint64_t _closed;
  PyObject *handles;
  PyObject *_query_lookups;
  int _cancelled;
  int event_thread;
  struct __pyx_obj_6cyares_13socket_handle_SocketHandle *socket_handle;
};



/* "resulttypes.pxd":29
 * 
 * 
 * cdef class ares_optval_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly str val
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_optval_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *(*__pyx_new)(ares_dns_rr_t const *, ares_dns_rr_key_t, size_t);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_optval_result *__pyx_vtabptr_6cyares_11resulttypes_ares_optval_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_18ares_optval_result_new(ares_dns_rr_t const *, ares_dns_rr_key_t, size_t);


/* "resulttypes.pxd":58
 * 
 * 
 * cdef class ares_query_opt_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t udp_size
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_opt_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_opt_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_opt_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_21ares_query_opt_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":79
 *         #
 * 
 * cdef class ares_query_hinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes cpu
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_hinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_hinfo_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_f_6cyares_11resulttypes_23ares_query_hinfo_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":97
 * 
 * 
 * cdef class ares_query_sig_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t type_covered
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_sig_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_sig_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_sig_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *__pyx_f_6cyares_11resulttypes_21ares_query_sig_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":132
 * 
 * 
 * cdef class ares_query_tlsa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint8_t  cert_usage
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_tlsa_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_tlsa_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_f_6cyares_11resulttypes_22ares_query_tlsa_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":158
 * 
 * 
 * cdef class ares_query_svcb_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_svcb_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_svcb_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_svcb_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *__pyx_f_6cyares_11resulttypes_22ares_query_svcb_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":180
 * 
 * # same as svcb but don't let the https thingy fool you...
 * cdef class ares_query_https_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_https_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_https_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_https_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *__pyx_f_6cyares_11resulttypes_23ares_query_https_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":200
 * 
 * 
 * cdef class ares_query_uri_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t priority
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_uri_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_uri_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_uri_result;
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *__pyx_f_6cyares_11resulttypes_21ares_query_uri_result_new(ares_dns_rr_t const *);


/* "resulttypes.pxd":221
 * 
 * 
 * cdef class ares_query_raw_rr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly uint16_t ty
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_raw_rr_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_raw_rr_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_raw_rr_result;


/* "resulttypes.pxd":232
 * 
 * 
 * cdef class ares_query_a_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_a_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result *(*old_new)(struct ares_addrttl *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_a_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_a_result;


/* "resulttypes.pxd":246
 * 
 * 
 * cdef class ares_query_aaaa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_aaaa_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result *(*old_new)(struct ares_addr6ttl *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_aaaa_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_aaaa_result;


/* "resulttypes.pxd":258
 * 
 * 
 * cdef class  ares_query_caa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int critical
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_caa_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result *(*old_new)(struct ares_caa_reply *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_caa_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_caa_result;


/* "resulttypes.pxd":272
 * 
 * 
 * cdef class ares_query_cname_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes cname
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_cname_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result *(*old_new)(struct hostent *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_cname_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_cname_result;


/* "resulttypes.pxd":283
 * 
 * 
 * cdef class ares_query_mx_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_mx_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result *(*old_new)(struct ares_mx_reply *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_mx_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_mx_result;


/* "resulttypes.pxd":296
 * 
 * 
 * cdef class ares_query_naptr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes flags
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_naptr_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result *(*old_new)(struct ares_naptr_reply *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_naptr_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_naptr_result;


/* "resulttypes.pxd":313
 * 
 * 
 * cdef class ares_query_ns_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef readonly bytes host
 *     cdef readonly int ttl
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ns_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result *(*old_new)(char *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ns_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_ns_result;


/* "resulttypes.pxd":323
 * 
 * 
 * cdef class ares_query_ptr_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes name
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ptr_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result *(*old_new)(struct hostent *, PyObject *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ptr_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_ptr_result;


/* "resulttypes.pxd":341
 * 
 * 
 * cdef class ares_query_soa_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes nsname
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_soa_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result *(*old_new)(struct ares_soa_reply *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_soa_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_soa_result;


/* "resulttypes.pxd":361
 * 
 * 
 * cdef class ares_query_srv_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes host
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_srv_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result *(*old_new)(struct ares_srv_reply *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result *(*__pyx_new)(ares_dns_rr_t const *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_srv_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_srv_result;


/* "resulttypes.pxd":376
 *     )
 * 
 * cdef class ares_query_txt_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes text
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_txt_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result *(*old_new)(struct ares_txt_ext *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result *(*from_object)(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result *);
  struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result *(*__pyx_new)(ares_dns_rr_t const *, size_t);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_txt_result *__pyx_vtabptr_6cyares_11resulttypes_ares_query_txt_result;


/* "resulttypes.pxd":396
 * #
 * 
 * cdef class ares_host_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes name
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_host_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_host_result *(*__pyx_new)(struct hostent *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_host_result *__pyx_vtabptr_6cyares_11resulttypes_ares_host_result;


/* "resulttypes.pxd":406
 * 
 * 
 * cdef class ares_nameinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly bytes node
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_nameinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result *(*__pyx_new)(char *, char *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_nameinfo_result *__pyx_vtabptr_6cyares_11resulttypes_ares_nameinfo_result;


/* "resulttypes.pxd":415
 * 
 * 
 * cdef class ares_addrinfo_node_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int ttl
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_node_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result *(*__pyx_new)(struct ares_addrinfo_node *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_node_result *__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_node_result;


/* "resulttypes.pxd":427
 *     cdef ares_addrinfo_node_result new(ares_addrinfo_node* ares_node)
 * 
 * cdef class ares_addrinfo_cname_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int ttl
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_cname_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result *(*__pyx_new)(struct ares_addrinfo_cname *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_cname_result *__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_cname_result;


/* "resulttypes.pxd":437
 * 
 * 
 * cdef class ares_addrinfo_result(AresResult):             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly list cnames
*/

struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_result {
  struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result *(*__pyx_new)(struct ares_addrinfo *);
};
static struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_result *__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_result;


/* "socket_handle.pxd":4
 * 
 * 
 * cdef class SocketHandle:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
*/

struct __pyx_vtabstruct_6cyares_13socket_handle_SocketHandle {
  struct __pyx_obj_6cyares_13socket_handle_SocketHandle *(*__pyx_new)(PyObject *);
  void (*handle_cb)(struct __pyx_obj_6cyares_13socket_handle_SocketHandle *, ares_socket_t, int, int);
};
static struct __pyx_vtabstruct_6cyares_13socket_handle_SocketHandle *__pyx_vtabptr_6cyares_13socket_handle_SocketHandle;


/* "cyares/channel.pyx":137
 * 
 * 
 * cdef class Channel:             # <<<<<<<<<<<<<<
 *     def __init__(
 *         self,
*/

struct __pyx_vtabstruct_6cyares_7channel_Channel {
  void (*cancel)(struct __pyx_obj_6cyares_7channel_Channel *, int __pyx_skip_dispatch);
  void *(*_malloc)(struct __pyx_obj_6cyares_7channel_Channel *, size_t);
  PyObject *(*_query)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*_search)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*_Channel__create_future)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *);
  ares_status_t (*_Channel__wait)(struct __pyx_obj_6cyares_7channel_Channel *, int);
};
static struct __pyx_vtabstruct_6cyares_7channel_Channel *__pyx_vtabptr_6cyares_7channel_Channel;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetRuntimeVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __Pyx_tp_new_fallback(PyObject* type_obj, PyObject* args, PyObject* kwargs);
#endif
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    newfunc tp_new = __Pyx_PyType_TryGetSlot((PyTypeObject*)type_obj, tp_new, newfunc);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    if (!tp_new) return __Pyx_tp_new_fallback(type_obj, args, kwargs);
#else
    assert(tp_new != NULL);
#endif
    return tp_new((PyTypeObject*)type_obj, args, kwargs);
}

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallCFunction.proto */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_Unicode.proto */
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectDelAttr.proto */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RejectKeywords.proto */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* py_set_remove.proto */
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key);

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* ReRaiseException.proto */
static CYTHON_INLINE void __Pyx_ReraiseException(void);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* LimitedApiGetTypeDict.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_1_2
#define __PYX_HAVE_RT_ImportType_proto_3_1_2
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_1_2(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_1_2(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_1_2 {
   __Pyx_ImportType_CheckSize_Error_3_1_2 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_1_2 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_1_2 = 2
};
static PyTypeObject *__Pyx_ImportType_3_1_2(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_1_2 check_size);
#endif

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL ||\
        (CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL)
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned short __Pyx_PyLong_As_unsigned_short(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint64_t(uint64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ares_socket_t(ares_socket_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ares_status_t(ares_status_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_1_2(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_18ares_optval_result_new(ares_dns_rr_t const *__pyx_v_dns_rr, ares_dns_rr_key_t __pyx_v_key, size_t __pyx_v_idx); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_21ares_query_opt_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_f_6cyares_11resulttypes_23ares_query_hinfo_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *__pyx_f_6cyares_11resulttypes_21ares_query_sig_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_f_6cyares_11resulttypes_22ares_query_tlsa_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *__pyx_f_6cyares_11resulttypes_22ares_query_svcb_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *__pyx_f_6cyares_11resulttypes_23ares_query_https_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *__pyx_f_6cyares_11resulttypes_21ares_query_uri_result_new(ares_dns_rr_t const *__pyx_v_dns_rr); /* proto*/
static void __pyx_f_6cyares_7channel_7Channel_cancel(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void *__pyx_f_6cyares_7channel_7Channel__malloc(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, size_t __pyx_v_size); /* proto*/
static PyObject *__pyx_f_6cyares_7channel_7Channel__Channel__create_future(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_callback); /* proto*/
static PyObject *__pyx_f_6cyares_7channel_7Channel__query(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_qname, PyObject *__pyx_v_qtype, int __pyx_v_qclass, PyObject *__pyx_v_callback); /* proto*/
static PyObject *__pyx_f_6cyares_7channel_7Channel__search(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_qname, PyObject *__pyx_v_qtype, int __pyx_v_qclass, PyObject *__pyx_v_callback); /* proto*/
static ares_status_t __pyx_f_6cyares_7channel_7Channel__Channel__wait(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_timeout_ms); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "libc.time" */

/* Module declarations from "cyares.ares" */

/* Module declarations from "cyares.exception" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cyares.inc" */

/* Module declarations from "cyares.resulttypes" */
static CYTHON_INLINE PyObject *__pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(ares_dns_rr_t const *, ares_dns_rr_key_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_optresults(ares_dns_rr_t const *, ares_dns_rr_key_t); /*proto*/

/* Module declarations from "cyares.socket_handle" */
static void (*__pyx_f_6cyares_13socket_handle___socket_state_callback)(void *, ares_socket_t, int, int); /*proto*/

/* Module declarations from "cython" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "libc.math" */

/* Module declarations from "cyares.callbacks" */
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_a)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_aaaa)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_caa)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_cname)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_mx)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_naptr)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_ns)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_ptr)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_soa)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_srv)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_query_on_txt)(void *, int, int, unsigned char *, int); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_getaddrinfo)(void *, int, int, struct ares_addrinfo *); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_gethostbyname)(void *, int, int, struct hostent *); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_nameinfo)(void *, int, int, char *, char *); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_gethostbyaddr)(void *, int, int, struct hostent *); /*proto*/
static void (*__pyx_f_6cyares_9callbacks___callback_dns_rec__any)(void *, ares_status_t, size_t, ares_dns_record_t const *); /*proto*/

/* Module declarations from "cyares.channel" */
static PyObject *__pyx_v_6cyares_7channel_Future = 0;
static int __pyx_v_6cyares_7channel_HAS_IDNA2008;
static int __pyx_v_6cyares_7channel_init_status;
static CYTHON_INLINE void __pyx_f_6cyares_7channel_raise_negative_timeout_err(PyObject *); /*proto*/
static int __pyx_f_6cyares_7channel_cyares_seconds_to_milliseconds(PyObject *); /*proto*/
static int __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(PyObject *, Py_buffer *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "cyares.channel"
extern int __pyx_module_is_main_cyares__channel;
int __pyx_module_is_main_cyares__channel = 0;

/* Implementation of "cyares.channel" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_ModuleNotFoundError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_range;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = ",";
static const char __pyx_k_A[] = "A";
static const char __pyx_k_Q[] = "\200\001\330\004\n\210+\220Q";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_q[] = "\200\001\330\004\013\210<\220q\230\001";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_Cs[] = "\200\001\360\020\000\005\014\320\013\034\230C\230s\240!";
static const char __pyx_k_MX[] = "MX";
static const char __pyx_k_NS[] = "NS";
static const char __pyx_k__2[] = ".";
static const char __pyx_k__3[] = "?";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_ip[] = "ip";
static const char __pyx_k_ms[] = "ms";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_tv[] = "tv";
static const char __pyx_k_ANY[] = "ANY";
static const char __pyx_k_A_q[] = "\200A\330\010\017\210q";
static const char __pyx_k_CAA[] = "CAA";
static const char __pyx_k_OPT[] = "OPT";
static const char __pyx_k_PTR[] = "PTR";
static const char __pyx_k_SIG[] = "SIG";
static const char __pyx_k_SOA[] = "SOA";
static const char __pyx_k_SRV[] = "SRV";
static const char __pyx_k_TXT[] = "TXT";
static const char __pyx_k_URI[] = "URI";
static const char __pyx_k_cpu[] = "cpu";
static const char __pyx_k_dev[] = "dev";
static const char __pyx_k_fut[] = "fut";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_sa4[] = "sa4";
static const char __pyx_k_sa6[] = "sa6";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_AAAA[] = "AAAA";
static const char __pyx_k_A_G1[] = "\200A\340\010\014\210G\2201";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_SVCB[] = "SVCB";
static const char __pyx_k_TLSA[] = "TLSA";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_func[] = "__func__";
static const char __pyx_k_host[] = "host";
static const char __pyx_k_idna[] = "idna";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_port[] = "port";
static const char __pyx_k_rfds[] = "rfds";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_view[] = "view";
static const char __pyx_k_wait[] = "wait";
static const char __pyx_k_wfds[] = "wfds";
static const char __pyx_k_CNAME[] = "CNAME";
static const char __pyx_k_HINFO[] = "HINFO";
static const char __pyx_k_HTTPS[] = "HTTPS";
static const char __pyx_k_NAPTR[] = "NAPTR";
static const char __pyx_k_addr4[] = "addr4";
static const char __pyx_k_addr6[] = "addr6";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_hints[] = "hints";
static const char __pyx_k_htonl[] = "__htonl";
static const char __pyx_k_htons[] = "__htons";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_maxtv[] = "maxtv";
static const char __pyx_k_ndots[] = "ndots";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_proto[] = "proto";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_socks[] = "socks";
static const char __pyx_k_tries[] = "tries";
static const char __pyx_k_A_1D_N[] = "\200A\330\010\023\2201\220D\230\001\330\010\014\210N\230!";
static const char __pyx_k_Future[] = "Future";
static const char __pyx_k_cancel[] = "cancel";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_family[] = "family";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_labels[] = "labels";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_reinit[] = "reinit";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_rotate[] = "rotate";
static const char __pyx_k_search[] = "search";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_weight[] = "weight";
static const char __pyx_k_Channel[] = "Channel";
static const char __pyx_k_address[] = "address";
static const char __pyx_k_bitmask[] = "bitmask";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_domains[] = "domains";
static const char __pyx_k_getsock[] = "getsock";
static const char __pyx_k_key_tag[] = "key_tag";
static const char __pyx_k_lookups[] = "lookups";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_read_fd[] = "read_fd";
static const char __pyx_k_servers[] = "servers";
static const char __pyx_k_service[] = "service";
static const char __pyx_k_timeout[] = "timeout";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_A_HG1A_L[] = "\200A\330\010\014\210H\220G\2301\230A\330\010\014\210L\230\001";
static const char __pyx_k_A_q_JoYa[] = "\200A\360\026\000\t\030\220q\230\004\230J\240o\260Y\270a";
static const char __pyx_k_Future_2[] = "_Future";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_add_note[] = "add_note";
static const char __pyx_k_callback[] = "callback";
static const char __pyx_k_flowinfo[] = "flowinfo";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_local_ip[] = "local_ip";
static const char __pyx_k_priority[] = "priority";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_scope_id[] = "scope_id";
static const char __pyx_k_selector[] = "selector";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_socktype[] = "socktype";
static const char __pyx_k_tcp_port[] = "tcp_port";
static const char __pyx_k_udp_port[] = "udp_port";
static const char __pyx_k_udp_size[] = "udp_size";
static const char __pyx_k_write_fd[] = "write_fd";
static const char __pyx_k_A_Q_2S_1A[] = "\200A\330\010\025\220[\240\001\240\024\240Q\330\010\013\2102\210S\220\001\330\014\022\220)\2301\230A";
static const char __pyx_k_A_q_J_7_a[] = "\200A\360\026\000\t\030\220q\230\004\230J\320&7\260\300a";
static const char __pyx_k_A_q_JoYoQ[] = "\200A\330\010\027\220q\230\004\230J\240o\260Y\270o\310Q";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_algorithm[] = "algorithm";
static const char __pyx_k_inception[] = "inception";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_local_dev[] = "local_dev";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_signature[] = "signature";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_cert_usage[] = "cert_usage";
static const char __pyx_k_expiration[] = "expiration";
static const char __pyx_k_process_fd[] = "process_fd";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_query_type[] = "query_type";
static const char __pyx_k_getaddrinfo[] = "getaddrinfo";
static const char __pyx_k_getnameinfo[] = "getnameinfo";
static const char __pyx_k_idna_decode[] = "idna_decode";
static const char __pyx_k_query_class[] = "query_class";
static const char __pyx_k_A_q_4z_A_Qaq[] = "\200A\340\010\031\230\021\230%\230q\240\001\330\010\032\230!\2304\230z\250\027\260\004\260A\330\010\035\230Q\230a\230q";
static const char __pyx_k_Channel_wait[] = "Channel.wait";
static const char __pyx_k_QUERY_TYPE_A[] = "QUERY_TYPE_A";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_event_thread[] = "event_thread";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_original_ttl[] = "original_ttl";
static const char __pyx_k_service_data[] = "service_data";
static const char __pyx_k_set_local_ip[] = "set_local_ip";
static const char __pyx_k_signers_name[] = "signers_name";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_type_covered[] = "type_covered";
static const char __pyx_k_Channel_debug[] = "Channel.debug";
static const char __pyx_k_Channel_query[] = "Channel.query";
static const char __pyx_k_QUERY_TYPE_MX[] = "QUERY_TYPE_MX";
static const char __pyx_k_QUERY_TYPE_NS[] = "QUERY_TYPE_NS";
static const char __pyx_k_gethostbyaddr[] = "gethostbyaddr";
static const char __pyx_k_gethostbyname[] = "gethostbyname";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_remove_future[] = "__remove_future";
static const char __pyx_k_set_local_dev[] = "set_local_dev";
static const char __pyx_k_sock_state_cb[] = "sock_state_cb";
static const char __pyx_k_Channel___exit[] = "Channel.__exit__";
static const char __pyx_k_Channel_cancel[] = "Channel.cancel";
static const char __pyx_k_Channel_reinit[] = "Channel.reinit";
static const char __pyx_k_Channel_search[] = "Channel.search";
static const char __pyx_k_Closed_Handles[] = "Closed Handles: {}";
static const char __pyx_k_QUERY_CLASS_HS[] = "QUERY_CLASS_HS";
static const char __pyx_k_QUERY_CLASS_IN[] = "QUERY_CLASS_IN";
static const char __pyx_k_QUERY_TYPE_ANY[] = "QUERY_TYPE_ANY";
static const char __pyx_k_QUERY_TYPE_CAA[] = "QUERY_TYPE_CAA";
static const char __pyx_k_QUERY_TYPE_PTR[] = "QUERY_TYPE_PTR";
static const char __pyx_k_QUERY_TYPE_SOA[] = "QUERY_TYPE_SOA";
static const char __pyx_k_QUERY_TYPE_SRV[] = "QUERY_TYPE_SRV";
static const char __pyx_k_QUERY_TYPE_TXT[] = "QUERY_TYPE_TXT";
static const char __pyx_k_buffer_carried[] = "buffer_carried";
static const char __pyx_k_cyares_channel[] = "cyares.channel";
static const char __pyx_k_Channel___enter[] = "Channel.__enter__";
static const char __pyx_k_Channel_getsock[] = "Channel.getsock";
static const char __pyx_k_Channel_timeout[] = "Channel.timeout";
static const char __pyx_k_QUERY_CLASS_ANY[] = "QUERY_CLASS_ANY";
static const char __pyx_k_QUERY_TYPE_AAAA[] = "QUERY_TYPE_AAAA";
static const char __pyx_k_Running_Handles[] = "Running Handles: {}";
static const char __pyx_k_Running_Queries[] = "Running Queries: {}";
static const char __pyx_k_process_read_fd[] = "process_read_fd";
static const char __pyx_k_resolvconf_path[] = "resolvconf_path";
static const char __pyx_k_running_queries[] = "running_queries";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_QUERY_CLASS_NONE[] = "QUERY_CLASS_NONE";
static const char __pyx_k_QUERY_TYPE_CNAME[] = "QUERY_TYPE_CNAME";
static const char __pyx_k_QUERY_TYPE_NAPTR[] = "QUERY_TYPE_NAPTR";
static const char __pyx_k_process_write_fd[] = "process_write_fd";
static const char __pyx_k_CYARES_SOCKET_BAD[] = "CYARES_SOCKET_BAD";
static const char __pyx_k_QUERY_CLASS_CHAOS[] = "QUERY_CLASS_CHAOS";
static const char __pyx_k_add_done_callback[] = "add_done_callback";
static const char __pyx_k_Channel_process_fd[] = "Channel.process_fd";
static const char __pyx_k_FFbbc_t7_6_X_JeS_a[] = "\320\004F\320Fb\320bc\330\010\017\210t\2207\230!\2306\240\034\250X\260\\\300\023\300J\310e\320S`\320`a";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_concurrent_futures[] = "concurrent.futures";
static const char __pyx_k_cyares_channel_pyx[] = "cyares/channel.pyx";
static const char __pyx_k_invalid_IP_address[] = "invalid IP address";
static const char __pyx_k_A_Q_7_4q_Q_7_4q_Q_a[] = "\200A\330\010\r\210Q\320\016#\2407\250!\2504\250q\330\010\r\210Q\320\016#\2407\250!\2504\250q\330\010\r\210Q\320\016\"\240'\250\021\250$\250a";
static const char __pyx_k_Channel_getaddrinfo[] = "Channel.getaddrinfo";
static const char __pyx_k_Channel_getnameinfo[] = "Channel.getnameinfo";
static const char __pyx_k_ModuleNotFoundError[] = "ModuleNotFoundError";
static const char __pyx_k_cyares_threadsafety[] = "cyares_threadsafety";
static const char __pyx_k_A_aq_QfAQ_4q_1_Qaq_q[] = "\200A\360\010\000\033\034\360\010\000\t\033\230$\320\036.\250a\250q\340\010%\240Q\240f\250A\250Q\330\010\032\230!\330\014\020\220\001\330\014\023\2204\220q\330\014\r\330\014\r\330\014\023\2201\340\010\035\230Q\230a\230q\330\010\017\210q";
static const char __pyx_k_Channel_set_local_ip[] = "Channel.set_local_ip";
static const char __pyx_k_GGccd_t81F_hl_ZuTaab[] = "\320\004G\320Gc\320cd\330\010\017\210t\2208\2301\230F\240,\250h\260l\300#\300Z\310u\320Ta\320ab";
static const char __pyx_k_Channel_gethostbyaddr[] = "Channel.gethostbyaddr";
static const char __pyx_k_Channel_gethostbyname[] = "Channel.gethostbyname";
static const char __pyx_k_Channel_set_local_dev[] = "Channel.set_local_dev";
static const char __pyx_k_Channel__remove_future[] = "_Channel__remove_future";
static const char __pyx_k_Invalid_IPv4_address_r[] = "Invalid IPv4 address %r";
static const char __pyx_k_Invalid_IPv6_address_r[] = "Invalid IPv6 address %r";
static const char __pyx_k_Channel___reduce_cython[] = "Channel.__reduce_cython__";
static const char __pyx_k_Channel___remove_future[] = "Channel.__remove_future";
static const char __pyx_k_Channel_process_read_fd[] = "Channel.process_read_fd";
static const char __pyx_k_socket_send_buffer_size[] = "socket_send_buffer_size";
static const char __pyx_k_Channel_process_write_fd[] = "Channel.process_write_fd";
static const char __pyx_k_Invalid_address_argument[] = "Invalid address argument";
static const char __pyx_k_Channel___setstate_cython[] = "Channel.__setstate_cython__";
static const char __pyx_k_socket_receive_buffer_size[] = "socket_receive_buffer_size";
static const char __pyx_k_invalid_query_type_specified[] = "invalid query type specified";
static const char __pyx_k_A_83a_T_q_s_A_T_7_Q_1_Cq_1_1A[] = "\320\004$\240A\360\010\000\t\014\2108\2203\220a\330\014\025\220T\230\027\240\002\240!\340\014\021\320\021/\250q\260\001\330\014\017\210s\220\"\220A\330\020\021\340\014\025\220T\230\027\240\001\240\021\340\010\013\2107\220#\220Q\330\014\023\2201\330\r\024\220C\220q\330\014\023\2201\340\014\022\220)\2301\230A";
static const char __pyx_k_A_Q_1_aq_1_z_uAQ_QfAQ_g_Q_QfAQ[] = "\200A\360\006\000\027\030\330\032\033\330\025\026\330\027\030\330\024\025\330\024\025\340\010\035\230Q\340\010#\2401\360\006\000\t\033\230$\320\036.\250a\250q\340\010\013\2101\330\014\017\210z\230\021\230&\240\001\330\020\027\220u\230A\230Q\330\014\035\230Q\230f\240A\240Q\330\014\026\220g\230\\\250\021\330\014\035\230Q\340\010%\240Q\240f\250A\250Q\340\010\r\210\\\230\021\330\010\r\210]\230!\330\010\r\210_\230A\330\010\r\210_\230A\340\010\030\230\001\330\014\020\220\001\330\014\023\2204\220q\330\014\r\330\014\r\210Q\330\014\r\330\014\023\2201\360\006\000\t\036\230Q\230a\230q\330\010\013\2101\330\014!\240\021\240!\2401\330\010\013\2101\330\014\017\320\017!\240\021\240!\330\010\017\210q";
static const char __pyx_k_sock_state_cb_must_be_callable[] = "sock_state_cb must be callable";
static const char __pyx_k_1_r_A_ZxuAQ_t1CuBa_j_1_A_AT_1G1[] = "\320\004!\240\021\360\010\000\t\014\2101\330\014\017\210r\220\023\220A\330\020\025\220Z\230x\240u\250A\250Q\330\020\025\220[\240\007\240t\2501\250C\250u\260B\260a\340\020\026\220j\240\001\240\021\360\006\000\r\022\220\032\2301\330\014\021\220\033\230A\340\010\024\220A\220T\230\032\2401\240G\2501\250A\340\010\013\2105\220\002\220(\230$\230b\240\001\330\014\023\2201\340\010\020\220\002\220(\230\"\230B\230i\240r\250\021";
static const char __pyx_k_9A_7_fAQ_aq_a_A_q_q_1JgT_q_aq_a[] = "\320\004;\2709\300A\360\014\000\t\032\230\021\230&\240\001\240\021\340\010\013\210>\230\021\230)\2407\250$\250f\260A\260Q\330\014\022\220$\320\026&\240a\240q\330\014\036\230a\330\020\024\220A\330\020\021\220\021\330\020\027\220q\330\020\021\330\020\021\330\020\027\220q\360\010\000\016\034\2301\230J\240g\250T\260\026\260q\270\001\330\014\022\220$\320\026&\240a\240q\330\014\036\230a\330\020\024\220A\330\020\021\220\021\330\020\027\220q\330\020\021\330\020\021\330\020\027\220q\360\010\000\r\"\240\021\240!\2401\330\014\022\220*\230A\230Q\340\010\035\230Q\230a\230q\330\010\017\210q";
static const char __pyx_k_A_3ay_1_q_Qd_1_t_7_fAS_Qaq_j_2Q[] = "\200A\360\010\000\033\034\360\014\000\t\014\2103\210a\210y\230\003\2301\330\014\020\220\007\220q\330\014\035\230Q\230d\240!\2401\330\014\017\210t\220>\240\021\240)\2507\260$\260f\270A\270S\300\001\330\020%\240Q\240a\240q\330\020\026\220j\240\001\320!;\2702\270Q\330\014\017\210~\230Q\330\014\017\210|\230<\240q\250\001\330\014!\240\021\240!\2401\330\014\022\220$\320\026&\240a\240q\330\014\034\230A\330\020\024\220A\330\020\033\2301\230A\330\020\027\220q\330\020\021\330\020\021\330\020\027\220q\340\r\020\220\001\220\031\230#\230Q\330\r\021\220\026\220z\240\034\250Q\330\014\035\230Q\230d\240!\2401\330\014\017\210t\220>\240\021\240*\250G\2604\260v\270Q\270c\300\021\330\020%\240Q\240a\240q\330\020\026\220j\240\001\320!;\2702\270Q\330\014\017\210\230a\330\014\017\210}\230L\250\001\250\021\330\014\017\320\017 \240\014\250A\250Q\330\014\017\320\017 \240\001\330\014!\240\021\240!\2401\330\014\022\220$\320\026&\240a\240q\330\014\034\230A\330\020\024\220A\330\020\033\2301\230A\330\020\027\220q\330\020\021\330\020\021\330\020\027\220q\360\006\000\r\023\220*\230A\230Q\340\010\017\210q";
static const char __pyx_k_Domain_names_being_idna_decoded[] = "Domain names being idna decoded should not exceed a size of 255";
static const char __pyx_k_c_ares_is_not_built_with_thread[] = "c-ares is not built with thread safety";
static const char __pyx_k_running_queries_is_an_immutable[] = "running_queries is an immutable property";
static const char __pyx_k_A_aq_Qiwd_4z_1EQR_q_F_1_4z1A_Q_j[] = "\200A\360\010\000\t\032\230\021\230$\230a\230q\330\010\t\330\014\017\210~\230Q\230i\240w\250d\260&\270\001\270\021\330\020\"\240!\2404\240z\260\036\270|\3101\310E\320QR\330\021\037\230q\240\t\250\027\260\004\260F\270!\2701\330\020\"\240!\2404\240z\3201A\300\024\300Q\340\020\026\220j\240\001\240\021\340\014!\240\021\240!\2401";
static const char __pyx_k_A_q_JgQ_E_aq_AYa_G1HE_AYa_G1E_vQ[] = "\200A\330\010\031\230\021\330\010\031\230\021\360\010\000\t\034\230<\240q\250\004\250J\260g\270Q\360\006\000\t\r\210E\220\025\220a\220q\330\014\017\320\017$\240A\240Y\250a\330\020\024\220G\2301\230H\240E\250\021\250!\330\014\017\320\017$\240A\240Y\250a\330\020\024\220G\2301\230E\240\021\240!\340\010\017\210v\220Q";
static const char __pyx_k_Note_that_Cython_is_deliberately[] = "Note that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.";
static const char __pyx_k_Provided_callbacks_must_be_calla[] = "Provided callbacks must be callable";
static const char __pyx_k_self_channel_self_options_cannot[] = "self.channel,self.options cannot be converted to a Python object for pickling";
static const char __pyx_k_sock_state_cb_and_event_thread_c[] = "sock_state_cb and event_thread cannot be used together";
static const char __pyx_k_timeout_must_be_an_integer_or_fl[] = "timeout must be an integer or float not %s";
static const char __pyx_k_timeout_needs_to_be_a_positive_n[] = "timeout needs to be a positive number or 0.0";
static const char __pyx_k_timeout_should_not_be_below_0_go[] = "timeout should not be below 0 got ";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_6cyares_7channel___htons(CYTHON_UNUSED PyObject *__pyx_self, unsigned short __pyx_v_s); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_2__htonl(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_l); /* proto */
static int __pyx_pf_6cyares_7channel_7Channel___init__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_flags, PyObject *__pyx_v_timeout, PyObject *__pyx_v_tries, PyObject *__pyx_v_ndots, PyObject *__pyx_v_tcp_port, PyObject *__pyx_v_udp_port, PyObject *__pyx_v_servers, PyObject *__pyx_v_domains, PyObject *__pyx_v_lookups, PyObject *__pyx_v_sock_state_cb, PyObject *__pyx_v_socket_send_buffer_size, PyObject *__pyx_v_socket_receive_buffer_size, int __pyx_v_rotate, PyObject *__pyx_v_local_ip, PyObject *__pyx_v_local_dev, PyObject *__pyx_v_resolvconf_path, int __pyx_v_event_thread); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_7servers___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static int __pyx_pf_6cyares_7channel_7Channel_7servers_2__set__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_servers); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_2cancel(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_4reinit(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static void __pyx_pf_6cyares_7channel_7Channel_6__dealloc__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_8__enter__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_10__exit__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_12__remove_future(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_fut); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_14debug(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_16query(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_query_type, PyObject *__pyx_v_callback, PyObject *__pyx_v_query_class); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_18search(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_query_type, PyObject *__pyx_v_callback, PyObject *__pyx_v_query_class); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_20process_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_read_fd, int __pyx_v_write_fd); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_22process_read_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_read_fd); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_24process_write_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_write_fd); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_26timeout(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_28getaddrinfo(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_callback, int __pyx_v_family, int __pyx_v_socktype, int __pyx_v_proto, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_30getnameinfo(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_address, int __pyx_v_flags, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_32gethostbyname(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_v_family, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_34gethostbyaddr(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_36set_local_dev(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_dev); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_38set_local_ip(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_ip); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_40getsock(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_42wait(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_15running_queries___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static int __pyx_pf_6cyares_7channel_7Channel_15running_queries_2__set__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ignore); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_12event_thread___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_44__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_7Channel_46__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6cyares_7channel_4cyares_threadsafety(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_tp_new_6cyares_7channel_Channel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  PyTypeObject *__pyx_ptype_6cyares_9exception_AresError;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_AresResult;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_optval_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_https_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_a_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_host_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result;
  PyTypeObject *__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result;
  PyTypeObject *__pyx_ptype_6cyares_13socket_handle_SocketHandle;
  PyObject *__pyx_type_6cyares_7channel_Channel;
  PyTypeObject *__pyx_ptype_6cyares_7channel_Channel;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  PyObject *__pyx_tuple[11];
  PyObject *__pyx_codeobj_tab[25];
  PyObject *__pyx_string_tab[236];
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1000;
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_n_u_A __pyx_string_tab[1]
#define __pyx_n_u_AAAA __pyx_string_tab[2]
#define __pyx_n_u_ANY __pyx_string_tab[3]
#define __pyx_n_u_CAA __pyx_string_tab[4]
#define __pyx_n_u_CNAME __pyx_string_tab[5]
#define __pyx_n_u_CYARES_SOCKET_BAD __pyx_string_tab[6]
#define __pyx_n_u_Channel __pyx_string_tab[7]
#define __pyx_n_u_Channel___enter __pyx_string_tab[8]
#define __pyx_n_u_Channel___exit __pyx_string_tab[9]
#define __pyx_n_u_Channel___reduce_cython __pyx_string_tab[10]
#define __pyx_n_u_Channel___remove_future __pyx_string_tab[11]
#define __pyx_n_u_Channel___setstate_cython __pyx_string_tab[12]
#define __pyx_n_u_Channel__remove_future __pyx_string_tab[13]
#define __pyx_n_u_Channel_cancel __pyx_string_tab[14]
#define __pyx_n_u_Channel_debug __pyx_string_tab[15]
#define __pyx_n_u_Channel_getaddrinfo __pyx_string_tab[16]
#define __pyx_n_u_Channel_gethostbyaddr __pyx_string_tab[17]
#define __pyx_n_u_Channel_gethostbyname __pyx_string_tab[18]
#define __pyx_n_u_Channel_getnameinfo __pyx_string_tab[19]
#define __pyx_n_u_Channel_getsock __pyx_string_tab[20]
#define __pyx_n_u_Channel_process_fd __pyx_string_tab[21]
#define __pyx_n_u_Channel_process_read_fd __pyx_string_tab[22]
#define __pyx_n_u_Channel_process_write_fd __pyx_string_tab[23]
#define __pyx_n_u_Channel_query __pyx_string_tab[24]
#define __pyx_n_u_Channel_reinit __pyx_string_tab[25]
#define __pyx_n_u_Channel_search __pyx_string_tab[26]
#define __pyx_n_u_Channel_set_local_dev __pyx_string_tab[27]
#define __pyx_n_u_Channel_set_local_ip __pyx_string_tab[28]
#define __pyx_n_u_Channel_timeout __pyx_string_tab[29]
#define __pyx_n_u_Channel_wait __pyx_string_tab[30]
#define __pyx_kp_u_Closed_Handles __pyx_string_tab[31]
#define __pyx_kp_u_Domain_names_being_idna_decoded __pyx_string_tab[32]
#define __pyx_n_u_Future __pyx_string_tab[33]
#define __pyx_n_u_Future_2 __pyx_string_tab[34]
#define __pyx_n_u_HINFO __pyx_string_tab[35]
#define __pyx_n_u_HTTPS __pyx_string_tab[36]
#define __pyx_kp_u_Invalid_IPv4_address_r __pyx_string_tab[37]
#define __pyx_kp_u_Invalid_IPv6_address_r __pyx_string_tab[38]
#define __pyx_kp_u_Invalid_address_argument __pyx_string_tab[39]
#define __pyx_n_u_KeyError __pyx_string_tab[40]
#define __pyx_n_u_MX __pyx_string_tab[41]
#define __pyx_n_u_ModuleNotFoundError __pyx_string_tab[42]
#define __pyx_n_u_NAPTR __pyx_string_tab[43]
#define __pyx_n_u_NS __pyx_string_tab[44]
#define __pyx_n_u_None __pyx_string_tab[45]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[46]
#define __pyx_n_u_OPT __pyx_string_tab[47]
#define __pyx_n_u_PTR __pyx_string_tab[48]
#define __pyx_kp_u_Provided_callbacks_must_be_calla __pyx_string_tab[49]
#define __pyx_n_u_QUERY_CLASS_ANY __pyx_string_tab[50]
#define __pyx_n_u_QUERY_CLASS_CHAOS __pyx_string_tab[51]
#define __pyx_n_u_QUERY_CLASS_HS __pyx_string_tab[52]
#define __pyx_n_u_QUERY_CLASS_IN __pyx_string_tab[53]
#define __pyx_n_u_QUERY_CLASS_NONE __pyx_string_tab[54]
#define __pyx_n_u_QUERY_TYPE_A __pyx_string_tab[55]
#define __pyx_n_u_QUERY_TYPE_AAAA __pyx_string_tab[56]
#define __pyx_n_u_QUERY_TYPE_ANY __pyx_string_tab[57]
#define __pyx_n_u_QUERY_TYPE_CAA __pyx_string_tab[58]
#define __pyx_n_u_QUERY_TYPE_CNAME __pyx_string_tab[59]
#define __pyx_n_u_QUERY_TYPE_MX __pyx_string_tab[60]
#define __pyx_n_u_QUERY_TYPE_NAPTR __pyx_string_tab[61]
#define __pyx_n_u_QUERY_TYPE_NS __pyx_string_tab[62]
#define __pyx_n_u_QUERY_TYPE_PTR __pyx_string_tab[63]
#define __pyx_n_u_QUERY_TYPE_SOA __pyx_string_tab[64]
#define __pyx_n_u_QUERY_TYPE_SRV __pyx_string_tab[65]
#define __pyx_n_u_QUERY_TYPE_TXT __pyx_string_tab[66]
#define __pyx_kp_u_Running_Handles __pyx_string_tab[67]
#define __pyx_kp_u_Running_Queries __pyx_string_tab[68]
#define __pyx_n_u_RuntimeError __pyx_string_tab[69]
#define __pyx_n_u_SIG __pyx_string_tab[70]
#define __pyx_n_u_SOA __pyx_string_tab[71]
#define __pyx_n_u_SRV __pyx_string_tab[72]
#define __pyx_n_u_SVCB __pyx_string_tab[73]
#define __pyx_n_u_TLSA __pyx_string_tab[74]
#define __pyx_n_u_TXT __pyx_string_tab[75]
#define __pyx_n_u_TypeError __pyx_string_tab[76]
#define __pyx_n_u_URI __pyx_string_tab[77]
#define __pyx_n_u_ValueError __pyx_string_tab[78]
#define __pyx_kp_u__2 __pyx_string_tab[79]
#define __pyx_kp_u__3 __pyx_string_tab[80]
#define __pyx_n_u_add_done_callback __pyx_string_tab[81]
#define __pyx_kp_u_add_note __pyx_string_tab[82]
#define __pyx_n_u_addr __pyx_string_tab[83]
#define __pyx_n_u_addr4 __pyx_string_tab[84]
#define __pyx_n_u_addr6 __pyx_string_tab[85]
#define __pyx_n_u_address __pyx_string_tab[86]
#define __pyx_n_u_algorithm __pyx_string_tab[87]
#define __pyx_n_u_args __pyx_string_tab[88]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[89]
#define __pyx_n_u_bitmask __pyx_string_tab[90]
#define __pyx_n_u_buffer_carried __pyx_string_tab[91]
#define __pyx_kp_u_c_ares_is_not_built_with_thread __pyx_string_tab[92]
#define __pyx_n_u_callback __pyx_string_tab[93]
#define __pyx_n_u_cancel __pyx_string_tab[94]
#define __pyx_n_u_cert_usage __pyx_string_tab[95]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[96]
#define __pyx_n_u_concurrent_futures __pyx_string_tab[97]
#define __pyx_n_u_cpu __pyx_string_tab[98]
#define __pyx_n_u_cyares_channel __pyx_string_tab[99]
#define __pyx_kp_u_cyares_channel_pyx __pyx_string_tab[100]
#define __pyx_n_u_cyares_threadsafety __pyx_string_tab[101]
#define __pyx_n_u_data __pyx_string_tab[102]
#define __pyx_n_u_debug __pyx_string_tab[103]
#define __pyx_n_u_decode __pyx_string_tab[104]
#define __pyx_n_u_dev __pyx_string_tab[105]
#define __pyx_kp_u_disable __pyx_string_tab[106]
#define __pyx_n_u_domains __pyx_string_tab[107]
#define __pyx_kp_u_enable __pyx_string_tab[108]
#define __pyx_n_u_encode __pyx_string_tab[109]
#define __pyx_n_u_enter __pyx_string_tab[110]
#define __pyx_n_u_event_thread __pyx_string_tab[111]
#define __pyx_n_u_exit __pyx_string_tab[112]
#define __pyx_n_u_expiration __pyx_string_tab[113]
#define __pyx_n_u_family __pyx_string_tab[114]
#define __pyx_n_u_flags __pyx_string_tab[115]
#define __pyx_n_u_flowinfo __pyx_string_tab[116]
#define __pyx_n_u_format __pyx_string_tab[117]
#define __pyx_n_u_func __pyx_string_tab[118]
#define __pyx_n_u_fut __pyx_string_tab[119]
#define __pyx_kp_u_gc __pyx_string_tab[120]
#define __pyx_n_u_getaddrinfo __pyx_string_tab[121]
#define __pyx_n_u_gethostbyaddr __pyx_string_tab[122]
#define __pyx_n_u_gethostbyname __pyx_string_tab[123]
#define __pyx_n_u_getnameinfo __pyx_string_tab[124]
#define __pyx_n_u_getsock __pyx_string_tab[125]
#define __pyx_n_u_getstate __pyx_string_tab[126]
#define __pyx_n_u_hints __pyx_string_tab[127]
#define __pyx_n_u_host __pyx_string_tab[128]
#define __pyx_n_u_htonl __pyx_string_tab[129]
#define __pyx_n_u_htons __pyx_string_tab[130]
#define __pyx_n_u_i __pyx_string_tab[131]
#define __pyx_n_u_id __pyx_string_tab[132]
#define __pyx_n_u_idna __pyx_string_tab[133]
#define __pyx_n_u_idna_decode __pyx_string_tab[134]
#define __pyx_n_u_inception __pyx_string_tab[135]
#define __pyx_kp_u_invalid_IP_address __pyx_string_tab[136]
#define __pyx_kp_u_invalid_query_type_specified __pyx_string_tab[137]
#define __pyx_n_u_ip __pyx_string_tab[138]
#define __pyx_n_u_is_coroutine __pyx_string_tab[139]
#define __pyx_kp_u_isenabled __pyx_string_tab[140]
#define __pyx_n_u_key_tag __pyx_string_tab[141]
#define __pyx_n_u_l __pyx_string_tab[142]
#define __pyx_n_u_labels __pyx_string_tab[143]
#define __pyx_n_u_local_dev __pyx_string_tab[144]
#define __pyx_n_u_local_ip __pyx_string_tab[145]
#define __pyx_n_u_lookups __pyx_string_tab[146]
#define __pyx_n_u_main __pyx_string_tab[147]
#define __pyx_n_u_match __pyx_string_tab[148]
#define __pyx_n_u_maxtv __pyx_string_tab[149]
#define __pyx_n_u_module __pyx_string_tab[150]
#define __pyx_n_u_ms __pyx_string_tab[151]
#define __pyx_n_u_name __pyx_string_tab[152]
#define __pyx_n_u_name_2 __pyx_string_tab[153]
#define __pyx_n_u_ndots __pyx_string_tab[154]
#define __pyx_n_u_options __pyx_string_tab[155]
#define __pyx_n_u_original_ttl __pyx_string_tab[156]
#define __pyx_n_u_os __pyx_string_tab[157]
#define __pyx_n_u_params __pyx_string_tab[158]
#define __pyx_n_u_pop __pyx_string_tab[159]
#define __pyx_n_u_port __pyx_string_tab[160]
#define __pyx_n_u_print __pyx_string_tab[161]
#define __pyx_n_u_priority __pyx_string_tab[162]
#define __pyx_n_u_process_fd __pyx_string_tab[163]
#define __pyx_n_u_process_read_fd __pyx_string_tab[164]
#define __pyx_n_u_process_write_fd __pyx_string_tab[165]
#define __pyx_n_u_proto __pyx_string_tab[166]
#define __pyx_n_u_pyx_capi __pyx_string_tab[167]
#define __pyx_n_u_pyx_state __pyx_string_tab[168]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[169]
#define __pyx_n_u_qualname __pyx_string_tab[170]
#define __pyx_n_u_query __pyx_string_tab[171]
#define __pyx_n_u_query_class __pyx_string_tab[172]
#define __pyx_n_u_query_type __pyx_string_tab[173]
#define __pyx_n_u_r __pyx_string_tab[174]
#define __pyx_n_u_range __pyx_string_tab[175]
#define __pyx_n_u_read_fd __pyx_string_tab[176]
#define __pyx_n_u_reduce __pyx_string_tab[177]
#define __pyx_n_u_reduce_cython __pyx_string_tab[178]
#define __pyx_n_u_reduce_ex __pyx_string_tab[179]
#define __pyx_n_u_reinit __pyx_string_tab[180]
#define __pyx_n_u_remove_future __pyx_string_tab[181]
#define __pyx_n_u_resolvconf_path __pyx_string_tab[182]
#define __pyx_n_u_return __pyx_string_tab[183]
#define __pyx_n_u_rfds __pyx_string_tab[184]
#define __pyx_n_u_rotate __pyx_string_tab[185]
#define __pyx_n_u_running_queries __pyx_string_tab[186]
#define __pyx_kp_u_running_queries_is_an_immutable __pyx_string_tab[187]
#define __pyx_n_u_s __pyx_string_tab[188]
#define __pyx_n_u_sa4 __pyx_string_tab[189]
#define __pyx_n_u_sa6 __pyx_string_tab[190]
#define __pyx_n_u_scope_id __pyx_string_tab[191]
#define __pyx_n_u_search __pyx_string_tab[192]
#define __pyx_n_u_selector __pyx_string_tab[193]
#define __pyx_n_u_self __pyx_string_tab[194]
#define __pyx_kp_u_self_channel_self_options_cannot __pyx_string_tab[195]
#define __pyx_n_u_servers __pyx_string_tab[196]
#define __pyx_n_u_service __pyx_string_tab[197]
#define __pyx_n_u_service_data __pyx_string_tab[198]
#define __pyx_n_u_set_local_dev __pyx_string_tab[199]
#define __pyx_n_u_set_local_ip __pyx_string_tab[200]
#define __pyx_n_u_set_name __pyx_string_tab[201]
#define __pyx_n_u_setstate __pyx_string_tab[202]
#define __pyx_n_u_setstate_cython __pyx_string_tab[203]
#define __pyx_n_u_signature __pyx_string_tab[204]
#define __pyx_n_u_signers_name __pyx_string_tab[205]
#define __pyx_n_u_sock_state_cb __pyx_string_tab[206]
#define __pyx_kp_u_sock_state_cb_and_event_thread_c __pyx_string_tab[207]
#define __pyx_kp_u_sock_state_cb_must_be_callable __pyx_string_tab[208]
#define __pyx_n_u_socket_receive_buffer_size __pyx_string_tab[209]
#define __pyx_n_u_socket_send_buffer_size __pyx_string_tab[210]
#define __pyx_n_u_socks __pyx_string_tab[211]
#define __pyx_n_u_socktype __pyx_string_tab[212]
#define __pyx_n_u_status __pyx_string_tab[213]
#define __pyx_kp_u_stringsource __pyx_string_tab[214]
#define __pyx_n_u_t __pyx_string_tab[215]
#define __pyx_n_u_target __pyx_string_tab[216]
#define __pyx_n_u_tcp_port __pyx_string_tab[217]
#define __pyx_n_u_test __pyx_string_tab[218]
#define __pyx_n_u_timeout __pyx_string_tab[219]
#define __pyx_kp_u_timeout_must_be_an_integer_or_fl __pyx_string_tab[220]
#define __pyx_kp_u_timeout_needs_to_be_a_positive_n __pyx_string_tab[221]
#define __pyx_kp_u_timeout_should_not_be_below_0_go __pyx_string_tab[222]
#define __pyx_n_u_tries __pyx_string_tab[223]
#define __pyx_n_u_tv __pyx_string_tab[224]
#define __pyx_n_u_type __pyx_string_tab[225]
#define __pyx_n_u_type_covered __pyx_string_tab[226]
#define __pyx_n_u_udp_port __pyx_string_tab[227]
#define __pyx_n_u_udp_size __pyx_string_tab[228]
#define __pyx_n_u_val __pyx_string_tab[229]
#define __pyx_n_u_version __pyx_string_tab[230]
#define __pyx_n_u_view __pyx_string_tab[231]
#define __pyx_n_u_wait __pyx_string_tab[232]
#define __pyx_n_u_weight __pyx_string_tab[233]
#define __pyx_n_u_wfds __pyx_string_tab[234]
#define __pyx_n_u_write_fd __pyx_string_tab[235]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_9exception_AresError);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_AresResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_optval_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_a_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_host_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_13socket_handle_SocketHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cyares_7channel_Channel);
  Py_CLEAR(clear_module_state->__pyx_type_6cyares_7channel_Channel);
  for (int i=0; i<11; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<25; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<236; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1000);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_9exception_AresError);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_AresResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_optval_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_a_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_host_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_13socket_handle_SocketHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cyares_7channel_Channel);
  Py_VISIT(traverse_module_state->__pyx_type_6cyares_7channel_Channel);
  for (int i=0; i<11; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<25; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<236; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_float_0_0);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_0);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_1000);
  return 0;
}
#endif
/* #### Code section: module_code ### */

/* "resulttypes.pxd":19
 * 
 * # custom
 * cdef inline bytes cyares_dns_rr_get_bytes(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):             # <<<<<<<<<<<<<<
 *     return PyBytes_FromString(ares_dns_rr_get_str(dns_rr, key))
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(ares_dns_rr_t const *__pyx_v_dns_rr, ares_dns_rr_key_t __pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cyares_dns_rr_get_bytes", 0);

  /* "resulttypes.pxd":20
 * # custom
 * cdef inline bytes cyares_dns_rr_get_bytes(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):
 *     return PyBytes_FromString(ares_dns_rr_get_str(dns_rr, key))             # <<<<<<<<<<<<<<
 * 
 * # cdef inline str cyares_dns_rr_get_abin(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromString(ares_dns_rr_get_str(__pyx_v_dns_rr, __pyx_v_key)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "resulttypes.pxd":19
 * 
 * # custom
 * cdef inline bytes cyares_dns_rr_get_bytes(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):             # <<<<<<<<<<<<<<
 *     return PyBytes_FromString(ares_dns_rr_get_str(dns_rr, key))
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.cyares_dns_rr_get_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":34
 *         readonly uint16_t id
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_optval_result new(
 *         const ares_dns_rr_t* dns_rr,
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_18ares_optval_result_new(ares_dns_rr_t const *__pyx_v_dns_rr, ares_dns_rr_key_t __pyx_v_key, size_t __pyx_v_idx) {
  size_t __pyx_v_len;
  uint8_t const *__pyx_v_cstr;
  struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":42
 *         cdef size_t len
 *         cdef const uint8_t* cstr
 *         cdef ares_optval_result r = ares_optval_result.__new__(ares_optval_result)             # <<<<<<<<<<<<<<
 *         r.id = ares_dns_rr_get_opt(dns_rr, key, idx, &cstr, &len)
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_optval_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_optval_result)))) __PYX_ERR(1, 42, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":43
 *         cdef const uint8_t* cstr
 *         cdef ares_optval_result r = ares_optval_result.__new__(ares_optval_result)
 *         r.id = ares_dns_rr_get_opt(dns_rr, key, idx, &cstr, &len)             # <<<<<<<<<<<<<<
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)
 *         r._attrs = ("id", "val")
*/
  __pyx_v_r->id = ares_dns_rr_get_opt(__pyx_v_dns_rr, __pyx_v_key, __pyx_v_idx, (&__pyx_v_cstr), (&__pyx_v_len));

  /* "resulttypes.pxd":44
 *         cdef ares_optval_result r = ares_optval_result.__new__(ares_optval_result)
 *         r.id = ares_dns_rr_get_opt(dns_rr, key, idx, &cstr, &len)
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)             # <<<<<<<<<<<<<<
 *         r._attrs = ("id", "val")
 *         return r
*/
  __pyx_t_1 = cyares_unicode_from_uchar_and_size(__pyx_v_cstr, ((Py_ssize_t)__pyx_v_len)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->val);
  __Pyx_DECREF(__pyx_v_r->val);
  __pyx_v_r->val = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":45
 *         r.id = ares_dns_rr_get_opt(dns_rr, key, idx, &cstr, &len)
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)
 *         r._attrs = ("id", "val")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[0];

  /* "resulttypes.pxd":46
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)
 *         r._attrs = ("id", "val")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":34
 *         readonly uint16_t id
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_optval_result new(
 *         const ares_dns_rr_t* dns_rr,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_optval_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":49
 * 
 * 
 * cdef inline list cyares_dns_rr_get_optresults(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef size_t size = ares_dns_rr_get_opt_cnt(dns_rr, key)
*/

static CYTHON_INLINE PyObject *__pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_optresults(ares_dns_rr_t const *__pyx_v_dns_rr, ares_dns_rr_key_t __pyx_v_key) {
  size_t __pyx_v_i;
  size_t __pyx_v_size;
  PyObject *__pyx_v_optvals = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cyares_dns_rr_get_optresults", 0);

  /* "resulttypes.pxd":51
 * cdef inline list cyares_dns_rr_get_optresults(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):
 *     cdef size_t i
 *     cdef size_t size = ares_dns_rr_get_opt_cnt(dns_rr, key)             # <<<<<<<<<<<<<<
 *     cdef list optvals = PyList_New(<Py_ssize_t>size)
 *     for i in range(size):
*/
  __pyx_v_size = ares_dns_rr_get_opt_cnt(__pyx_v_dns_rr, __pyx_v_key);

  /* "resulttypes.pxd":52
 *     cdef size_t i
 *     cdef size_t size = ares_dns_rr_get_opt_cnt(dns_rr, key)
 *     cdef list optvals = PyList_New(<Py_ssize_t>size)             # <<<<<<<<<<<<<<
 *     for i in range(size):
 *         PyList_SET_ITEM(optvals, <Py_ssize_t>i, ares_optval_result.new(dns_rr, key, i))
*/
  __pyx_t_1 = PyList_New(((Py_ssize_t)__pyx_v_size)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_optvals = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":53
 *     cdef size_t size = ares_dns_rr_get_opt_cnt(dns_rr, key)
 *     cdef list optvals = PyList_New(<Py_ssize_t>size)
 *     for i in range(size):             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(optvals, <Py_ssize_t>i, ares_optval_result.new(dns_rr, key, i))
 *     return optvals
*/
  __pyx_t_2 = __pyx_v_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "resulttypes.pxd":54
 *     cdef list optvals = PyList_New(<Py_ssize_t>size)
 *     for i in range(size):
 *         PyList_SET_ITEM(optvals, <Py_ssize_t>i, ares_optval_result.new(dns_rr, key, i))             # <<<<<<<<<<<<<<
 *     return optvals
 * 
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_6cyares_11resulttypes_18ares_optval_result_new(__pyx_v_dns_rr, __pyx_v_key, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_v_optvals, ((Py_ssize_t)__pyx_v_i), __pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "resulttypes.pxd":55
 *     for i in range(size):
 *         PyList_SET_ITEM(optvals, <Py_ssize_t>i, ares_optval_result.new(dns_rr, key, i))
 *     return optvals             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_optvals);
  __pyx_r = __pyx_v_optvals;
  goto __pyx_L0;

  /* "resulttypes.pxd":49
 * 
 * 
 * cdef inline list cyares_dns_rr_get_optresults(const ares_dns_rr_t* dns_rr, ares_dns_rr_key_t key):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef size_t size = ares_dns_rr_get_opt_cnt(dns_rr, key)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.cyares_dns_rr_get_optresults", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_optvals);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":65
 *         readonly list options
 *         readonly str type
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_optval_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_f_6cyares_11resulttypes_21ares_query_opt_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":69
 *         const ares_dns_rr_t* dns_rr
 *     ):
 *         cdef ares_query_opt_result r = ares_query_opt_result.__new__(ares_query_opt_result)             # <<<<<<<<<<<<<<
 *         r.udp_size = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_UDP_SIZE)
 *         r.version = ares_dns_rr_get_u8(dns_rr, ARES_RR_OPT_VERSION)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result)))) __PYX_ERR(1, 69, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":70
 *     ):
 *         cdef ares_query_opt_result r = ares_query_opt_result.__new__(ares_query_opt_result)
 *         r.udp_size = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_UDP_SIZE)             # <<<<<<<<<<<<<<
 *         r.version = ares_dns_rr_get_u8(dns_rr, ARES_RR_OPT_VERSION)
 *         r.flags = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_FLAGS)
*/
  __pyx_v_r->udp_size = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_OPT_UDP_SIZE);

  /* "resulttypes.pxd":71
 *         cdef ares_query_opt_result r = ares_query_opt_result.__new__(ares_query_opt_result)
 *         r.udp_size = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_UDP_SIZE)
 *         r.version = ares_dns_rr_get_u8(dns_rr, ARES_RR_OPT_VERSION)             # <<<<<<<<<<<<<<
 *         r.flags = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_FLAGS)
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)
*/
  __pyx_v_r->version = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_OPT_VERSION);

  /* "resulttypes.pxd":72
 *         r.udp_size = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_UDP_SIZE)
 *         r.version = ares_dns_rr_get_u8(dns_rr, ARES_RR_OPT_VERSION)
 *         r.flags = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_FLAGS)             # <<<<<<<<<<<<<<
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)
 *         r.type = "OPT"
*/
  __pyx_v_r->flags = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_OPT_FLAGS);

  /* "resulttypes.pxd":73
 *         r.version = ares_dns_rr_get_u8(dns_rr, ARES_RR_OPT_VERSION)
 *         r.flags = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_FLAGS)
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)             # <<<<<<<<<<<<<<
 *         r.type = "OPT"
 *         r._attrs = ("udp_size", "version", "flags", "options", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_optresults(__pyx_v_dns_rr, ARES_RR_OPT_OPTIONS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->options);
  __Pyx_DECREF(__pyx_v_r->options);
  __pyx_v_r->options = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":74
 *         r.flags = ares_dns_rr_get_u16(dns_rr, ARES_RR_OPT_FLAGS)
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)
 *         r.type = "OPT"             # <<<<<<<<<<<<<<
 *         r._attrs = ("udp_size", "version", "flags", "options", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_OPT);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_OPT);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_OPT;

  /* "resulttypes.pxd":75
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)
 *         r.type = "OPT"
 *         r._attrs = ("udp_size", "version", "flags", "options", "type")             # <<<<<<<<<<<<<<
 *         return r
 *         #
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[1];

  /* "resulttypes.pxd":76
 *         r.type = "OPT"
 *         r._attrs = ("udp_size", "version", "flags", "options", "type")
 *         return r             # <<<<<<<<<<<<<<
 *         #
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  if (!(likely(((((PyObject *)__pyx_v_r)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_r), __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_optval_result))))) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_optval_result *)__pyx_v_r);
  goto __pyx_L0;

  /* "resulttypes.pxd":65
 *         readonly list options
 *         readonly str type
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_optval_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_opt_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":85
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_hinfo_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_f_6cyares_11resulttypes_23ares_query_hinfo_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":89
 *         const ares_dns_rr_t* dns_rr
 *     ):
 *         cdef ares_query_hinfo_result r = ares_query_hinfo_result.__new__(ares_query_hinfo_result)             # <<<<<<<<<<<<<<
 *         r.cpu = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_CPU)
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result)))) __PYX_ERR(1, 89, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":90
 *     ):
 *         cdef ares_query_hinfo_result r = ares_query_hinfo_result.__new__(ares_query_hinfo_result)
 *         r.cpu = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_CPU)             # <<<<<<<<<<<<<<
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)
 *         r.type = "HINFO"
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_HINFO_CPU); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->cpu);
  __Pyx_DECREF(__pyx_v_r->cpu);
  __pyx_v_r->cpu = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":91
 *         cdef ares_query_hinfo_result r = ares_query_hinfo_result.__new__(ares_query_hinfo_result)
 *         r.cpu = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_CPU)
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)             # <<<<<<<<<<<<<<
 *         r.type = "HINFO"
 *         r._attrs = ("cpu", "os", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_HINFO_OS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->os);
  __Pyx_DECREF(__pyx_v_r->os);
  __pyx_v_r->os = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":92
 *         r.cpu = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_CPU)
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)
 *         r.type = "HINFO"             # <<<<<<<<<<<<<<
 *         r._attrs = ("cpu", "os", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_HINFO);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_HINFO);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_HINFO;

  /* "resulttypes.pxd":93
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)
 *         r.type = "HINFO"
 *         r._attrs = ("cpu", "os", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[2];

  /* "resulttypes.pxd":94
 *         r.type = "HINFO"
 *         r._attrs = ("cpu", "os", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":85
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_hinfo_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_hinfo_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":110
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_sig_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *__pyx_f_6cyares_11resulttypes_21ares_query_sig_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  size_t __pyx_v_len;
  uint8_t const *__pyx_v_cstr;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":116
 *         cdef size_t len
 *         cdef const uint8_t* cstr
 *         cdef ares_query_sig_result r = ares_query_sig_result.__new__(ares_query_sig_result)             # <<<<<<<<<<<<<<
 *         r.type_covered = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_TYPE_COVERED)
 *         r.algorithm = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_ALGORITHM)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result)))) __PYX_ERR(1, 116, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":117
 *         cdef const uint8_t* cstr
 *         cdef ares_query_sig_result r = ares_query_sig_result.__new__(ares_query_sig_result)
 *         r.type_covered = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_TYPE_COVERED)             # <<<<<<<<<<<<<<
 *         r.algorithm = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_ALGORITHM)
 *         r.labels = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_LABELS)
*/
  __pyx_v_r->type_covered = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_SIG_TYPE_COVERED);

  /* "resulttypes.pxd":118
 *         cdef ares_query_sig_result r = ares_query_sig_result.__new__(ares_query_sig_result)
 *         r.type_covered = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_TYPE_COVERED)
 *         r.algorithm = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_ALGORITHM)             # <<<<<<<<<<<<<<
 *         r.labels = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_LABELS)
 *         r.original_ttl = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_ORIGINAL_TTL)
*/
  __pyx_v_r->algorithm = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_SIG_ALGORITHM);

  /* "resulttypes.pxd":119
 *         r.type_covered = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_TYPE_COVERED)
 *         r.algorithm = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_ALGORITHM)
 *         r.labels = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_LABELS)             # <<<<<<<<<<<<<<
 *         r.original_ttl = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_ORIGINAL_TTL)
 *         r.expiration = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_EXPIRATION)
*/
  __pyx_v_r->labels = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_SIG_LABELS);

  /* "resulttypes.pxd":120
 *         r.algorithm = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_ALGORITHM)
 *         r.labels = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_LABELS)
 *         r.original_ttl = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_ORIGINAL_TTL)             # <<<<<<<<<<<<<<
 *         r.expiration = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_EXPIRATION)
 *         r.inception = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_INCEPTION)
*/
  __pyx_v_r->original_ttl = ares_dns_rr_get_u32(__pyx_v_dns_rr, ARES_RR_SIG_ORIGINAL_TTL);

  /* "resulttypes.pxd":121
 *         r.labels = ares_dns_rr_get_u8(dns_rr, ARES_RR_SIG_LABELS)
 *         r.original_ttl = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_ORIGINAL_TTL)
 *         r.expiration = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_EXPIRATION)             # <<<<<<<<<<<<<<
 *         r.inception = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_INCEPTION)
 *         r.key_tag = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_KEY_TAG)
*/
  __pyx_v_r->expiration = ares_dns_rr_get_u32(__pyx_v_dns_rr, ARES_RR_SIG_EXPIRATION);

  /* "resulttypes.pxd":122
 *         r.original_ttl = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_ORIGINAL_TTL)
 *         r.expiration = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_EXPIRATION)
 *         r.inception = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_INCEPTION)             # <<<<<<<<<<<<<<
 *         r.key_tag = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_KEY_TAG)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_SIG_SIGNATURE, &len)
*/
  __pyx_v_r->inception = ares_dns_rr_get_u32(__pyx_v_dns_rr, ARES_RR_SIG_INCEPTION);

  /* "resulttypes.pxd":123
 *         r.expiration = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_EXPIRATION)
 *         r.inception = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_INCEPTION)
 *         r.key_tag = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_KEY_TAG)             # <<<<<<<<<<<<<<
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_SIG_SIGNATURE, &len)
 *         r.signature = cyares_unicode_from_uchar_and_size(cstr, len)
*/
  __pyx_v_r->key_tag = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_SIG_KEY_TAG);

  /* "resulttypes.pxd":124
 *         r.inception = ares_dns_rr_get_u32(dns_rr, ARES_RR_SIG_INCEPTION)
 *         r.key_tag = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_KEY_TAG)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_SIG_SIGNATURE, &len)             # <<<<<<<<<<<<<<
 *         r.signature = cyares_unicode_from_uchar_and_size(cstr, len)
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)
*/
  __pyx_v_cstr = ares_dns_rr_get_bin(__pyx_v_dns_rr, ARES_RR_SIG_SIGNATURE, (&__pyx_v_len));

  /* "resulttypes.pxd":125
 *         r.key_tag = ares_dns_rr_get_u16(dns_rr, ARES_RR_SIG_KEY_TAG)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_SIG_SIGNATURE, &len)
 *         r.signature = cyares_unicode_from_uchar_and_size(cstr, len)             # <<<<<<<<<<<<<<
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)
 *         r.type = "SIG"
*/
  __pyx_t_1 = cyares_unicode_from_uchar_and_size(__pyx_v_cstr, __pyx_v_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->signature);
  __Pyx_DECREF(__pyx_v_r->signature);
  __pyx_v_r->signature = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":126
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_SIG_SIGNATURE, &len)
 *         r.signature = cyares_unicode_from_uchar_and_size(cstr, len)
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)             # <<<<<<<<<<<<<<
 *         r.type = "SIG"
 *         r._attrs = ("type_covered", "algorithm", "labels", "original_ttl", "expiration", "inception", "key_tag", "signature", "signers_name", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_SIG_SIGNERS_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->signers_name);
  __Pyx_DECREF(__pyx_v_r->signers_name);
  __pyx_v_r->signers_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":127
 *         r.signature = cyares_unicode_from_uchar_and_size(cstr, len)
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)
 *         r.type = "SIG"             # <<<<<<<<<<<<<<
 *         r._attrs = ("type_covered", "algorithm", "labels", "original_ttl", "expiration", "inception", "key_tag", "signature", "signers_name", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_SIG);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_SIG);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_SIG;

  /* "resulttypes.pxd":128
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)
 *         r.type = "SIG"
 *         r._attrs = ("type_covered", "algorithm", "labels", "original_ttl", "expiration", "inception", "key_tag", "signature", "signers_name", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[3];

  /* "resulttypes.pxd":129
 *         r.type = "SIG"
 *         r._attrs = ("type_covered", "algorithm", "labels", "original_ttl", "expiration", "inception", "key_tag", "signature", "signers_name", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":110
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_sig_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_sig_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":140
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_tlsa_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_f_6cyares_11resulttypes_22ares_query_tlsa_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  size_t __pyx_v_l;
  uint8_t const *__pyx_v_cstr;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":146
 *         cdef size_t l
 *         cdef const uint8_t* cstr
 *         cdef ares_query_tlsa_result r = ares_query_tlsa_result.__new__(ares_query_tlsa_result)             # <<<<<<<<<<<<<<
 *         r.cert_usage = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_CERT_USAGE)
 *         r.selector = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_SELECTOR)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result)))) __PYX_ERR(1, 146, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":147
 *         cdef const uint8_t* cstr
 *         cdef ares_query_tlsa_result r = ares_query_tlsa_result.__new__(ares_query_tlsa_result)
 *         r.cert_usage = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_CERT_USAGE)             # <<<<<<<<<<<<<<
 *         r.selector = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_SELECTOR)
 *         r.match = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_MATCH)
*/
  __pyx_v_r->cert_usage = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_TLSA_CERT_USAGE);

  /* "resulttypes.pxd":148
 *         cdef ares_query_tlsa_result r = ares_query_tlsa_result.__new__(ares_query_tlsa_result)
 *         r.cert_usage = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_CERT_USAGE)
 *         r.selector = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_SELECTOR)             # <<<<<<<<<<<<<<
 *         r.match = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_MATCH)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_TLSA_DATA, &l)
*/
  __pyx_v_r->selector = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_TLSA_SELECTOR);

  /* "resulttypes.pxd":149
 *         r.cert_usage = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_CERT_USAGE)
 *         r.selector = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_SELECTOR)
 *         r.match = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_MATCH)             # <<<<<<<<<<<<<<
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_TLSA_DATA, &l)
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)
*/
  __pyx_v_r->match = ares_dns_rr_get_u8(__pyx_v_dns_rr, ARES_RR_TLSA_MATCH);

  /* "resulttypes.pxd":150
 *         r.selector = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_SELECTOR)
 *         r.match = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_MATCH)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_TLSA_DATA, &l)             # <<<<<<<<<<<<<<
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)
 *         r.type = "TLSA"
*/
  __pyx_v_cstr = ares_dns_rr_get_bin(__pyx_v_dns_rr, ARES_RR_TLSA_DATA, (&__pyx_v_l));

  /* "resulttypes.pxd":151
 *         r.match = ares_dns_rr_get_u8(dns_rr, ARES_RR_TLSA_MATCH)
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_TLSA_DATA, &l)
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)             # <<<<<<<<<<<<<<
 *         r.type = "TLSA"
 *         r._attrs = ("cert_usage", "selector", "match", "data", "type")
*/
  __pyx_t_1 = cyares_unicode_from_uchar_and_size(__pyx_v_cstr, __pyx_v_l); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->data);
  __Pyx_DECREF(__pyx_v_r->data);
  __pyx_v_r->data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":152
 *         cstr = ares_dns_rr_get_bin(dns_rr, ARES_RR_TLSA_DATA, &l)
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)
 *         r.type = "TLSA"             # <<<<<<<<<<<<<<
 *         r._attrs = ("cert_usage", "selector", "match", "data", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_TLSA);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_TLSA);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_TLSA;

  /* "resulttypes.pxd":153
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)
 *         r.type = "TLSA"
 *         r._attrs = ("cert_usage", "selector", "match", "data", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[4];

  /* "resulttypes.pxd":154
 *         r.type = "TLSA"
 *         r._attrs = ("cert_usage", "selector", "match", "data", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":140
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_tlsa_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_tlsa_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":165
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_svcb_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *__pyx_f_6cyares_11resulttypes_22ares_query_svcb_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":169
 *         const ares_dns_rr_t* dns_rr
 *     ):
 *         cdef ares_query_svcb_result r = ares_query_svcb_result.__new__(ares_query_svcb_result)             # <<<<<<<<<<<<<<
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_SVCB_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SVCB_TARGET)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result)))) __PYX_ERR(1, 169, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":170
 *     ):
 *         cdef ares_query_svcb_result r = ares_query_svcb_result.__new__(ares_query_svcb_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_SVCB_PRIORITY)             # <<<<<<<<<<<<<<
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SVCB_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)
*/
  __pyx_v_r->priority = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_SVCB_PRIORITY);

  /* "resulttypes.pxd":171
 *         cdef ares_query_svcb_result r = ares_query_svcb_result.__new__(ares_query_svcb_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_SVCB_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SVCB_TARGET)             # <<<<<<<<<<<<<<
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)
 *         r.type = "SVCB"
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_SVCB_TARGET); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->target);
  __Pyx_DECREF(__pyx_v_r->target);
  __pyx_v_r->target = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":172
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_SVCB_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SVCB_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)             # <<<<<<<<<<<<<<
 *         r.type = "SVCB"
 *         r._attrs = ("priority", "target", "params", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_optresults(__pyx_v_dns_rr, ARES_RR_SVCB_PARAMS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->params);
  __Pyx_DECREF(__pyx_v_r->params);
  __pyx_v_r->params = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":173
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SVCB_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)
 *         r.type = "SVCB"             # <<<<<<<<<<<<<<
 *         r._attrs = ("priority", "target", "params", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_SVCB);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_SVCB);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_SVCB;

  /* "resulttypes.pxd":174
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)
 *         r.type = "SVCB"
 *         r._attrs = ("priority", "target", "params", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[5];

  /* "resulttypes.pxd":175
 *         r.type = "SVCB"
 *         r._attrs = ("priority", "target", "params", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":165
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_svcb_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_svcb_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":187
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_https_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *__pyx_f_6cyares_11resulttypes_23ares_query_https_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":191
 *         const ares_dns_rr_t* dns_rr
 *     ):
 *         cdef ares_query_https_result r = ares_query_https_result.__new__(ares_query_https_result)             # <<<<<<<<<<<<<<
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_HTTPS_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HTTPS_TARGET)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result)))) __PYX_ERR(1, 191, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":192
 *     ):
 *         cdef ares_query_https_result r = ares_query_https_result.__new__(ares_query_https_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_HTTPS_PRIORITY)             # <<<<<<<<<<<<<<
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HTTPS_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_HTTPS_PARAMS)
*/
  __pyx_v_r->priority = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_HTTPS_PRIORITY);

  /* "resulttypes.pxd":193
 *         cdef ares_query_https_result r = ares_query_https_result.__new__(ares_query_https_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_HTTPS_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HTTPS_TARGET)             # <<<<<<<<<<<<<<
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_HTTPS_PARAMS)
 *         r.type = "HTTPS"
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_HTTPS_TARGET); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->target);
  __Pyx_DECREF(__pyx_v_r->target);
  __pyx_v_r->target = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":194
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_HTTPS_PRIORITY)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HTTPS_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_HTTPS_PARAMS)             # <<<<<<<<<<<<<<
 *         r.type = "HTTPS"
 *         r._attrs = ("priority", "target", "params", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_optresults(__pyx_v_dns_rr, ARES_RR_HTTPS_PARAMS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->params);
  __Pyx_DECREF(__pyx_v_r->params);
  __pyx_v_r->params = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":195
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HTTPS_TARGET)
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_HTTPS_PARAMS)
 *         r.type = "HTTPS"             # <<<<<<<<<<<<<<
 *         r._attrs = ("priority", "target", "params", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_HTTPS);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_HTTPS);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_HTTPS;

  /* "resulttypes.pxd":196
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_HTTPS_PARAMS)
 *         r.type = "HTTPS"
 *         r._attrs = ("priority", "target", "params", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[5];

  /* "resulttypes.pxd":197
 *         r.type = "HTTPS"
 *         r._attrs = ("priority", "target", "params", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":187
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_https_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_https_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "resulttypes.pxd":207
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_uri_result new(
 *         const ares_dns_rr_t* dns_rr
*/

static CYTHON_INLINE struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *__pyx_f_6cyares_11resulttypes_21ares_query_uri_result_new(ares_dns_rr_t const *__pyx_v_dns_rr) {
  struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *__pyx_v_r = 0;
  struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);

  /* "resulttypes.pxd":211
 *         const ares_dns_rr_t* dns_rr
 *     ):
 *         cdef ares_query_uri_result r = ares_query_uri_result.__new__(ares_query_uri_result)             # <<<<<<<<<<<<<<
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_PRIORITY)
 *         r.weight = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_WEIGHT)
*/
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result), __pyx_mstate_global->__pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result)))) __PYX_ERR(1, 211, __pyx_L1_error)
  __pyx_v_r = ((struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":212
 *     ):
 *         cdef ares_query_uri_result r = ares_query_uri_result.__new__(ares_query_uri_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_PRIORITY)             # <<<<<<<<<<<<<<
 *         r.weight = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_WEIGHT)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)
*/
  __pyx_v_r->priority = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_URI_PRIORITY);

  /* "resulttypes.pxd":213
 *         cdef ares_query_uri_result r = ares_query_uri_result.__new__(ares_query_uri_result)
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_PRIORITY)
 *         r.weight = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_WEIGHT)             # <<<<<<<<<<<<<<
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)
 *         r.type = "URI"
*/
  __pyx_v_r->weight = ares_dns_rr_get_u16(__pyx_v_dns_rr, ARES_RR_URI_WEIGHT);

  /* "resulttypes.pxd":214
 *         r.priority = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_PRIORITY)
 *         r.weight = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_WEIGHT)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)             # <<<<<<<<<<<<<<
 *         r.type = "URI"
 *         r._attrs =  ("priority", "target", "weight", "type")
*/
  __pyx_t_1 = __pyx_f_6cyares_11resulttypes_cyares_dns_rr_get_bytes(__pyx_v_dns_rr, ARES_RR_URI_TARGET); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_r->target);
  __Pyx_DECREF(__pyx_v_r->target);
  __pyx_v_r->target = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "resulttypes.pxd":215
 *         r.weight = ares_dns_rr_get_u16(dns_rr, ARES_RR_URI_WEIGHT)
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)
 *         r.type = "URI"             # <<<<<<<<<<<<<<
 *         r._attrs =  ("priority", "target", "weight", "type")
 *         return r
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_URI);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_URI);
  __Pyx_GOTREF(__pyx_v_r->type);
  __Pyx_DECREF(__pyx_v_r->type);
  __pyx_v_r->type = __pyx_mstate_global->__pyx_n_u_URI;

  /* "resulttypes.pxd":216
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)
 *         r.type = "URI"
 *         r._attrs =  ("priority", "target", "weight", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GOTREF(__pyx_v_r->__pyx_base._attrs);
  __Pyx_DECREF(__pyx_v_r->__pyx_base._attrs);
  __pyx_v_r->__pyx_base._attrs = __pyx_mstate_global->__pyx_tuple[6];

  /* "resulttypes.pxd":217
 *         r.type = "URI"
 *         r._attrs =  ("priority", "target", "weight", "type")
 *         return r             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_r);
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  /* "resulttypes.pxd":207
 *         readonly str type
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline ares_query_uri_result new(
 *         const ares_dns_rr_t* dns_rr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.resulttypes.ares_query_uri_result.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_r);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":41
 * 
 * # PYTEST PURPOSES ONLY, USING THESE IS DISCOURAGED!!!
 * def __htons(unsigned short s):             # <<<<<<<<<<<<<<
 *     return cyares_htons(s)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_1__htons(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_1__htons = {"__htons", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_1__htons, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_1__htons(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  unsigned short __pyx_v_s;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__htons (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_s,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 41, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 41, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__htons", 0) < 0) __PYX_ERR(0, 41, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__htons", 1, 1, 1, i); __PYX_ERR(0, 41, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 41, __pyx_L3_error)
    }
    __pyx_v_s = __Pyx_PyLong_As_unsigned_short(values[0]); if (unlikely((__pyx_v_s == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 41, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__htons", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 41, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.__htons", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel___htons(__pyx_self, __pyx_v_s);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel___htons(CYTHON_UNUSED PyObject *__pyx_self, unsigned short __pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__htons", 0);

  /* "cyares/channel.pyx":42
 * # PYTEST PURPOSES ONLY, USING THESE IS DISCOURAGED!!!
 * def __htons(unsigned short s):
 *     return cyares_htons(s)             # <<<<<<<<<<<<<<
 * 
 * def __htonl(unsigned long l):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(cyares_htons(__pyx_v_s)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":41
 * 
 * # PYTEST PURPOSES ONLY, USING THESE IS DISCOURAGED!!!
 * def __htons(unsigned short s):             # <<<<<<<<<<<<<<
 *     return cyares_htons(s)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.__htons", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":44
 *     return cyares_htons(s)
 * 
 * def __htonl(unsigned long l):             # <<<<<<<<<<<<<<
 *     return cyares_htonl(l)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_3__htonl(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_3__htonl = {"__htonl", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_3__htonl, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_3__htonl(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  unsigned long __pyx_v_l;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__htonl (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_l,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 44, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 44, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__htonl", 0) < 0) __PYX_ERR(0, 44, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__htonl", 1, 1, 1, i); __PYX_ERR(0, 44, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 44, __pyx_L3_error)
    }
    __pyx_v_l = __Pyx_PyLong_As_unsigned_long(values[0]); if (unlikely((__pyx_v_l == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 44, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__htonl", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 44, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.__htonl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_2__htonl(__pyx_self, __pyx_v_l);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_2__htonl(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_l) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__htonl", 0);

  /* "cyares/channel.pyx":45
 * 
 * def __htonl(unsigned long l):
 *     return cyares_htonl(l)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(cyares_htonl(__pyx_v_l)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":44
 *     return cyares_htons(s)
 * 
 * def __htonl(unsigned long l):             # <<<<<<<<<<<<<<
 *     return cyares_htonl(l)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.__htonl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":48
 * 
 * 
 * cdef inline void raise_negative_timeout_err(object obj):             # <<<<<<<<<<<<<<
 *     PyErr_SetObject(ValueError, f"timeout should not be below 0 got {obj}" )
 * 
*/

static CYTHON_INLINE void __pyx_f_6cyares_7channel_raise_negative_timeout_err(PyObject *__pyx_v_obj) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("raise_negative_timeout_err", 0);

  /* "cyares/channel.pyx":49
 * 
 * cdef inline void raise_negative_timeout_err(object obj):
 *     PyErr_SetObject(ValueError, f"timeout should not be below 0 got {obj}" )             # <<<<<<<<<<<<<<
 * 
 * cdef int cyares_seconds_to_milliseconds(object obj) except -1:
*/
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_obj, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_timeout_should_not_be_below_0_go, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyErr_SetObject(__pyx_builtin_ValueError, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":48
 * 
 * 
 * cdef inline void raise_negative_timeout_err(object obj):             # <<<<<<<<<<<<<<
 *     PyErr_SetObject(ValueError, f"timeout should not be below 0 got {obj}" )
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cyares.channel.raise_negative_timeout_err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "cyares/channel.pyx":51
 *     PyErr_SetObject(ValueError, f"timeout should not be below 0 got {obj}" )
 * 
 * cdef int cyares_seconds_to_milliseconds(object obj) except -1:             # <<<<<<<<<<<<<<
 *     cdef double _d
 *     if isinstance(obj, float):
*/

static int __pyx_f_6cyares_7channel_cyares_seconds_to_milliseconds(PyObject *__pyx_v_obj) {
  double __pyx_v__d;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cyares_seconds_to_milliseconds", 0);

  /* "cyares/channel.pyx":53
 * cdef int cyares_seconds_to_milliseconds(object obj) except -1:
 *     cdef double _d
 *     if isinstance(obj, float):             # <<<<<<<<<<<<<<
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
*/
  __pyx_t_1 = PyFloat_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":54
 *     cdef double _d
 *     if isinstance(obj, float):
 *         if obj < 0:             # <<<<<<<<<<<<<<
 *             raise_negative_timeout_err(obj)
 *             return -1
*/
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_obj, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "cyares/channel.pyx":55
 *     if isinstance(obj, float):
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)             # <<<<<<<<<<<<<<
 *             return -1
 * 
*/
      __pyx_f_6cyares_7channel_raise_negative_timeout_err(__pyx_v_obj); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 55, __pyx_L1_error)

      /* "cyares/channel.pyx":56
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
 *             return -1             # <<<<<<<<<<<<<<
 * 
 *         _d = <double>obj
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyares/channel.pyx":54
 *     cdef double _d
 *     if isinstance(obj, float):
 *         if obj < 0:             # <<<<<<<<<<<<<<
 *             raise_negative_timeout_err(obj)
 *             return -1
*/
    }

    /* "cyares/channel.pyx":58
 *             return -1
 * 
 *         _d = <double>obj             # <<<<<<<<<<<<<<
 *         return <int>floor(_d) + <int>(fmod(_d, 1.0) * 1000)
 * 
*/
    __pyx_t_3 = __Pyx_PyFloat_AsDouble(__pyx_v_obj); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_v__d = ((double)__pyx_t_3);

    /* "cyares/channel.pyx":59
 * 
 *         _d = <double>obj
 *         return <int>floor(_d) + <int>(fmod(_d, 1.0) * 1000)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_r = (((int)floor(__pyx_v__d)) + ((int)(fmod(__pyx_v__d, 1.0) * 1000.0)));
    goto __pyx_L0;

    /* "cyares/channel.pyx":53
 * cdef int cyares_seconds_to_milliseconds(object obj) except -1:
 *     cdef double _d
 *     if isinstance(obj, float):             # <<<<<<<<<<<<<<
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
*/
  }

  /* "cyares/channel.pyx":62
 * 
 * 
 *     elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
*/
  __pyx_t_1 = PyLong_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":63
 * 
 *     elif isinstance(obj, int):
 *         if obj < 0:             # <<<<<<<<<<<<<<
 *             raise_negative_timeout_err(obj)
 *             return -1
*/
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_obj, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "cyares/channel.pyx":64
 *     elif isinstance(obj, int):
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)             # <<<<<<<<<<<<<<
 *             return -1
 *         return <int>(obj * 1000)
*/
      __pyx_f_6cyares_7channel_raise_negative_timeout_err(__pyx_v_obj); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)

      /* "cyares/channel.pyx":65
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
 *             return -1             # <<<<<<<<<<<<<<
 *         return <int>(obj * 1000)
 *     else:
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "cyares/channel.pyx":63
 * 
 *     elif isinstance(obj, int):
 *         if obj < 0:             # <<<<<<<<<<<<<<
 *             raise_negative_timeout_err(obj)
 *             return -1
*/
    }

    /* "cyares/channel.pyx":66
 *             raise_negative_timeout_err(obj)
 *             return -1
 *         return <int>(obj * 1000)             # <<<<<<<<<<<<<<
 *     else:
 *         PyErr_SetObject(TypeError, "timeout must be an integer or float not %s" % type(obj).__name__)
*/
    __pyx_t_2 = __Pyx_PyLong_MultiplyObjC(__pyx_v_obj, __pyx_mstate_global->__pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((int)__pyx_t_4);
    goto __pyx_L0;

    /* "cyares/channel.pyx":62
 * 
 * 
 *     elif isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         if obj < 0:
 *             raise_negative_timeout_err(obj)
*/
  }

  /* "cyares/channel.pyx":68
 *         return <int>(obj * 1000)
 *     else:
 *         PyErr_SetObject(TypeError, "timeout must be an integer or float not %s" % type(obj).__name__)             # <<<<<<<<<<<<<<
 *         return -1
 * 
*/
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_timeout_must_be_an_integer_or_fl, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    PyErr_SetObject(__pyx_builtin_TypeError, __pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "cyares/channel.pyx":69
 *     else:
 *         PyErr_SetObject(TypeError, "timeout must be an integer or float not %s" % type(obj).__name__)
 *         return -1             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_r = -1;
    goto __pyx_L0;
  }

  /* "cyares/channel.pyx":51
 *     PyErr_SetObject(ValueError, f"timeout should not be below 0 got {obj}" )
 * 
 * cdef int cyares_seconds_to_milliseconds(object obj) except -1:             # <<<<<<<<<<<<<<
 *     cdef double _d
 *     if isinstance(obj, float):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyares.channel.cyares_seconds_to_milliseconds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":74
 * 
 * 
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:             # <<<<<<<<<<<<<<
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:
*/

static int __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(PyObject *__pyx_v_obj, Py_buffer *__pyx_v_view) {
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cyares_get_domain_name_buffer", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "cyares/channel.pyx":75
 * 
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:
 *     if PyUnicode_Check(obj):             # <<<<<<<<<<<<<<
 *         if HAS_IDNA2008:
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):
*/
  __pyx_t_1 = PyUnicode_Check(__pyx_v_obj);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":76
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:             # <<<<<<<<<<<<<<
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):
 * 
*/
    if (__pyx_v_6cyares_7channel_HAS_IDNA2008) {

      /* "cyares/channel.pyx":77
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):             # <<<<<<<<<<<<<<
 * 
 *              # Not going to risk users using earlier versions of idna even if idna is newer
*/
      __pyx_t_1 = (!PyUnicode_IS_COMPACT_ASCII(__pyx_v_obj));
      if (__pyx_t_1) {

        /* "cyares/channel.pyx":86
 *              # SEE: https://github.com/kjd/idna/security/advisories/GHSA-jjg7-2v4v-x38h
 * 
 *                 if PyUnicode_GetLength(obj) > 253:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(ValueError, "Domain names being idna decoded should not exceed a size of 255")
 *                     return -1
*/
        __pyx_t_2 = PyUnicode_GetLength(__pyx_v_obj); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 86, __pyx_L1_error)
        __pyx_t_1 = (__pyx_t_2 > 0xFD);
        if (__pyx_t_1) {

          /* "cyares/channel.pyx":87
 * 
 *                 if PyUnicode_GetLength(obj) > 253:
 *                     PyErr_SetObject(ValueError, "Domain names being idna decoded should not exceed a size of 255")             # <<<<<<<<<<<<<<
 *                     return -1
 * 
*/
          PyErr_SetObject(__pyx_builtin_ValueError, __pyx_mstate_global->__pyx_kp_u_Domain_names_being_idna_decoded);

          /* "cyares/channel.pyx":88
 *                 if PyUnicode_GetLength(obj) > 253:
 *                     PyErr_SetObject(ValueError, "Domain names being idna decoded should not exceed a size of 255")
 *                     return -1             # <<<<<<<<<<<<<<
 * 
 *                 try:
*/
          __pyx_r = -1;
          goto __pyx_L0;

          /* "cyares/channel.pyx":86
 *              # SEE: https://github.com/kjd/idna/security/advisories/GHSA-jjg7-2v4v-x38h
 * 
 *                 if PyUnicode_GetLength(obj) > 253:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(ValueError, "Domain names being idna decoded should not exceed a size of 255")
 *                     return -1
*/
        }

        /* "cyares/channel.pyx":90
 *                     return -1
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = idna_decode(obj)
 * 
*/
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_5);
          /*try:*/ {

            /* "cyares/channel.pyx":91
 * 
 *                 try:
 *                     obj = idna_decode(obj)             # <<<<<<<<<<<<<<
 * 
 *                 except Exception as e:
*/
            __pyx_t_7 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_idna_decode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 91, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_9 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_8))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
              assert(__pyx_t_7);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
              __pyx_t_9 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_obj};
              __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 91, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_6);
            }
            __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "cyares/channel.pyx":90
 *                     return -1
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = idna_decode(obj)
 * 
*/
          }
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L12_try_end;
          __pyx_L7_error:;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "cyares/channel.pyx":93
 *                     obj = idna_decode(obj)
 * 
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(e, str(e))
 *                     return -1
*/
          __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
          if (__pyx_t_10) {
            __Pyx_AddTraceback("cyares.channel.cyares_get_domain_name_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(0, 93, __pyx_L9_except_error)
            __Pyx_XGOTREF(__pyx_t_6);
            __Pyx_XGOTREF(__pyx_t_8);
            __Pyx_XGOTREF(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __pyx_v_e = __pyx_t_8;
            /*try:*/ {

              /* "cyares/channel.pyx":94
 * 
 *                 except Exception as e:
 *                     PyErr_SetObject(e, str(e))             # <<<<<<<<<<<<<<
 *                     return -1
 *             else:
*/
              __pyx_t_11 = __Pyx_PyObject_Unicode(__pyx_v_e); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 94, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_11);
              PyErr_SetObject(__pyx_v_e, __pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "cyares/channel.pyx":95
 *                 except Exception as e:
 *                     PyErr_SetObject(e, str(e))
 *                     return -1             # <<<<<<<<<<<<<<
 *             else:
 *                 # Python should have it's own backup
*/
              __pyx_r = -1;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              goto __pyx_L17_return;
            }

            /* "cyares/channel.pyx":93
 *                     obj = idna_decode(obj)
 * 
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(e, str(e))
 *                     return -1
*/
            /*finally:*/ {
              __pyx_L18_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                 __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
                if ( unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
                __Pyx_XGOTREF(__pyx_t_14);
                __Pyx_XGOTREF(__pyx_t_15);
                __Pyx_XGOTREF(__pyx_t_16);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __pyx_t_10 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
                }
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_14);
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
                __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
                __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
                goto __pyx_L9_except_error;
              }
              __pyx_L17_return: {
                __pyx_t_12 = __pyx_r;
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
                __pyx_r = __pyx_t_12;
                goto __pyx_L10_except_return;
              }
            }
          }
          goto __pyx_L9_except_error;

          /* "cyares/channel.pyx":90
 *                     return -1
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = idna_decode(obj)
 * 
*/
          __pyx_L9_except_error:;
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
          goto __pyx_L1_error;
          __pyx_L10_except_return:;
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
          goto __pyx_L0;
          __pyx_L12_try_end:;
        }

        /* "cyares/channel.pyx":77
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):             # <<<<<<<<<<<<<<
 * 
 *              # Not going to risk users using earlier versions of idna even if idna is newer
*/
        goto __pyx_L5;
      }

      /* "cyares/channel.pyx":98
 *             else:
 *                 # Python should have it's own backup
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = obj.encode("idna")
 *                 except Exception as e:
*/
      /*else*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_3);
          /*try:*/ {

            /* "cyares/channel.pyx":99
 *                 # Python should have it's own backup
 *                 try:
 *                     obj = obj.encode("idna")             # <<<<<<<<<<<<<<
 *                 except Exception as e:
 *                     PyErr_SetObject(e, str(e))
*/
            __pyx_t_8 = __pyx_v_obj;
            __Pyx_INCREF(__pyx_t_8);
            __pyx_t_9 = 0;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_mstate_global->__pyx_n_u_idna};
              __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_7);
            }
            __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "cyares/channel.pyx":98
 *             else:
 *                 # Python should have it's own backup
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = obj.encode("idna")
 *                 except Exception as e:
*/
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L29_try_end;
          __pyx_L24_error:;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "cyares/channel.pyx":100
 *                 try:
 *                     obj = obj.encode("idna")
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(e, str(e))
 *                     return -1
*/
          __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
          if (__pyx_t_12) {
            __Pyx_AddTraceback("cyares.channel.cyares_get_domain_name_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(0, 100, __pyx_L26_except_error)
            __Pyx_XGOTREF(__pyx_t_7);
            __Pyx_XGOTREF(__pyx_t_8);
            __Pyx_XGOTREF(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_8);
            __pyx_v_e = __pyx_t_8;
            /*try:*/ {

              /* "cyares/channel.pyx":101
 *                     obj = obj.encode("idna")
 *                 except Exception as e:
 *                     PyErr_SetObject(e, str(e))             # <<<<<<<<<<<<<<
 *                     return -1
 *     return cyares_get_buffer(obj, view)
*/
              __pyx_t_11 = __Pyx_PyObject_Unicode(__pyx_v_e); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 101, __pyx_L35_error)
              __Pyx_GOTREF(__pyx_t_11);
              PyErr_SetObject(__pyx_v_e, __pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "cyares/channel.pyx":102
 *                 except Exception as e:
 *                     PyErr_SetObject(e, str(e))
 *                     return -1             # <<<<<<<<<<<<<<
 *     return cyares_get_buffer(obj, view)
 * 
*/
              __pyx_r = -1;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              goto __pyx_L34_return;
            }

            /* "cyares/channel.pyx":100
 *                 try:
 *                     obj = obj.encode("idna")
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     PyErr_SetObject(e, str(e))
 *                     return -1
*/
            /*finally:*/ {
              __pyx_L35_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0;
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                 __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
                if ( unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_16);
                __Pyx_XGOTREF(__pyx_t_15);
                __Pyx_XGOTREF(__pyx_t_14);
                __pyx_t_12 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
                }
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_14);
                __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_15, __pyx_t_14);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_ErrRestore(__pyx_t_19, __pyx_t_18, __pyx_t_17);
                __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0;
                __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_20;
                goto __pyx_L26_except_error;
              }
              __pyx_L34_return: {
                __pyx_t_10 = __pyx_r;
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
                __pyx_r = __pyx_t_10;
                goto __pyx_L27_except_return;
              }
            }
          }
          goto __pyx_L26_except_error;

          /* "cyares/channel.pyx":98
 *             else:
 *                 # Python should have it's own backup
 *                 try:             # <<<<<<<<<<<<<<
 *                     obj = obj.encode("idna")
 *                 except Exception as e:
*/
          __pyx_L26_except_error:;
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
          goto __pyx_L1_error;
          __pyx_L27_except_return:;
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
          goto __pyx_L0;
          __pyx_L29_try_end:;
        }
      }
      __pyx_L5:;

      /* "cyares/channel.pyx":76
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:             # <<<<<<<<<<<<<<
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):
 * 
*/
    }

    /* "cyares/channel.pyx":75
 * 
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:
 *     if PyUnicode_Check(obj):             # <<<<<<<<<<<<<<
 *         if HAS_IDNA2008:
 *             if not PyUnicode_IS_COMPACT_ASCII(obj):
*/
  }

  /* "cyares/channel.pyx":103
 *                     PyErr_SetObject(e, str(e))
 *                     return -1
 *     return cyares_get_buffer(obj, view)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_10 = cyares_get_buffer(__pyx_v_obj, __pyx_v_view); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 103, __pyx_L1_error)
  __pyx_r = __pyx_t_10;
  goto __pyx_L0;

  /* "cyares/channel.pyx":74
 * 
 * 
 * cdef int cyares_get_domain_name_buffer(object obj, Py_buffer* view) except -1:             # <<<<<<<<<<<<<<
 *     if PyUnicode_Check(obj):
 *         if HAS_IDNA2008:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cyares.channel.cyares_get_domain_name_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":138
 * 
 * cdef class Channel:
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         flags = None,
*/

/* Python wrapper */
static int __pyx_pw_6cyares_7channel_7Channel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6cyares_7channel_7Channel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_timeout = 0;
  PyObject *__pyx_v_tries = 0;
  PyObject *__pyx_v_ndots = 0;
  PyObject *__pyx_v_tcp_port = 0;
  PyObject *__pyx_v_udp_port = 0;
  PyObject *__pyx_v_servers = 0;
  PyObject *__pyx_v_domains = 0;
  PyObject *__pyx_v_lookups = 0;
  PyObject *__pyx_v_sock_state_cb = 0;
  PyObject *__pyx_v_socket_send_buffer_size = 0;
  PyObject *__pyx_v_socket_receive_buffer_size = 0;
  int __pyx_v_rotate;
  PyObject *__pyx_v_local_ip = 0;
  PyObject *__pyx_v_local_dev = 0;
  PyObject *__pyx_v_resolvconf_path = 0;
  int __pyx_v_event_thread;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_timeout,&__pyx_mstate_global->__pyx_n_u_tries,&__pyx_mstate_global->__pyx_n_u_ndots,&__pyx_mstate_global->__pyx_n_u_tcp_port,&__pyx_mstate_global->__pyx_n_u_udp_port,&__pyx_mstate_global->__pyx_n_u_servers,&__pyx_mstate_global->__pyx_n_u_domains,&__pyx_mstate_global->__pyx_n_u_lookups,&__pyx_mstate_global->__pyx_n_u_sock_state_cb,&__pyx_mstate_global->__pyx_n_u_socket_send_buffer_size,&__pyx_mstate_global->__pyx_n_u_socket_receive_buffer_size,&__pyx_mstate_global->__pyx_n_u_rotate,&__pyx_mstate_global->__pyx_n_u_local_ip,&__pyx_mstate_global->__pyx_n_u_local_dev,&__pyx_mstate_global->__pyx_n_u_resolvconf_path,&__pyx_mstate_global->__pyx_n_u_event_thread,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 138, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 17:
        values[16] = __Pyx_ArgRef_VARARGS(__pyx_args, 16);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[16])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 16:
        values[15] = __Pyx_ArgRef_VARARGS(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 138, __pyx_L3_error)

      /* "cyares/channel.pyx":140
 *     def __init__(
 *         self,
 *         flags = None,             # <<<<<<<<<<<<<<
 *         timeout = None,
 *         tries = None,
*/
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":141
 *         self,
 *         flags = None,
 *         timeout = None,             # <<<<<<<<<<<<<<
 *         tries = None,
 *         ndots = None,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":142
 *         flags = None,
 *         timeout = None,
 *         tries = None,             # <<<<<<<<<<<<<<
 *         ndots = None,
 *         tcp_port = None,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":143
 *         timeout = None,
 *         tries = None,
 *         ndots = None,             # <<<<<<<<<<<<<<
 *         tcp_port = None,
 *         udp_port = None,
*/
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":144
 *         tries = None,
 *         ndots = None,
 *         tcp_port = None,             # <<<<<<<<<<<<<<
 *         udp_port = None,
 *         servers = None,
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":145
 *         ndots = None,
 *         tcp_port = None,
 *         udp_port = None,             # <<<<<<<<<<<<<<
 *         servers = None,
 *         domains = None,
*/
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":146
 *         tcp_port = None,
 *         udp_port = None,
 *         servers = None,             # <<<<<<<<<<<<<<
 *         domains = None,
 *         lookups = None,
*/
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":147
 *         udp_port = None,
 *         servers = None,
 *         domains = None,             # <<<<<<<<<<<<<<
 *         lookups = None,
 *         sock_state_cb = None,
*/
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":148
 *         servers = None,
 *         domains = None,
 *         lookups = None,             # <<<<<<<<<<<<<<
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,
*/
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":149
 *         domains = None,
 *         lookups = None,
 *         sock_state_cb = None,             # <<<<<<<<<<<<<<
 *         socket_send_buffer_size = None,
 *         socket_receive_buffer_size = None,
*/
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":150
 *         lookups = None,
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,             # <<<<<<<<<<<<<<
 *         socket_receive_buffer_size = None,
 *         bint rotate = False,
*/
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":151
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,
 *         socket_receive_buffer_size = None,             # <<<<<<<<<<<<<<
 *         bint rotate = False,
 *         local_ip = None,
*/
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":153
 *         socket_receive_buffer_size = None,
 *         bint rotate = False,
 *         local_ip = None,             # <<<<<<<<<<<<<<
 *         local_dev = None,
 *         resolvconf_path = None,
*/
      if (!values[13]) values[13] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":154
 *         bint rotate = False,
 *         local_ip = None,
 *         local_dev = None,             # <<<<<<<<<<<<<<
 *         resolvconf_path = None,
 *         bint event_thread = False
*/
      if (!values[14]) values[14] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":155
 *         local_ip = None,
 *         local_dev = None,
 *         resolvconf_path = None,             # <<<<<<<<<<<<<<
 *         bint event_thread = False
 *     ):
*/
      if (!values[15]) values[15] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case 17:
        values[16] = __Pyx_ArgRef_VARARGS(__pyx_args, 16);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[16])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 16:
        values[15] = __Pyx_ArgRef_VARARGS(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "cyares/channel.pyx":140
 *     def __init__(
 *         self,
 *         flags = None,             # <<<<<<<<<<<<<<
 *         timeout = None,
 *         tries = None,
*/
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":141
 *         self,
 *         flags = None,
 *         timeout = None,             # <<<<<<<<<<<<<<
 *         tries = None,
 *         ndots = None,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":142
 *         flags = None,
 *         timeout = None,
 *         tries = None,             # <<<<<<<<<<<<<<
 *         ndots = None,
 *         tcp_port = None,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":143
 *         timeout = None,
 *         tries = None,
 *         ndots = None,             # <<<<<<<<<<<<<<
 *         tcp_port = None,
 *         udp_port = None,
*/
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":144
 *         tries = None,
 *         ndots = None,
 *         tcp_port = None,             # <<<<<<<<<<<<<<
 *         udp_port = None,
 *         servers = None,
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":145
 *         ndots = None,
 *         tcp_port = None,
 *         udp_port = None,             # <<<<<<<<<<<<<<
 *         servers = None,
 *         domains = None,
*/
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":146
 *         tcp_port = None,
 *         udp_port = None,
 *         servers = None,             # <<<<<<<<<<<<<<
 *         domains = None,
 *         lookups = None,
*/
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":147
 *         udp_port = None,
 *         servers = None,
 *         domains = None,             # <<<<<<<<<<<<<<
 *         lookups = None,
 *         sock_state_cb = None,
*/
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":148
 *         servers = None,
 *         domains = None,
 *         lookups = None,             # <<<<<<<<<<<<<<
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,
*/
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":149
 *         domains = None,
 *         lookups = None,
 *         sock_state_cb = None,             # <<<<<<<<<<<<<<
 *         socket_send_buffer_size = None,
 *         socket_receive_buffer_size = None,
*/
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":150
 *         lookups = None,
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,             # <<<<<<<<<<<<<<
 *         socket_receive_buffer_size = None,
 *         bint rotate = False,
*/
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":151
 *         sock_state_cb = None,
 *         socket_send_buffer_size = None,
 *         socket_receive_buffer_size = None,             # <<<<<<<<<<<<<<
 *         bint rotate = False,
 *         local_ip = None,
*/
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":153
 *         socket_receive_buffer_size = None,
 *         bint rotate = False,
 *         local_ip = None,             # <<<<<<<<<<<<<<
 *         local_dev = None,
 *         resolvconf_path = None,
*/
      if (!values[13]) values[13] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":154
 *         bint rotate = False,
 *         local_ip = None,
 *         local_dev = None,             # <<<<<<<<<<<<<<
 *         resolvconf_path = None,
 *         bint event_thread = False
*/
      if (!values[14]) values[14] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":155
 *         local_ip = None,
 *         local_dev = None,
 *         resolvconf_path = None,             # <<<<<<<<<<<<<<
 *         bint event_thread = False
 *     ):
*/
      if (!values[15]) values[15] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_flags = values[0];
    __pyx_v_timeout = values[1];
    __pyx_v_tries = values[2];
    __pyx_v_ndots = values[3];
    __pyx_v_tcp_port = values[4];
    __pyx_v_udp_port = values[5];
    __pyx_v_servers = values[6];
    __pyx_v_domains = values[7];
    __pyx_v_lookups = values[8];
    __pyx_v_sock_state_cb = values[9];
    __pyx_v_socket_send_buffer_size = values[10];
    __pyx_v_socket_receive_buffer_size = values[11];
    if (values[12]) {
      __pyx_v_rotate = __Pyx_PyObject_IsTrue(values[12]); if (unlikely((__pyx_v_rotate == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 152, __pyx_L3_error)
    } else {

      /* "cyares/channel.pyx":152
 *         socket_send_buffer_size = None,
 *         socket_receive_buffer_size = None,
 *         bint rotate = False,             # <<<<<<<<<<<<<<
 *         local_ip = None,
 *         local_dev = None,
*/
      __pyx_v_rotate = ((int)0);
    }
    __pyx_v_local_ip = values[13];
    __pyx_v_local_dev = values[14];
    __pyx_v_resolvconf_path = values[15];
    if (values[16]) {
      __pyx_v_event_thread = __Pyx_PyObject_IsTrue(values[16]); if (unlikely((__pyx_v_event_thread == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L3_error)
    } else {

      /* "cyares/channel.pyx":156
 *         local_dev = None,
 *         resolvconf_path = None,
 *         bint event_thread = False             # <<<<<<<<<<<<<<
 *     ):
 *         cdef Py_buffer view
*/
      __pyx_v_event_thread = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 17, __pyx_nargs); __PYX_ERR(0, 138, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel___init__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_flags, __pyx_v_timeout, __pyx_v_tries, __pyx_v_ndots, __pyx_v_tcp_port, __pyx_v_udp_port, __pyx_v_servers, __pyx_v_domains, __pyx_v_lookups, __pyx_v_sock_state_cb, __pyx_v_socket_send_buffer_size, __pyx_v_socket_receive_buffer_size, __pyx_v_rotate, __pyx_v_local_ip, __pyx_v_local_dev, __pyx_v_resolvconf_path, __pyx_v_event_thread);

  /* "cyares/channel.pyx":138
 * 
 * cdef class Channel:
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         flags = None,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6cyares_7channel_7Channel___init__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_flags, PyObject *__pyx_v_timeout, PyObject *__pyx_v_tries, PyObject *__pyx_v_ndots, PyObject *__pyx_v_tcp_port, PyObject *__pyx_v_udp_port, PyObject *__pyx_v_servers, PyObject *__pyx_v_domains, PyObject *__pyx_v_lookups, PyObject *__pyx_v_sock_state_cb, PyObject *__pyx_v_socket_send_buffer_size, PyObject *__pyx_v_socket_receive_buffer_size, int __pyx_v_rotate, PyObject *__pyx_v_local_ip, PyObject *__pyx_v_local_dev, PyObject *__pyx_v_resolvconf_path, int __pyx_v_event_thread) {
  Py_buffer __pyx_v_view;
  int __pyx_v_optmask;
  char **__pyx_v_strs;
  PyObject *__pyx_v_i = 0;
  int __pyx_v_r;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  unsigned short __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cyares/channel.pyx":159
 *     ):
 *         cdef Py_buffer view
 *         cdef int optmask = 0             # <<<<<<<<<<<<<<
 *         cdef char** strs = NULL
 *         cdef object i
*/
  __pyx_v_optmask = 0;

  /* "cyares/channel.pyx":160
 *         cdef Py_buffer view
 *         cdef int optmask = 0
 *         cdef char** strs = NULL             # <<<<<<<<<<<<<<
 *         cdef object i
 *         self.event_thread = event_thread
*/
  __pyx_v_strs = NULL;

  /* "cyares/channel.pyx":162
 *         cdef char** strs = NULL
 *         cdef object i
 *         self.event_thread = event_thread             # <<<<<<<<<<<<<<
 * 
 *         self._cancelled = False
*/
  __pyx_v_self->event_thread = __pyx_v_event_thread;

  /* "cyares/channel.pyx":164
 *         self.event_thread = event_thread
 * 
 *         self._cancelled = False             # <<<<<<<<<<<<<<
 *         self.handles = set()
 *         self._query_lookups = {
*/
  __pyx_v_self->_cancelled = 0;

  /* "cyares/channel.pyx":165
 * 
 *         self._cancelled = False
 *         self.handles = set()             # <<<<<<<<<<<<<<
 *         self._query_lookups = {
 *             "A":T_A,
*/
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->handles);
  __Pyx_DECREF(__pyx_v_self->handles);
  __pyx_v_self->handles = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":167
 *         self.handles = set()
 *         self._query_lookups = {
 *             "A":T_A,             # <<<<<<<<<<<<<<
 *             "AAAA":T_AAAA,
 *             "ANY":T_ANY,
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int(T_A); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_A, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":168
 *         self._query_lookups = {
 *             "A":T_A,
 *             "AAAA":T_AAAA,             # <<<<<<<<<<<<<<
 *             "ANY":T_ANY,
 *             "CAA":T_CAA,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_AAAA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_AAAA, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":169
 *             "A":T_A,
 *             "AAAA":T_AAAA,
 *             "ANY":T_ANY,             # <<<<<<<<<<<<<<
 *             "CAA":T_CAA,
 *             "CNAME":T_CNAME,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_ANY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ANY, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":170
 *             "AAAA":T_AAAA,
 *             "ANY":T_ANY,
 *             "CAA":T_CAA,             # <<<<<<<<<<<<<<
 *             "CNAME":T_CNAME,
 *             "MX":T_MX,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_CAA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_CAA, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":171
 *             "ANY":T_ANY,
 *             "CAA":T_CAA,
 *             "CNAME":T_CNAME,             # <<<<<<<<<<<<<<
 *             "MX":T_MX,
 *             "NAPTR":T_NAPTR,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_CNAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_CNAME, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":172
 *             "CAA":T_CAA,
 *             "CNAME":T_CNAME,
 *             "MX":T_MX,             # <<<<<<<<<<<<<<
 *             "NAPTR":T_NAPTR,
 *             "NS":T_NS,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_MX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_MX, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":173
 *             "CNAME":T_CNAME,
 *             "MX":T_MX,
 *             "NAPTR":T_NAPTR,             # <<<<<<<<<<<<<<
 *             "NS":T_NS,
 *             "PTR":T_PTR,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_NAPTR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_NAPTR, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":174
 *             "MX":T_MX,
 *             "NAPTR":T_NAPTR,
 *             "NS":T_NS,             # <<<<<<<<<<<<<<
 *             "PTR":T_PTR,
 *             "SOA":T_SOA,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_NS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_NS, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":175
 *             "NAPTR":T_NAPTR,
 *             "NS":T_NS,
 *             "PTR":T_PTR,             # <<<<<<<<<<<<<<
 *             "SOA":T_SOA,
 *             "SRV":T_SRV,
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_PTR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PTR, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":176
 *             "NS":T_NS,
 *             "PTR":T_PTR,
 *             "SOA":T_SOA,             # <<<<<<<<<<<<<<
 *             "SRV":T_SRV,
 *             "TXT":T_TXT
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_SOA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_SOA, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":177
 *             "PTR":T_PTR,
 *             "SOA":T_SOA,
 *             "SRV":T_SRV,             # <<<<<<<<<<<<<<
 *             "TXT":T_TXT
 *         }
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_SRV); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_SRV, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":178
 *             "SOA":T_SOA,
 *             "SRV":T_SRV,
 *             "TXT":T_TXT             # <<<<<<<<<<<<<<
 *         }
 *         self._closed = 0
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(T_TXT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_TXT, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyares/channel.pyx":166
 *         self._cancelled = False
 *         self.handles = set()
 *         self._query_lookups = {             # <<<<<<<<<<<<<<
 *             "A":T_A,
 *             "AAAA":T_AAAA,
*/
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_query_lookups);
  __Pyx_DECREF(__pyx_v_self->_query_lookups);
  __pyx_v_self->_query_lookups = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":180
 *             "TXT":T_TXT
 *         }
 *         self._closed = 0             # <<<<<<<<<<<<<<
 *         self._running = 0
 * 
*/
  __pyx_v_self->_closed = 0;

  /* "cyares/channel.pyx":181
 *         }
 *         self._closed = 0
 *         self._running = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags is not None:
*/
  __pyx_v_self->_running = 0;

  /* "cyares/channel.pyx":183
 *         self._running = 0
 * 
 *         if flags is not None:             # <<<<<<<<<<<<<<
 *             self.options.flags = flags
 *             optmask |= ARES_OPT_FLAGS
*/
  __pyx_t_3 = (__pyx_v_flags != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":184
 * 
 *         if flags is not None:
 *             self.options.flags = flags             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_FLAGS
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
    __pyx_v_self->options.flags = __pyx_t_4;

    /* "cyares/channel.pyx":185
 *         if flags is not None:
 *             self.options.flags = flags
 *             optmask |= ARES_OPT_FLAGS             # <<<<<<<<<<<<<<
 * 
 *         if timeout is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_FLAGS);

    /* "cyares/channel.pyx":183
 *         self._running = 0
 * 
 *         if flags is not None:             # <<<<<<<<<<<<<<
 *             self.options.flags = flags
 *             optmask |= ARES_OPT_FLAGS
*/
  }

  /* "cyares/channel.pyx":187
 *             optmask |= ARES_OPT_FLAGS
 * 
 *         if timeout is not None:             # <<<<<<<<<<<<<<
 *             self.options.timeout = int(timeout * 1000)
 *             optmask |= ARES_OPT_TIMEOUTMS
*/
  __pyx_t_3 = (__pyx_v_timeout != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":188
 * 
 *         if timeout is not None:
 *             self.options.timeout = int(timeout * 1000)             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_TIMEOUTMS
 * 
*/
    __pyx_t_1 = __Pyx_PyLong_MultiplyObjC(__pyx_v_timeout, __pyx_mstate_global->__pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->options.timeout = __pyx_t_4;

    /* "cyares/channel.pyx":189
 *         if timeout is not None:
 *             self.options.timeout = int(timeout * 1000)
 *             optmask |= ARES_OPT_TIMEOUTMS             # <<<<<<<<<<<<<<
 * 
 *         if tries is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_TIMEOUTMS);

    /* "cyares/channel.pyx":187
 *             optmask |= ARES_OPT_FLAGS
 * 
 *         if timeout is not None:             # <<<<<<<<<<<<<<
 *             self.options.timeout = int(timeout * 1000)
 *             optmask |= ARES_OPT_TIMEOUTMS
*/
  }

  /* "cyares/channel.pyx":191
 *             optmask |= ARES_OPT_TIMEOUTMS
 * 
 *         if tries is not None:             # <<<<<<<<<<<<<<
 *             self.options.tries = tries
 *             optmask |= ARES_OPT_TRIES
*/
  __pyx_t_3 = (__pyx_v_tries != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":192
 * 
 *         if tries is not None:
 *             self.options.tries = tries             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_TRIES
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_tries); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 192, __pyx_L1_error)
    __pyx_v_self->options.tries = __pyx_t_4;

    /* "cyares/channel.pyx":193
 *         if tries is not None:
 *             self.options.tries = tries
 *             optmask |= ARES_OPT_TRIES             # <<<<<<<<<<<<<<
 * 
 *         if ndots is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_TRIES);

    /* "cyares/channel.pyx":191
 *             optmask |= ARES_OPT_TIMEOUTMS
 * 
 *         if tries is not None:             # <<<<<<<<<<<<<<
 *             self.options.tries = tries
 *             optmask |= ARES_OPT_TRIES
*/
  }

  /* "cyares/channel.pyx":195
 *             optmask |= ARES_OPT_TRIES
 * 
 *         if ndots is not None:             # <<<<<<<<<<<<<<
 *             self.options.ndots = ndots
 *             optmask |= ARES_OPT_NDOTS
*/
  __pyx_t_3 = (__pyx_v_ndots != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":196
 * 
 *         if ndots is not None:
 *             self.options.ndots = ndots             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_NDOTS
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_ndots); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L1_error)
    __pyx_v_self->options.ndots = __pyx_t_4;

    /* "cyares/channel.pyx":197
 *         if ndots is not None:
 *             self.options.ndots = ndots
 *             optmask |= ARES_OPT_NDOTS             # <<<<<<<<<<<<<<
 * 
 *         if tcp_port is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_NDOTS);

    /* "cyares/channel.pyx":195
 *             optmask |= ARES_OPT_TRIES
 * 
 *         if ndots is not None:             # <<<<<<<<<<<<<<
 *             self.options.ndots = ndots
 *             optmask |= ARES_OPT_NDOTS
*/
  }

  /* "cyares/channel.pyx":199
 *             optmask |= ARES_OPT_NDOTS
 * 
 *         if tcp_port is not None:             # <<<<<<<<<<<<<<
 *             self.options.tcp_port = tcp_port
 *             optmask |= ARES_OPT_TCP_PORT
*/
  __pyx_t_3 = (__pyx_v_tcp_port != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":200
 * 
 *         if tcp_port is not None:
 *             self.options.tcp_port = tcp_port             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_TCP_PORT
 * 
*/
    __pyx_t_5 = __Pyx_PyLong_As_unsigned_short(__pyx_v_tcp_port); if (unlikely((__pyx_t_5 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 200, __pyx_L1_error)
    __pyx_v_self->options.tcp_port = __pyx_t_5;

    /* "cyares/channel.pyx":201
 *         if tcp_port is not None:
 *             self.options.tcp_port = tcp_port
 *             optmask |= ARES_OPT_TCP_PORT             # <<<<<<<<<<<<<<
 * 
 *         if udp_port is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_TCP_PORT);

    /* "cyares/channel.pyx":199
 *             optmask |= ARES_OPT_NDOTS
 * 
 *         if tcp_port is not None:             # <<<<<<<<<<<<<<
 *             self.options.tcp_port = tcp_port
 *             optmask |= ARES_OPT_TCP_PORT
*/
  }

  /* "cyares/channel.pyx":203
 *             optmask |= ARES_OPT_TCP_PORT
 * 
 *         if udp_port is not None:             # <<<<<<<<<<<<<<
 *             self.options.udp_port = udp_port
 *             optmask |= ARES_OPT_UDP_PORT
*/
  __pyx_t_3 = (__pyx_v_udp_port != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":204
 * 
 *         if udp_port is not None:
 *             self.options.udp_port = udp_port             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_UDP_PORT
 * 
*/
    __pyx_t_5 = __Pyx_PyLong_As_unsigned_short(__pyx_v_udp_port); if (unlikely((__pyx_t_5 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
    __pyx_v_self->options.udp_port = __pyx_t_5;

    /* "cyares/channel.pyx":205
 *         if udp_port is not None:
 *             self.options.udp_port = udp_port
 *             optmask |= ARES_OPT_UDP_PORT             # <<<<<<<<<<<<<<
 * 
 *         if socket_send_buffer_size is not None:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_UDP_PORT);

    /* "cyares/channel.pyx":203
 *             optmask |= ARES_OPT_TCP_PORT
 * 
 *         if udp_port is not None:             # <<<<<<<<<<<<<<
 *             self.options.udp_port = udp_port
 *             optmask |= ARES_OPT_UDP_PORT
*/
  }

  /* "cyares/channel.pyx":207
 *             optmask |= ARES_OPT_UDP_PORT
 * 
 *         if socket_send_buffer_size is not None:             # <<<<<<<<<<<<<<
 *             self.options.socket_send_buffer_size = socket_send_buffer_size
 *             optmask |= ARES_OPT_SOCK_SNDBUF
*/
  __pyx_t_3 = (__pyx_v_socket_send_buffer_size != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":208
 * 
 *         if socket_send_buffer_size is not None:
 *             self.options.socket_send_buffer_size = socket_send_buffer_size             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_SOCK_SNDBUF
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_socket_send_buffer_size); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 208, __pyx_L1_error)
    __pyx_v_self->options.socket_send_buffer_size = __pyx_t_4;

    /* "cyares/channel.pyx":209
 *         if socket_send_buffer_size is not None:
 *             self.options.socket_send_buffer_size = socket_send_buffer_size
 *             optmask |= ARES_OPT_SOCK_SNDBUF             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_SOCK_SNDBUF);

    /* "cyares/channel.pyx":207
 *             optmask |= ARES_OPT_UDP_PORT
 * 
 *         if socket_send_buffer_size is not None:             # <<<<<<<<<<<<<<
 *             self.options.socket_send_buffer_size = socket_send_buffer_size
 *             optmask |= ARES_OPT_SOCK_SNDBUF
*/
  }

  /* "cyares/channel.pyx":212
 * 
 * 
 *         if socket_receive_buffer_size is not None:             # <<<<<<<<<<<<<<
 *             self.options.socket_receive_buffer_size = socket_receive_buffer_size
 *             optmask |= ARES_OPT_SOCK_RCVBUF
*/
  __pyx_t_3 = (__pyx_v_socket_receive_buffer_size != Py_None);
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":213
 * 
 *         if socket_receive_buffer_size is not None:
 *             self.options.socket_receive_buffer_size = socket_receive_buffer_size             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_SOCK_RCVBUF
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_socket_receive_buffer_size); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 213, __pyx_L1_error)
    __pyx_v_self->options.socket_receive_buffer_size = __pyx_t_4;

    /* "cyares/channel.pyx":214
 *         if socket_receive_buffer_size is not None:
 *             self.options.socket_receive_buffer_size = socket_receive_buffer_size
 *             optmask |= ARES_OPT_SOCK_RCVBUF             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_SOCK_RCVBUF);

    /* "cyares/channel.pyx":212
 * 
 * 
 *         if socket_receive_buffer_size is not None:             # <<<<<<<<<<<<<<
 *             self.options.socket_receive_buffer_size = socket_receive_buffer_size
 *             optmask |= ARES_OPT_SOCK_RCVBUF
*/
  }

  /* "cyares/channel.pyx":217
 * 
 * 
 *         if sock_state_cb:             # <<<<<<<<<<<<<<
 *             if not callable(sock_state_cb):
 *                 raise TypeError("sock_state_cb must be callable")
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_sock_state_cb); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 217, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "cyares/channel.pyx":218
 * 
 *         if sock_state_cb:
 *             if not callable(sock_state_cb):             # <<<<<<<<<<<<<<
 *                 raise TypeError("sock_state_cb must be callable")
 * 
*/
    __pyx_t_3 = __Pyx_PyCallable_Check(__pyx_v_sock_state_cb); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 218, __pyx_L1_error)
    __pyx_t_6 = (!__pyx_t_3);
    if (unlikely(__pyx_t_6)) {

      /* "cyares/channel.pyx":219
 *         if sock_state_cb:
 *             if not callable(sock_state_cb):
 *                 raise TypeError("sock_state_cb must be callable")             # <<<<<<<<<<<<<<
 * 
 *             # This must be kept alive while the channel is alive.
*/
      __pyx_t_1 = NULL;
      __Pyx_INCREF(__pyx_builtin_TypeError);
      __pyx_t_7 = __pyx_builtin_TypeError; 
      __pyx_t_8 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_sock_state_cb_must_be_callable};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 219, __pyx_L1_error)

      /* "cyares/channel.pyx":218
 * 
 *         if sock_state_cb:
 *             if not callable(sock_state_cb):             # <<<<<<<<<<<<<<
 *                 raise TypeError("sock_state_cb must be callable")
 * 
*/
    }

    /* "cyares/channel.pyx":222
 * 
 *             # This must be kept alive while the channel is alive.
 *             self.socket_handle = SocketHandle.new(sock_state_cb)             # <<<<<<<<<<<<<<
 * 
 *             self.options.sock_state_cb = __socket_state_callback
*/
    __pyx_t_2 = ((PyObject *)__pyx_vtabptr_6cyares_13socket_handle_SocketHandle->__pyx_new(__pyx_v_sock_state_cb)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->socket_handle);
    __Pyx_DECREF((PyObject *)__pyx_v_self->socket_handle);
    __pyx_v_self->socket_handle = ((struct __pyx_obj_6cyares_13socket_handle_SocketHandle *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyares/channel.pyx":224
 *             self.socket_handle = SocketHandle.new(sock_state_cb)
 * 
 *             self.options.sock_state_cb = __socket_state_callback             # <<<<<<<<<<<<<<
 *             self.options.sock_state_cb_data = <void*>self.socket_handle
 *             optmask |= ARES_OPT_SOCK_STATE_CB
*/
    __pyx_v_self->options.sock_state_cb = __pyx_f_6cyares_13socket_handle___socket_state_callback;

    /* "cyares/channel.pyx":225
 * 
 *             self.options.sock_state_cb = __socket_state_callback
 *             self.options.sock_state_cb_data = <void*>self.socket_handle             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_SOCK_STATE_CB
 *         else:
*/
    __pyx_v_self->options.sock_state_cb_data = ((void *)__pyx_v_self->socket_handle);

    /* "cyares/channel.pyx":226
 *             self.options.sock_state_cb = __socket_state_callback
 *             self.options.sock_state_cb_data = <void*>self.socket_handle
 *             optmask |= ARES_OPT_SOCK_STATE_CB             # <<<<<<<<<<<<<<
 *         else:
 *             self.socket_handle = None
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_SOCK_STATE_CB);

    /* "cyares/channel.pyx":217
 * 
 * 
 *         if sock_state_cb:             # <<<<<<<<<<<<<<
 *             if not callable(sock_state_cb):
 *                 raise TypeError("sock_state_cb must be callable")
*/
    goto __pyx_L11;
  }

  /* "cyares/channel.pyx":228
 *             optmask |= ARES_OPT_SOCK_STATE_CB
 *         else:
 *             self.socket_handle = None             # <<<<<<<<<<<<<<
 * 
 *         if event_thread:
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->socket_handle);
    __Pyx_DECREF((PyObject *)__pyx_v_self->socket_handle);
    __pyx_v_self->socket_handle = ((struct __pyx_obj_6cyares_13socket_handle_SocketHandle *)Py_None);
  }
  __pyx_L11:;

  /* "cyares/channel.pyx":230
 *             self.socket_handle = None
 * 
 *         if event_thread:             # <<<<<<<<<<<<<<
 *             if not ares_threadsafety():
 *                 raise RuntimeError("c-ares is not built with thread safety")
*/
  if (__pyx_v_event_thread) {

    /* "cyares/channel.pyx":231
 * 
 *         if event_thread:
 *             if not ares_threadsafety():             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("c-ares is not built with thread safety")
 *             if sock_state_cb:
*/
    __pyx_t_6 = (!ares_threadsafety());
    if (unlikely(__pyx_t_6)) {

      /* "cyares/channel.pyx":232
 *         if event_thread:
 *             if not ares_threadsafety():
 *                 raise RuntimeError("c-ares is not built with thread safety")             # <<<<<<<<<<<<<<
 *             if sock_state_cb:
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")
*/
      __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_builtin_RuntimeError);
      __pyx_t_1 = __pyx_builtin_RuntimeError; 
      __pyx_t_8 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_c_ares_is_not_built_with_thread};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 232, __pyx_L1_error)

      /* "cyares/channel.pyx":231
 * 
 *         if event_thread:
 *             if not ares_threadsafety():             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("c-ares is not built with thread safety")
 *             if sock_state_cb:
*/
    }

    /* "cyares/channel.pyx":233
 *             if not ares_threadsafety():
 *                 raise RuntimeError("c-ares is not built with thread safety")
 *             if sock_state_cb:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")
 *             optmask |= ARES_OPT_EVENT_THREAD
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_sock_state_cb); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 233, __pyx_L1_error)
    if (unlikely(__pyx_t_6)) {

      /* "cyares/channel.pyx":234
 *                 raise RuntimeError("c-ares is not built with thread safety")
 *             if sock_state_cb:
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_EVENT_THREAD
 *             self.options.evsys = ARES_EVSYS_DEFAULT
*/
      __pyx_t_1 = NULL;
      __Pyx_INCREF(__pyx_builtin_RuntimeError);
      __pyx_t_7 = __pyx_builtin_RuntimeError; 
      __pyx_t_8 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_sock_state_cb_and_event_thread_c};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 234, __pyx_L1_error)

      /* "cyares/channel.pyx":233
 *             if not ares_threadsafety():
 *                 raise RuntimeError("c-ares is not built with thread safety")
 *             if sock_state_cb:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")
 *             optmask |= ARES_OPT_EVENT_THREAD
*/
    }

    /* "cyares/channel.pyx":235
 *             if sock_state_cb:
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")
 *             optmask |= ARES_OPT_EVENT_THREAD             # <<<<<<<<<<<<<<
 *             self.options.evsys = ARES_EVSYS_DEFAULT
 * 
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_EVENT_THREAD);

    /* "cyares/channel.pyx":236
 *                 raise RuntimeError("sock_state_cb and event_thread cannot be used together")
 *             optmask |= ARES_OPT_EVENT_THREAD
 *             self.options.evsys = ARES_EVSYS_DEFAULT             # <<<<<<<<<<<<<<
 * 
 *         if lookups:
*/
    __pyx_v_self->options.evsys = ARES_EVSYS_DEFAULT;

    /* "cyares/channel.pyx":230
 *             self.socket_handle = None
 * 
 *         if event_thread:             # <<<<<<<<<<<<<<
 *             if not ares_threadsafety():
 *                 raise RuntimeError("c-ares is not built with thread safety")
*/
  }

  /* "cyares/channel.pyx":238
 *             self.options.evsys = ARES_EVSYS_DEFAULT
 * 
 *         if lookups:             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(lookups, &view)
 *             self.options.lookups = <char*>view.buf
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_lookups); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 238, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":239
 * 
 *         if lookups:
 *             cyares_get_buffer(lookups, &view)             # <<<<<<<<<<<<<<
 *             self.options.lookups = <char*>view.buf
 *             optmask |= ARES_OPT_LOOKUPS
*/
    __pyx_t_4 = cyares_get_buffer(__pyx_v_lookups, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 239, __pyx_L1_error)

    /* "cyares/channel.pyx":240
 *         if lookups:
 *             cyares_get_buffer(lookups, &view)
 *             self.options.lookups = <char*>view.buf             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_LOOKUPS
 *             cyares_release_buffer(&view)
*/
    __pyx_v_self->options.lookups = ((char *)__pyx_v_view.buf);

    /* "cyares/channel.pyx":241
 *             cyares_get_buffer(lookups, &view)
 *             self.options.lookups = <char*>view.buf
 *             optmask |= ARES_OPT_LOOKUPS             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&view)
 * 
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_LOOKUPS);

    /* "cyares/channel.pyx":242
 *             self.options.lookups = <char*>view.buf
 *             optmask |= ARES_OPT_LOOKUPS
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *         if domains:
*/
    cyares_release_buffer((&__pyx_v_view));

    /* "cyares/channel.pyx":238
 *             self.options.evsys = ARES_EVSYS_DEFAULT
 * 
 *         if lookups:             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(lookups, &view)
 *             self.options.lookups = <char*>view.buf
*/
  }

  /* "cyares/channel.pyx":244
 *             cyares_release_buffer(&view)
 * 
 *         if domains:             # <<<<<<<<<<<<<<
 *             strs = <char**>PyMem_Malloc(sizeof(char*) *  len(domains))
 * 
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_domains); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":245
 * 
 *         if domains:
 *             strs = <char**>PyMem_Malloc(sizeof(char*) *  len(domains))             # <<<<<<<<<<<<<<
 * 
 *             for i in domains:
*/
    __pyx_t_9 = PyObject_Length(__pyx_v_domains); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
    __pyx_v_strs = ((char **)PyMem_Malloc(((sizeof(char *)) * __pyx_t_9)));

    /* "cyares/channel.pyx":247
 *             strs = <char**>PyMem_Malloc(sizeof(char*) *  len(domains))
 * 
 *             for i in domains:             # <<<<<<<<<<<<<<
 *                 cyares_get_buffer(i, &view)
 *                 strs[i] = <char*>view.buf
*/
    if (likely(PyList_CheckExact(__pyx_v_domains)) || PyTuple_CheckExact(__pyx_v_domains)) {
      __pyx_t_2 = __pyx_v_domains; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_domains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 247, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 247, __pyx_L1_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          __pyx_t_7 = __Pyx_PyList_GetItemRef(__pyx_t_2, __pyx_t_9);
          ++__pyx_t_9;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 247, __pyx_L1_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9));
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_9);
          #endif
          ++__pyx_t_9;
        }
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 247, __pyx_L1_error)
      } else {
        __pyx_t_7 = __pyx_t_10(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 247, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "cyares/channel.pyx":248
 * 
 *             for i in domains:
 *                 cyares_get_buffer(i, &view)             # <<<<<<<<<<<<<<
 *                 strs[i] = <char*>view.buf
 *                 cyares_release_buffer(&view)
*/
      __pyx_t_4 = cyares_get_buffer(__pyx_v_i, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 248, __pyx_L1_error)

      /* "cyares/channel.pyx":249
 *             for i in domains:
 *                 cyares_get_buffer(i, &view)
 *                 strs[i] = <char*>view.buf             # <<<<<<<<<<<<<<
 *                 cyares_release_buffer(&view)
 * 
*/
      __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
      (__pyx_v_strs[__pyx_t_11]) = ((char *)__pyx_v_view.buf);

      /* "cyares/channel.pyx":250
 *                 cyares_get_buffer(i, &view)
 *                 strs[i] = <char*>view.buf
 *                 cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *             self.options.domains = strs
*/
      cyares_release_buffer((&__pyx_v_view));

      /* "cyares/channel.pyx":247
 *             strs = <char**>PyMem_Malloc(sizeof(char*) *  len(domains))
 * 
 *             for i in domains:             # <<<<<<<<<<<<<<
 *                 cyares_get_buffer(i, &view)
 *                 strs[i] = <char*>view.buf
*/
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyares/channel.pyx":252
 *                 cyares_release_buffer(&view)
 * 
 *             self.options.domains = strs             # <<<<<<<<<<<<<<
 *             self.options.ndomains = len(domains)
 *             optmask |= ARES_OPT_DOMAINS
*/
    __pyx_v_self->options.domains = __pyx_v_strs;

    /* "cyares/channel.pyx":253
 * 
 *             self.options.domains = strs
 *             self.options.ndomains = len(domains)             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_DOMAINS
 * 
*/
    __pyx_t_9 = PyObject_Length(__pyx_v_domains); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 253, __pyx_L1_error)
    __pyx_v_self->options.ndomains = __pyx_t_9;

    /* "cyares/channel.pyx":254
 *             self.options.domains = strs
 *             self.options.ndomains = len(domains)
 *             optmask |= ARES_OPT_DOMAINS             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_DOMAINS);

    /* "cyares/channel.pyx":244
 *             cyares_release_buffer(&view)
 * 
 *         if domains:             # <<<<<<<<<<<<<<
 *             strs = <char**>PyMem_Malloc(sizeof(char*) *  len(domains))
 * 
*/
  }

  /* "cyares/channel.pyx":257
 * 
 * 
 *         if rotate:             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_ROTATE
 * 
*/
  if (__pyx_v_rotate) {

    /* "cyares/channel.pyx":258
 * 
 *         if rotate:
 *             optmask |= ARES_OPT_ROTATE             # <<<<<<<<<<<<<<
 * 
 *         if resolvconf_path:
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_ROTATE);

    /* "cyares/channel.pyx":257
 * 
 * 
 *         if rotate:             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_ROTATE
 * 
*/
  }

  /* "cyares/channel.pyx":260
 *             optmask |= ARES_OPT_ROTATE
 * 
 *         if resolvconf_path:             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_RESOLVCONF
 *             cyares_get_buffer(resolvconf_path, &view)
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_resolvconf_path); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 260, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":261
 * 
 *         if resolvconf_path:
 *             optmask |= ARES_OPT_RESOLVCONF             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(resolvconf_path, &view)
 *             self.options.resolvconf_path = <char*>view.buf
*/
    __pyx_v_optmask = (__pyx_v_optmask | ARES_OPT_RESOLVCONF);

    /* "cyares/channel.pyx":262
 *         if resolvconf_path:
 *             optmask |= ARES_OPT_RESOLVCONF
 *             cyares_get_buffer(resolvconf_path, &view)             # <<<<<<<<<<<<<<
 *             self.options.resolvconf_path = <char*>view.buf
 *             cyares_release_buffer(&view)
*/
    __pyx_t_4 = cyares_get_buffer(__pyx_v_resolvconf_path, (&__pyx_v_view)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 262, __pyx_L1_error)

    /* "cyares/channel.pyx":263
 *             optmask |= ARES_OPT_RESOLVCONF
 *             cyares_get_buffer(resolvconf_path, &view)
 *             self.options.resolvconf_path = <char*>view.buf             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&view)
 * 
*/
    __pyx_v_self->options.resolvconf_path = ((char *)__pyx_v_view.buf);

    /* "cyares/channel.pyx":264
 *             cyares_get_buffer(resolvconf_path, &view)
 *             self.options.resolvconf_path = <char*>view.buf
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *         if local_ip:
*/
    cyares_release_buffer((&__pyx_v_view));

    /* "cyares/channel.pyx":260
 *             optmask |= ARES_OPT_ROTATE
 * 
 *         if resolvconf_path:             # <<<<<<<<<<<<<<
 *             optmask |= ARES_OPT_RESOLVCONF
 *             cyares_get_buffer(resolvconf_path, &view)
*/
  }

  /* "cyares/channel.pyx":266
 *             cyares_release_buffer(&view)
 * 
 *         if local_ip:             # <<<<<<<<<<<<<<
 *             self.set_local_ip(local_ip)
 * 
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_local_ip); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 266, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":267
 * 
 *         if local_ip:
 *             self.set_local_ip(local_ip)             # <<<<<<<<<<<<<<
 * 
 *         if local_dev:
*/
    __pyx_t_7 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_local_ip};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_local_ip, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyares/channel.pyx":266
 *             cyares_release_buffer(&view)
 * 
 *         if local_ip:             # <<<<<<<<<<<<<<
 *             self.set_local_ip(local_ip)
 * 
*/
  }

  /* "cyares/channel.pyx":269
 *             self.set_local_ip(local_ip)
 * 
 *         if local_dev:             # <<<<<<<<<<<<<<
 *             self.set_local_dev(local_dev)
 * 
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_local_dev); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 269, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":270
 * 
 *         if local_dev:
 *             self.set_local_dev(local_dev)             # <<<<<<<<<<<<<<
 * 
 *         r = ares_init_options(&self.channel, &self.options, optmask)
*/
    __pyx_t_7 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_local_dev};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_local_dev, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyares/channel.pyx":269
 *             self.set_local_ip(local_ip)
 * 
 *         if local_dev:             # <<<<<<<<<<<<<<
 *             self.set_local_dev(local_dev)
 * 
*/
  }

  /* "cyares/channel.pyx":272
 *             self.set_local_dev(local_dev)
 * 
 *         r = ares_init_options(&self.channel, &self.options, optmask)             # <<<<<<<<<<<<<<
 *         if r != ARES_SUCCESS:
 *             raise AresError(r)
*/
  __pyx_v_r = ares_init_options((&__pyx_v_self->channel), (&__pyx_v_self->options), __pyx_v_optmask);

  /* "cyares/channel.pyx":273
 * 
 *         r = ares_init_options(&self.channel, &self.options, optmask)
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  __pyx_t_6 = (__pyx_v_r != ARES_SUCCESS);
  if (unlikely(__pyx_t_6)) {

    /* "cyares/channel.pyx":274
 *         r = ares_init_options(&self.channel, &self.options, optmask)
 *         if r != ARES_SUCCESS:
 *             raise AresError(r)             # <<<<<<<<<<<<<<
 * 
 *         if servers:
*/
    __pyx_t_7 = NULL;
    __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError);
    __pyx_t_1 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError); 
    __pyx_t_12 = __Pyx_PyLong_From_int(__pyx_v_r); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_12};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_2), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 274, __pyx_L1_error)

    /* "cyares/channel.pyx":273
 * 
 *         r = ares_init_options(&self.channel, &self.options, optmask)
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  }

  /* "cyares/channel.pyx":276
 *             raise AresError(r)
 * 
 *         if servers:             # <<<<<<<<<<<<<<
 *             self.servers = servers
 * 
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_servers); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 276, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":277
 * 
 *         if servers:
 *             self.servers = servers             # <<<<<<<<<<<<<<
 * 
 *         if strs != NULL:
*/
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_servers, __pyx_v_servers) < 0) __PYX_ERR(0, 277, __pyx_L1_error)

    /* "cyares/channel.pyx":276
 *             raise AresError(r)
 * 
 *         if servers:             # <<<<<<<<<<<<<<
 *             self.servers = servers
 * 
*/
  }

  /* "cyares/channel.pyx":279
 *             self.servers = servers
 * 
 *         if strs != NULL:             # <<<<<<<<<<<<<<
 *             # be sure to throw an issue on github if this becomes a future problem
 *             PyMem_Free(strs)
*/
  __pyx_t_6 = (__pyx_v_strs != NULL);
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":281
 *         if strs != NULL:
 *             # be sure to throw an issue on github if this becomes a future problem
 *             PyMem_Free(strs)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    PyMem_Free(__pyx_v_strs);

    /* "cyares/channel.pyx":279
 *             self.servers = servers
 * 
 *         if strs != NULL:             # <<<<<<<<<<<<<<
 *             # be sure to throw an issue on github if this becomes a future problem
 *             PyMem_Free(strs)
*/
  }

  /* "cyares/channel.pyx":138
 * 
 * cdef class Channel:
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         flags = None,
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cyares.channel.Channel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":300
 *     # dns+tls://one.one.one.one?ipaddr=1.1.1.1
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def servers(self):
 *         cdef char* data = ares_get_servers_csv(self.channel)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_7servers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6cyares_7channel_7Channel_7servers_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_7servers___get__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_7servers___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  char *__pyx_v_data;
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_servers = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "cyares/channel.pyx":302
 *     @property
 *     def servers(self):
 *         cdef char* data = ares_get_servers_csv(self.channel)             # <<<<<<<<<<<<<<
 *         cdef bytes s = PyBytes_FromString(data)
 *         ares_free_string(data)
*/
  __pyx_v_data = ares_get_servers_csv(__pyx_v_self->channel);

  /* "cyares/channel.pyx":303
 *     def servers(self):
 *         cdef char* data = ares_get_servers_csv(self.channel)
 *         cdef bytes s = PyBytes_FromString(data)             # <<<<<<<<<<<<<<
 *         ares_free_string(data)
 *         cdef str servers = s.decode('utf-8')
*/
  __pyx_t_1 = PyBytes_FromString(__pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":304
 *         cdef char* data = ares_get_servers_csv(self.channel)
 *         cdef bytes s = PyBytes_FromString(data)
 *         ares_free_string(data)             # <<<<<<<<<<<<<<
 *         cdef str servers = s.decode('utf-8')
 *         return servers.split(",")
*/
  ares_free_string(__pyx_v_data);

  /* "cyares/channel.pyx":305
 *         cdef bytes s = PyBytes_FromString(data)
 *         ares_free_string(data)
 *         cdef str servers = s.decode('utf-8')             # <<<<<<<<<<<<<<
 *         return servers.split(",")
 * 
*/
  if (unlikely(__pyx_v_s == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 305, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_s, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_servers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":306
 *         ares_free_string(data)
 *         cdef str servers = s.decode('utf-8')
 *         return servers.split(",")             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Split(__pyx_v_servers, __Pyx_NoneAsNull(__pyx_mstate_global->__pyx_kp_u_), -1L); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":300
 *     # dns+tls://one.one.one.one?ipaddr=1.1.1.1
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def servers(self):
 *         cdef char* data = ares_get_servers_csv(self.channel)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.servers.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_servers);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":309
 * 
 * 
 *     @servers.setter             # <<<<<<<<<<<<<<
 *     def servers(self, list servers):
 *         cdef int r
*/

/* Python wrapper */
static int __pyx_pw_6cyares_7channel_7Channel_7servers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_servers); /*proto*/
static int __pyx_pw_6cyares_7channel_7Channel_7servers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_servers) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_servers), (&PyList_Type), 1, "servers", 1))) __PYX_ERR(0, 310, __pyx_L1_error)
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_7servers_2__set__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), ((PyObject*)__pyx_v_servers));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6cyares_7channel_7Channel_7servers_2__set__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_servers) {
  int __pyx_v_r;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_v_csv_list = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "cyares/channel.pyx":313
 *         cdef int r
 *         cdef Py_buffer view
 *         cdef str csv_list  = ",".join(servers)             # <<<<<<<<<<<<<<
 * 
 *         cyares_get_buffer(csv_list, &view)
*/
  __pyx_t_1 = PyUnicode_Join(__pyx_mstate_global->__pyx_kp_u_, __pyx_v_servers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_csv_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":315
 *         cdef str csv_list  = ",".join(servers)
 * 
 *         cyares_get_buffer(csv_list, &view)             # <<<<<<<<<<<<<<
 * 
 *         r = ares_set_servers_csv(self.channel, <const char*>view.buf)
*/
  __pyx_t_2 = cyares_get_buffer(__pyx_v_csv_list, (&__pyx_v_view)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 315, __pyx_L1_error)

  /* "cyares/channel.pyx":317
 *         cyares_get_buffer(csv_list, &view)
 * 
 *         r = ares_set_servers_csv(self.channel, <const char*>view.buf)             # <<<<<<<<<<<<<<
 * 
 *         cyares_release_buffer(&view)
*/
  __pyx_v_r = ares_set_servers_csv(__pyx_v_self->channel, ((char const *)__pyx_v_view.buf));

  /* "cyares/channel.pyx":319
 *         r = ares_set_servers_csv(self.channel, <const char*>view.buf)
 * 
 *         cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *         if r != ARES_SUCCESS:
*/
  cyares_release_buffer((&__pyx_v_view));

  /* "cyares/channel.pyx":321
 *         cyares_release_buffer(&view)
 * 
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  __pyx_t_3 = (__pyx_v_r != ARES_SUCCESS);
  if (unlikely(__pyx_t_3)) {

    /* "cyares/channel.pyx":322
 * 
 *         if r != ARES_SUCCESS:
 *             raise AresError(r)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_4 = NULL;
    __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError);
    __pyx_t_5 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError); 
    __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_r); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_1), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 322, __pyx_L1_error)

    /* "cyares/channel.pyx":321
 *         cyares_release_buffer(&view)
 * 
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  }

  /* "cyares/channel.pyx":309
 * 
 * 
 *     @servers.setter             # <<<<<<<<<<<<<<
 *     def servers(self, list servers):
 *         cdef int r
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cyares.channel.Channel.servers.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_csv_list);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":325
 * 
 * 
 *     cpdef void cancel(self) noexcept:             # <<<<<<<<<<<<<<
 *         ares_cancel(self.channel)
 *         self._cancelled = True
*/

static PyObject *__pyx_pw_6cyares_7channel_7Channel_3cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_6cyares_7channel_7Channel_cancel(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_cancel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_6cyares_7channel_7Channel_3cancel)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cyares/channel.pyx":326
 * 
 *     cpdef void cancel(self) noexcept:
 *         ares_cancel(self.channel)             # <<<<<<<<<<<<<<
 *         self._cancelled = True
 * 
*/
  ares_cancel(__pyx_v_self->channel);

  /* "cyares/channel.pyx":327
 *     cpdef void cancel(self) noexcept:
 *         ares_cancel(self.channel)
 *         self._cancelled = True             # <<<<<<<<<<<<<<
 * 
 *     def reinit(self):
*/
  __pyx_v_self->_cancelled = 1;

  /* "cyares/channel.pyx":325
 * 
 * 
 *     cpdef void cancel(self) noexcept:             # <<<<<<<<<<<<<<
 *         ares_cancel(self.channel)
 *         self._cancelled = True
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cyares.channel.Channel.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_3cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_3cancel = {"cancel", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_3cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_3cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("cancel", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_2cancel(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_2cancel(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6cyares_7channel_7Channel_cancel(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":329
 *         self._cancelled = True
 * 
 *     def reinit(self):             # <<<<<<<<<<<<<<
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_5reinit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_5reinit = {"reinit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_5reinit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_5reinit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reinit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("reinit", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("reinit", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_4reinit(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_4reinit(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  int __pyx_v_r;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reinit", 0);

  /* "cyares/channel.pyx":330
 * 
 *     def reinit(self):
 *         cdef int r = ares_reinit(self.channel)             # <<<<<<<<<<<<<<
 *         if r != ARES_SUCCESS:
 *             raise AresError(r)
*/
  __pyx_v_r = ares_reinit(__pyx_v_self->channel);

  /* "cyares/channel.pyx":331
 *     def reinit(self):
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  __pyx_t_1 = (__pyx_v_r != ARES_SUCCESS);
  if (unlikely(__pyx_t_1)) {

    /* "cyares/channel.pyx":332
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:
 *             raise AresError(r)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError);
    __pyx_t_4 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError); 
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_r); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_2), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 332, __pyx_L1_error)

    /* "cyares/channel.pyx":331
 *     def reinit(self):
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             raise AresError(r)
 * 
*/
  }

  /* "cyares/channel.pyx":329
 *         self._cancelled = True
 * 
 *     def reinit(self):             # <<<<<<<<<<<<<<
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyares.channel.Channel.reinit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":335
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if (not self._cancelled) and self.handles:
 *             self.cancel()
*/

/* Python wrapper */
static void __pyx_pw_6cyares_7channel_7Channel_7__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6cyares_7channel_7Channel_7__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6cyares_7channel_7Channel_6__dealloc__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6cyares_7channel_7Channel_6__dealloc__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyares/channel.pyx":336
 * 
 *     def __dealloc__(self):
 *         if (not self._cancelled) and self.handles:             # <<<<<<<<<<<<<<
 *             self.cancel()
 * 
*/
  __pyx_t_2 = (!__pyx_v_self->_cancelled);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->handles != Py_None)&&(__Pyx_PySet_GET_SIZE(__pyx_v_self->handles) != 0);
  if (unlikely(((!CYTHON_ASSUME_SAFE_MACROS) && __pyx_t_2 < 0))) __PYX_ERR(0, 336, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":337
 *     def __dealloc__(self):
 *         if (not self._cancelled) and self.handles:
 *             self.cancel()             # <<<<<<<<<<<<<<
 * 
 *         # If your not using an event_thread
*/
    ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->cancel(__pyx_v_self, 0);

    /* "cyares/channel.pyx":336
 * 
 *     def __dealloc__(self):
 *         if (not self._cancelled) and self.handles:             # <<<<<<<<<<<<<<
 *             self.cancel()
 * 
*/
  }

  /* "cyares/channel.pyx":347
 *         # so that use-after-free never sees the
 *         # light of day.
 *         if self.event_thread:             # <<<<<<<<<<<<<<
 *             self.__wait(-1)
 * 
*/
  if (__pyx_v_self->event_thread) {

    /* "cyares/channel.pyx":348
 *         # light of day.
 *         if self.event_thread:
 *             self.__wait(-1)             # <<<<<<<<<<<<<<
 * 
 *         ares_destroy(self.channel)
*/
    ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__wait(__pyx_v_self, -1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 348, __pyx_L1_error)

    /* "cyares/channel.pyx":347
 *         # so that use-after-free never sees the
 *         # light of day.
 *         if self.event_thread:             # <<<<<<<<<<<<<<
 *             self.__wait(-1)
 * 
*/
  }

  /* "cyares/channel.pyx":350
 *             self.__wait(-1)
 * 
 *         ares_destroy(self.channel)             # <<<<<<<<<<<<<<
 * 
 *     def __enter__(self):
*/
  ares_destroy(__pyx_v_self->channel);

  /* "cyares/channel.pyx":335
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if (not self._cancelled) and self.handles:
 *             self.cancel()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyares.channel.Channel.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
}

/* "cyares/channel.pyx":352
 *         ares_destroy(self.channel)
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_9__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_9__enter__ = {"__enter__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_9__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_9__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__enter__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__enter__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_8__enter__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_8__enter__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__", 0);

  /* "cyares/channel.pyx":353
 * 
 *     def __enter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, *args):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "cyares/channel.pyx":352
 *         ares_destroy(self.channel)
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":355
 *         return self
 * 
 *     def __exit__(self, *args):             # <<<<<<<<<<<<<<
 *         # TODO: Handle collection & Closure
 *         self.cancel()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_11__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_11__exit__ = {"__exit__", (PyCFunction)(void(*)(void))(PyCFunctionWithKeywords)__pyx_pw_6cyares_7channel_7Channel_11__exit__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_11__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__exit__", __pyx_kwds); return NULL;}
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_10__exit__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_10__exit__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__", 0);

  /* "cyares/channel.pyx":357
 *     def __exit__(self, *args):
 *         # TODO: Handle collection & Closure
 *         self.cancel()             # <<<<<<<<<<<<<<
 * 
 *     # wrote my own malloc function wrapper because
*/
  ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->cancel(__pyx_v_self, 0);

  /* "cyares/channel.pyx":355
 *         return self
 * 
 *     def __exit__(self, *args):             # <<<<<<<<<<<<<<
 *         # TODO: Handle collection & Closure
 *         self.cancel()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":363
 *     # allocating and then freeing memory...
 *     # cython will remember to throw the memory error
 *     cdef void* _malloc(self, size_t size) except NULL:             # <<<<<<<<<<<<<<
 *         cdef void* memory = <void*>PyMem_Malloc(size)
 *         if memory == NULL:
*/

static void *__pyx_f_6cyares_7channel_7Channel__malloc(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, size_t __pyx_v_size) {
  void *__pyx_v_memory;
  void *__pyx_r;
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyares/channel.pyx":364
 *     # cython will remember to throw the memory error
 *     cdef void* _malloc(self, size_t size) except NULL:
 *         cdef void* memory = <void*>PyMem_Malloc(size)             # <<<<<<<<<<<<<<
 *         if memory == NULL:
 *             PyErr_NoMemory()
*/
  __pyx_v_memory = ((void *)PyMem_Malloc(__pyx_v_size));

  /* "cyares/channel.pyx":365
 *     cdef void* _malloc(self, size_t size) except NULL:
 *         cdef void* memory = <void*>PyMem_Malloc(size)
 *         if memory == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *         return memory
*/
  __pyx_t_1 = (__pyx_v_memory == NULL);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":366
 *         cdef void* memory = <void*>PyMem_Malloc(size)
 *         if memory == NULL:
 *             PyErr_NoMemory()             # <<<<<<<<<<<<<<
 *         return memory
 * 
*/
    __pyx_t_2 = PyErr_NoMemory(); if (unlikely(__pyx_t_2 == ((PyObject *)0))) __PYX_ERR(0, 366, __pyx_L1_error)

    /* "cyares/channel.pyx":365
 *     cdef void* _malloc(self, size_t size) except NULL:
 *         cdef void* memory = <void*>PyMem_Malloc(size)
 *         if memory == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_NoMemory()
 *         return memory
*/
  }

  /* "cyares/channel.pyx":367
 *         if memory == NULL:
 *             PyErr_NoMemory()
 *         return memory             # <<<<<<<<<<<<<<
 * 
 *     # Not public to .pyi, please do not use... - Vizonex
*/
  __pyx_r = __pyx_v_memory;
  goto __pyx_L0;

  /* "cyares/channel.pyx":363
 *     # allocating and then freeing memory...
 *     # cython will remember to throw the memory error
 *     cdef void* _malloc(self, size_t size) except NULL:             # <<<<<<<<<<<<<<
 *         cdef void* memory = <void*>PyMem_Malloc(size)
 *         if memory == NULL:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyares.channel.Channel._malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyares/channel.pyx":370
 * 
 *     # Not public to .pyi, please do not use... - Vizonex
 *     def __remove_future(self, object fut):             # <<<<<<<<<<<<<<
 *         self.handles.remove(fut)
 *         self._closed += 1
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_13_Channel__remove_future(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_13_Channel__remove_future = {"_Channel__remove_future", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_13_Channel__remove_future, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_13_Channel__remove_future(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fut = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Channel__remove_future (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_fut,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 370, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 370, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_Channel__remove_future", 0) < 0) __PYX_ERR(0, 370, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_Channel__remove_future", 1, 1, 1, i); __PYX_ERR(0, 370, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 370, __pyx_L3_error)
    }
    __pyx_v_fut = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_Channel__remove_future", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 370, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel._Channel__remove_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_12__remove_future(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_fut);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_12__remove_future(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_fut) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Channel__remove_future", 0);

  /* "cyares/channel.pyx":371
 *     # Not public to .pyi, please do not use... - Vizonex
 *     def __remove_future(self, object fut):
 *         self.handles.remove(fut)             # <<<<<<<<<<<<<<
 *         self._closed += 1
 * 
*/
  if (unlikely(__pyx_v_self->handles == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "remove");
    __PYX_ERR(0, 371, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PySet_Remove(__pyx_v_self->handles, __pyx_v_fut); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 371, __pyx_L1_error)

  /* "cyares/channel.pyx":372
 *     def __remove_future(self, object fut):
 *         self.handles.remove(fut)
 *         self._closed += 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_v_self->_closed = (__pyx_v_self->_closed + 1);

  /* "cyares/channel.pyx":370
 * 
 *     # Not public to .pyi, please do not use... - Vizonex
 *     def __remove_future(self, object fut):             # <<<<<<<<<<<<<<
 *         self.handles.remove(fut)
 *         self._closed += 1
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyares.channel.Channel._Channel__remove_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":379
 *     # using such functions for non-development purposes is discouraged.
 * 
 *     def debug(self):             # <<<<<<<<<<<<<<
 *         print('Running Handles: {}'.format(self._running))
 *         print('Running Queries: {}'.format(self.running_queries))
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_15debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_15debug = {"debug", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_15debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_15debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("debug (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("debug", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("debug", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_14debug(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_14debug(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("debug", 0);

  /* "cyares/channel.pyx":380
 * 
 *     def debug(self):
 *         print('Running Handles: {}'.format(self._running))             # <<<<<<<<<<<<<<
 *         print('Running Queries: {}'.format(self.running_queries))
 *         print('Closed Handles: {}'.format(self._closed))
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__pyx_builtin_print);
  __pyx_t_3 = __pyx_builtin_print; 
  __pyx_t_5 = __pyx_mstate_global->__pyx_kp_u_Running_Handles;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_uint64_t(__pyx_v_self->_running); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyares/channel.pyx":381
 *     def debug(self):
 *         print('Running Handles: {}'.format(self._running))
 *         print('Running Queries: {}'.format(self.running_queries))             # <<<<<<<<<<<<<<
 *         print('Closed Handles: {}'.format(self._closed))
 * 
*/
  __pyx_t_3 = NULL;
  __Pyx_INCREF(__pyx_builtin_print);
  __pyx_t_4 = __pyx_builtin_print; 
  __pyx_t_6 = __pyx_mstate_global->__pyx_kp_u_Running_Queries;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_running_queries); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyares/channel.pyx":382
 *         print('Running Handles: {}'.format(self._running))
 *         print('Running Queries: {}'.format(self.running_queries))
 *         print('Closed Handles: {}'.format(self._closed))             # <<<<<<<<<<<<<<
 * 
 *     @cython.nonecheck(False)
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__pyx_builtin_print);
  __pyx_t_2 = __pyx_builtin_print; 
  __pyx_t_5 = __pyx_mstate_global->__pyx_kp_u_Closed_Handles;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_uint64_t(__pyx_v_self->_closed); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyares/channel.pyx":379
 *     # using such functions for non-development purposes is discouraged.
 * 
 *     def debug(self):             # <<<<<<<<<<<<<<
 *         print('Running Handles: {}'.format(self._running))
 *         print('Running Queries: {}'.format(self.running_queries))
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cyares.channel.Channel.debug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":384
 *         print('Closed Handles: {}'.format(self._closed))
 * 
 *     @cython.nonecheck(False)             # <<<<<<<<<<<<<<
 *     cdef object __create_future(self, object callback):
 *         cdef object fut = Future()
*/

static PyObject *__pyx_f_6cyares_7channel_7Channel__Channel__create_future(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_callback) {
  PyObject *__pyx_v_fut = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Channel__create_future", 0);

  /* "cyares/channel.pyx":386
 *     @cython.nonecheck(False)
 *     cdef object __create_future(self, object callback):
 *         cdef object fut = Future()             # <<<<<<<<<<<<<<
 *         self._running += 1
 *         self.handles.add(fut)
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__pyx_v_6cyares_7channel_Future);
  __pyx_t_3 = __pyx_v_6cyares_7channel_Future; 
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":387
 *     cdef object __create_future(self, object callback):
 *         cdef object fut = Future()
 *         self._running += 1             # <<<<<<<<<<<<<<
 *         self.handles.add(fut)
 * 
*/
  __pyx_v_self->_running = (__pyx_v_self->_running + 1);

  /* "cyares/channel.pyx":388
 *         cdef object fut = Future()
 *         self._running += 1
 *         self.handles.add(fut)             # <<<<<<<<<<<<<<
 * 
 *         # handle removal of finished futures...
*/
  if (unlikely(__pyx_v_self->handles == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(0, 388, __pyx_L1_error)
  }
  __pyx_t_5 = PySet_Add(__pyx_v_self->handles, __pyx_v_fut); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 388, __pyx_L1_error)

  /* "cyares/channel.pyx":391
 * 
 *         # handle removal of finished futures...
 *         fut.add_done_callback(self.__remove_future)             # <<<<<<<<<<<<<<
 * 
 *         if callback:
*/
  __pyx_t_3 = __pyx_v_fut;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_Channel__remove_future); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_add_done_callback, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyares/channel.pyx":393
 *         fut.add_done_callback(self.__remove_future)
 * 
 *         if callback:             # <<<<<<<<<<<<<<
 *             if not callable(callback):
 *                 raise TypeError("Provided callbacks must be callable")
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_callback); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 393, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":394
 * 
 *         if callback:
 *             if not callable(callback):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Provided callbacks must be callable")
 *             fut.add_done_callback(callback)
*/
    __pyx_t_6 = __Pyx_PyCallable_Check(__pyx_v_callback); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 394, __pyx_L1_error)
    __pyx_t_7 = (!__pyx_t_6);
    if (unlikely(__pyx_t_7)) {

      /* "cyares/channel.pyx":395
 *         if callback:
 *             if not callable(callback):
 *                 raise TypeError("Provided callbacks must be callable")             # <<<<<<<<<<<<<<
 *             fut.add_done_callback(callback)
 * 
*/
      __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_builtin_TypeError);
      __pyx_t_3 = __pyx_builtin_TypeError; 
      __pyx_t_4 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Provided_callbacks_must_be_calla};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 395, __pyx_L1_error)

      /* "cyares/channel.pyx":394
 * 
 *         if callback:
 *             if not callable(callback):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Provided callbacks must be callable")
 *             fut.add_done_callback(callback)
*/
    }

    /* "cyares/channel.pyx":396
 *             if not callable(callback):
 *                 raise TypeError("Provided callbacks must be callable")
 *             fut.add_done_callback(callback)             # <<<<<<<<<<<<<<
 * 
 *         # Up the objects refcount by 1 because were sending through a void object
*/
    __pyx_t_3 = __pyx_v_fut;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_callback};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_add_done_callback, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "cyares/channel.pyx":393
 *         fut.add_done_callback(self.__remove_future)
 * 
 *         if callback:             # <<<<<<<<<<<<<<
 *             if not callable(callback):
 *                 raise TypeError("Provided callbacks must be callable")
*/
  }

  /* "cyares/channel.pyx":400
 *         # Up the objects refcount by 1 because were sending through a void object
 *         # in a bit
 *         Py_INCREF(fut)             # <<<<<<<<<<<<<<
 *         return fut
 * 
*/
  Py_INCREF(__pyx_v_fut);

  /* "cyares/channel.pyx":401
 *         # in a bit
 *         Py_INCREF(fut)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     # _query is a lower-level C Function
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":384
 *         print('Closed Handles: {}'.format(self._closed))
 * 
 *     @cython.nonecheck(False)             # <<<<<<<<<<<<<<
 *     cdef object __create_future(self, object callback):
 *         cdef object fut = Future()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyares.channel.Channel._Channel__create_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":406
 *     # query is the upper-end and is meant to assist in
 *     # being a theoretical drop in replacement for pycares in aiodns
 *     cdef object _query(self, object qname, object qtype, int qclass, object callback):             # <<<<<<<<<<<<<<
 *         cdef int _qtype
 *         cdef object fut
*/

static PyObject *__pyx_f_6cyares_7channel_7Channel__query(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_qname, PyObject *__pyx_v_qtype, int __pyx_v_qclass, PyObject *__pyx_v_callback) {
  int __pyx_v__qtype;
  PyObject *__pyx_v_fut = 0;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_query", 0);

  /* "cyares/channel.pyx":411
 *         cdef Py_buffer view
 * 
 *         if isinstance(qtype, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
*/
  __pyx_t_1 = PyUnicode_Check(__pyx_v_qtype); 
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":412
 * 
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "cyares/channel.pyx":413
 *         if isinstance(qtype, str):
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 raise ValueError("invalid query type specified")
*/
        if (unlikely(__pyx_v_self->_query_lookups == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 413, __pyx_L4_error)
        }
        __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_self->_query_lookups, __pyx_v_qtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L4_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v__qtype = ((int)__pyx_t_6);

        /* "cyares/channel.pyx":412
 * 
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "cyares/channel.pyx":414
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 raise ValueError("invalid query type specified")
 *         else:
*/
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("cyares.channel.Channel._query", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 414, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);

        /* "cyares/channel.pyx":415
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
 *                 raise ValueError("invalid query type specified")             # <<<<<<<<<<<<<<
 *         else:
 *             _qtype = <int>qtype
*/
        __pyx_t_10 = NULL;
        __Pyx_INCREF(__pyx_builtin_ValueError);
        __pyx_t_11 = __pyx_builtin_ValueError; 
        __pyx_t_12 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_mstate_global->__pyx_kp_u_invalid_query_type_specified};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 415, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 415, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;

      /* "cyares/channel.pyx":412
 * 
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "cyares/channel.pyx":411
 *         cdef Py_buffer view
 * 
 *         if isinstance(qtype, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":417
 *                 raise ValueError("invalid query type specified")
 *         else:
 *             _qtype = <int>qtype             # <<<<<<<<<<<<<<
 * 
 *         if cyares_check_qclasses(qclass) < 0:
*/
  /*else*/ {
    __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_qtype); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)
    __pyx_v__qtype = ((int)__pyx_t_6);
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":419
 *             _qtype = <int>qtype
 * 
 *         if cyares_check_qclasses(qclass) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_6 = cyares_check_qclasses(__pyx_v_qclass); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 419, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_6 < 0);
  if (unlikely(__pyx_t_1)) {

    /* "cyares/channel.pyx":420
 * 
 *         if cyares_check_qclasses(qclass) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 420, __pyx_L1_error)

    /* "cyares/channel.pyx":419
 *             _qtype = <int>qtype
 * 
 *         if cyares_check_qclasses(qclass) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyares/channel.pyx":422
 *             raise
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_6 = __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(__pyx_v_qname, (&__pyx_v_view)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 422, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_6 < 0);
  if (unlikely(__pyx_t_1)) {

    /* "cyares/channel.pyx":423
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         fut = self.__create_future(callback)
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 423, __pyx_L1_error)

    /* "cyares/channel.pyx":422
 *             raise
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyares/channel.pyx":425
 *             raise
 * 
 *         fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *         if _qtype == T_A:
 *             ares_query(
*/
  __pyx_t_8 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_fut = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "cyares/channel.pyx":426
 * 
 *         fut = self.__create_future(callback)
 *         if _qtype == T_A:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_A);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":427
 *         fut = self.__create_future(callback)
 *         if _qtype == T_A:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_A, __pyx_f_6cyares_9callbacks___callback_query_on_a, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":426
 * 
 *         fut = self.__create_future(callback)
 *         if _qtype == T_A:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":436
 *             )
 * 
 *         elif _qtype == T_AAAA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_AAAA);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":437
 * 
 *         elif _qtype == T_AAAA:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_AAAA, __pyx_f_6cyares_9callbacks___callback_query_on_aaaa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":436
 *             )
 * 
 *         elif _qtype == T_AAAA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":446
 *             )
 * 
 *         elif _qtype == T_CAA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_CAA);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":447
 * 
 *         elif _qtype == T_CAA:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_CAA, __pyx_f_6cyares_9callbacks___callback_query_on_caa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":446
 *             )
 * 
 *         elif _qtype == T_CAA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":456
 *             )
 * 
 *         elif _qtype == T_CNAME:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_CNAME);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":457
 * 
 *         elif _qtype == T_CNAME:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_CNAME, __pyx_f_6cyares_9callbacks___callback_query_on_cname, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":456
 *             )
 * 
 *         elif _qtype == T_CNAME:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":466
 *             )
 * 
 *         elif _qtype == T_MX:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_MX);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":467
 * 
 *         elif _qtype == T_MX:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_MX, __pyx_f_6cyares_9callbacks___callback_query_on_mx, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":466
 *             )
 * 
 *         elif _qtype == T_MX:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":476
 *             )
 * 
 *         elif _qtype == T_NAPTR:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_NAPTR);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":477
 * 
 *         elif _qtype == T_NAPTR:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_NAPTR, __pyx_f_6cyares_9callbacks___callback_query_on_naptr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":476
 *             )
 * 
 *         elif _qtype == T_NAPTR:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":486
 *             )
 * 
 *         elif _qtype == T_NS:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_NS);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":487
 * 
 *         elif _qtype == T_NS:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_NS, __pyx_f_6cyares_9callbacks___callback_query_on_ns, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":486
 *             )
 * 
 *         elif _qtype == T_NS:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":496
 *             )
 * 
 *         elif _qtype == T_PTR:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_PTR);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":497
 * 
 *         elif _qtype == T_PTR:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_PTR, __pyx_f_6cyares_9callbacks___callback_query_on_ptr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":496
 *             )
 * 
 *         elif _qtype == T_PTR:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":506
 *             )
 * 
 *         elif _qtype == T_SOA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_SOA);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":507
 * 
 *         elif _qtype == T_SOA:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_SOA, __pyx_f_6cyares_9callbacks___callback_query_on_soa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":506
 *             )
 * 
 *         elif _qtype == T_SOA:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":516
 *             )
 * 
 *         elif _qtype == T_SRV:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_SRV);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":517
 * 
 *         elif _qtype == T_SRV:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_SRV, __pyx_f_6cyares_9callbacks___callback_query_on_srv, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":516
 *             )
 * 
 *         elif _qtype == T_SRV:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":526
 *             )
 * 
 *         elif _qtype == T_TXT:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_TXT);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":527
 * 
 *         elif _qtype == T_TXT:
 *             ares_query(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_query(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_TXT, __pyx_f_6cyares_9callbacks___callback_query_on_txt, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":526
 *             )
 * 
 *         elif _qtype == T_TXT:             # <<<<<<<<<<<<<<
 *             ares_query(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":536
 *             )
 * 
 *         elif _qtype == T_ANY:             # <<<<<<<<<<<<<<
 *             ares_query_dnsrec(
 *                 self.channel,
*/
  __pyx_t_1 = (__pyx_v__qtype == T_ANY);
  if (likely(__pyx_t_1)) {

    /* "cyares/channel.pyx":537
 * 
 *         elif _qtype == T_ANY:
 *             ares_query_dnsrec(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    (void)(ares_query_dnsrec(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), ((ares_dns_class_t)__pyx_v_qclass), ARES_REC_TYPE_ANY, __pyx_f_6cyares_9callbacks___callback_dns_rec__any, ((void *)__pyx_v_fut), NULL));

    /* "cyares/channel.pyx":536
 *             )
 * 
 *         elif _qtype == T_ANY:             # <<<<<<<<<<<<<<
 *             ares_query_dnsrec(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":548
 * 
 *         else:
 *             Py_DECREF(fut)             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid query type specified")
 * 
*/
  /*else*/ {
    Py_DECREF(__pyx_v_fut);

    /* "cyares/channel.pyx":549
 *         else:
 *             Py_DECREF(fut)
 *             raise ValueError("invalid query type specified")             # <<<<<<<<<<<<<<
 * 
 *         return fut
*/
    __pyx_t_7 = NULL;
    __Pyx_INCREF(__pyx_builtin_ValueError);
    __pyx_t_5 = __pyx_builtin_ValueError; 
    __pyx_t_12 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_invalid_query_type_specified};
      __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_L14:;

  /* "cyares/channel.pyx":551
 *             raise ValueError("invalid query type specified")
 * 
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":406
 *     # query is the upper-end and is meant to assist in
 *     # being a theoretical drop in replacement for pycares in aiodns
 *     cdef object _query(self, object qname, object qtype, int qclass, object callback):             # <<<<<<<<<<<<<<
 *         cdef int _qtype
 *         cdef object fut
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cyares.channel.Channel._query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":553
 *         return fut
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_17query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_17query = {"query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_17query, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_17query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_query_type = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_query_class = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name_2,&__pyx_mstate_global->__pyx_n_u_query_type,&__pyx_mstate_global->__pyx_n_u_callback,&__pyx_mstate_global->__pyx_n_u_query_class,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 553, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "query", 0) < 0) __PYX_ERR(0, 553, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("query", 0, 2, 4, i); __PYX_ERR(0, 553, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 553, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 553, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 553, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_name = values[0];
    __pyx_v_query_type = values[1];
    __pyx_v_callback = values[2];
    __pyx_v_query_class = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 553, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_16query(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_name, __pyx_v_query_type, __pyx_v_callback, __pyx_v_query_class);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_16query(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_query_type, PyObject *__pyx_v_callback, PyObject *__pyx_v_query_class) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query", 0);

  /* "cyares/channel.pyx":554
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _search(self, object qname, object qtype, int qclass, object callback):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_query_class == Py_None);
  if (__pyx_t_2) {
    __pyx_t_1 = C_IN;
  } else {
    __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_v_query_class); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 554, __pyx_L1_error)
    __pyx_t_1 = ((int)__pyx_t_3);
  }
  __pyx_t_4 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_query(__pyx_v_self, __pyx_v_name, __pyx_v_query_type, __pyx_t_1, __pyx_v_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":553
 *         return fut
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyares.channel.Channel.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":556
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
 *     cdef object _search(self, object qname, object qtype, int qclass, object callback):             # <<<<<<<<<<<<<<
 *         cdef int _qtype
 *         cdef object fut = self.__create_future(callback)
*/

static PyObject *__pyx_f_6cyares_7channel_7Channel__search(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_qname, PyObject *__pyx_v_qtype, int __pyx_v_qclass, PyObject *__pyx_v_callback) {
  int __pyx_v__qtype;
  PyObject *__pyx_v_fut = 0;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_search", 0);

  /* "cyares/channel.pyx":558
 *     cdef object _search(self, object qname, object qtype, int qclass, object callback):
 *         cdef int _qtype
 *         cdef object fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 * 
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":562
 * 
 *         # TODO: use match ...: cases when cython releases match support
 *         if isinstance(qtype, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
*/
  __pyx_t_2 = PyUnicode_Check(__pyx_v_qtype); 
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":563
 *         # TODO: use match ...: cases when cython releases match support
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "cyares/channel.pyx":564
 *         if isinstance(qtype, str):
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 raise ValueError("invalid query type specified")
*/
        if (unlikely(__pyx_v_self->_query_lookups == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 564, __pyx_L4_error)
        }
        __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_query_lookups, __pyx_v_qtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L4_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v__qtype = ((int)__pyx_t_6);

        /* "cyares/channel.pyx":563
 *         # TODO: use match ...: cases when cython releases match support
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyares/channel.pyx":565
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 raise ValueError("invalid query type specified")
 *         else:
*/
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("cyares.channel.Channel._search", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 565, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);

        /* "cyares/channel.pyx":566
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
 *                 raise ValueError("invalid query type specified")             # <<<<<<<<<<<<<<
 *         else:
 *             _qtype = <int>qtype
*/
        __pyx_t_10 = NULL;
        __Pyx_INCREF(__pyx_builtin_ValueError);
        __pyx_t_11 = __pyx_builtin_ValueError; 
        __pyx_t_12 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_mstate_global->__pyx_kp_u_invalid_query_type_specified};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 566, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 566, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;

      /* "cyares/channel.pyx":563
 *         # TODO: use match ...: cases when cython releases match support
 *         if isinstance(qtype, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 _qtype = <int>self._query_lookups[qtype]
 *             except KeyError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "cyares/channel.pyx":562
 * 
 *         # TODO: use match ...: cases when cython releases match support
 *         if isinstance(qtype, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 _qtype = <int>self._query_lookups[qtype]
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":568
 *                 raise ValueError("invalid query type specified")
 *         else:
 *             _qtype = <int>qtype             # <<<<<<<<<<<<<<
 * 
 *         if cyares_check_qclasses(qclass) < 0:
*/
  /*else*/ {
    __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_qtype); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L1_error)
    __pyx_v__qtype = ((int)__pyx_t_6);
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":570
 *             _qtype = <int>qtype
 * 
 *         if cyares_check_qclasses(qclass) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_6 = cyares_check_qclasses(__pyx_v_qclass); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 570, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_6 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyares/channel.pyx":571
 * 
 *         if cyares_check_qclasses(qclass) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 571, __pyx_L1_error)

    /* "cyares/channel.pyx":570
 *             _qtype = <int>qtype
 * 
 *         if cyares_check_qclasses(qclass) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyares/channel.pyx":573
 *             raise
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  __pyx_t_6 = __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(__pyx_v_qname, (&__pyx_v_view)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_6 < 0);
  if (unlikely(__pyx_t_2)) {

    /* "cyares/channel.pyx":574
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:
 *             raise             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __Pyx_ReraiseException();
    __PYX_ERR(0, 574, __pyx_L1_error)

    /* "cyares/channel.pyx":573
 *             raise
 * 
 *         if cyares_get_domain_name_buffer(qname, &view) < 0:             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
  }

  /* "cyares/channel.pyx":577
 * 
 * 
 *         if _qtype == T_A:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_A);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":578
 * 
 *         if _qtype == T_A:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_A, __pyx_f_6cyares_9callbacks___callback_query_on_a, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":577
 * 
 * 
 *         if _qtype == T_A:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":587
 *             )
 * 
 *         elif _qtype == T_AAAA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_AAAA);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":588
 * 
 *         elif _qtype == T_AAAA:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_AAAA, __pyx_f_6cyares_9callbacks___callback_query_on_aaaa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":587
 *             )
 * 
 *         elif _qtype == T_AAAA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":597
 *             )
 * 
 *         elif _qtype == T_CAA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_CAA);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":598
 * 
 *         elif _qtype == T_CAA:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_CAA, __pyx_f_6cyares_9callbacks___callback_query_on_caa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":597
 *             )
 * 
 *         elif _qtype == T_CAA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":607
 *             )
 * 
 *         elif _qtype == T_CNAME:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_CNAME);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":608
 * 
 *         elif _qtype == T_CNAME:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_CNAME, __pyx_f_6cyares_9callbacks___callback_query_on_cname, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":607
 *             )
 * 
 *         elif _qtype == T_CNAME:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":617
 *             )
 * 
 *         elif _qtype == T_MX:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_MX);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":618
 * 
 *         elif _qtype == T_MX:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_MX, __pyx_f_6cyares_9callbacks___callback_query_on_mx, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":617
 *             )
 * 
 *         elif _qtype == T_MX:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":627
 *             )
 * 
 *         elif _qtype == T_NAPTR:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_NAPTR);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":628
 * 
 *         elif _qtype == T_NAPTR:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_NAPTR, __pyx_f_6cyares_9callbacks___callback_query_on_naptr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":627
 *             )
 * 
 *         elif _qtype == T_NAPTR:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":637
 *             )
 * 
 *         elif _qtype == T_NS:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_NS);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":638
 * 
 *         elif _qtype == T_NS:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_NS, __pyx_f_6cyares_9callbacks___callback_query_on_ns, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":637
 *             )
 * 
 *         elif _qtype == T_NS:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":647
 *             )
 * 
 *         elif _qtype == T_PTR:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_PTR);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":648
 * 
 *         elif _qtype == T_PTR:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_PTR, __pyx_f_6cyares_9callbacks___callback_query_on_ptr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":647
 *             )
 * 
 *         elif _qtype == T_PTR:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":657
 *             )
 * 
 *         elif _qtype == T_SOA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_SOA);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":658
 * 
 *         elif _qtype == T_SOA:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_SOA, __pyx_f_6cyares_9callbacks___callback_query_on_soa, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":657
 *             )
 * 
 *         elif _qtype == T_SOA:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":667
 *             )
 * 
 *         elif _qtype == T_SRV:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_SRV);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":668
 * 
 *         elif _qtype == T_SRV:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_SRV, __pyx_f_6cyares_9callbacks___callback_query_on_srv, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":667
 *             )
 * 
 *         elif _qtype == T_SRV:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":677
 *             )
 * 
 *         elif _qtype == T_TXT:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
  __pyx_t_2 = (__pyx_v__qtype == T_TXT);
  if (likely(__pyx_t_2)) {

    /* "cyares/channel.pyx":678
 * 
 *         elif _qtype == T_TXT:
 *             ares_search(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <char*>view.buf,
*/
    ares_search(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_qclass, T_TXT, __pyx_f_6cyares_9callbacks___callback_query_on_txt, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":677
 *             )
 * 
 *         elif _qtype == T_TXT:             # <<<<<<<<<<<<<<
 *             ares_search(
 *                 self.channel,
*/
    goto __pyx_L14;
  }

  /* "cyares/channel.pyx":695
 * 
 *         else:
 *             raise ValueError("invalid query type specified")             # <<<<<<<<<<<<<<
 * 
 *         return fut
*/
  /*else*/ {
    __pyx_t_7 = NULL;
    __Pyx_INCREF(__pyx_builtin_ValueError);
    __pyx_t_1 = __pyx_builtin_ValueError; 
    __pyx_t_12 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_invalid_query_type_specified};
      __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 695, __pyx_L1_error)
  }
  __pyx_L14:;

  /* "cyares/channel.pyx":697
 *             raise ValueError("invalid query type specified")
 * 
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     def search(self, object name, object query_type, object callback = None , object query_class = None):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":556
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
 *     cdef object _search(self, object qname, object qtype, int qclass, object callback):             # <<<<<<<<<<<<<<
 *         cdef int _qtype
 *         cdef object fut = self.__create_future(callback)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cyares.channel.Channel._search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":699
 *         return fut
 * 
 *     def search(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_19search(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_19search = {"search", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_19search, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_19search(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_query_type = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_query_class = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("search (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name_2,&__pyx_mstate_global->__pyx_n_u_query_type,&__pyx_mstate_global->__pyx_n_u_callback,&__pyx_mstate_global->__pyx_n_u_query_class,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 699, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "search", 0) < 0) __PYX_ERR(0, 699, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("search", 0, 2, 4, i); __PYX_ERR(0, 699, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 699, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 699, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 699, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_name = values[0];
    __pyx_v_query_type = values[1];
    __pyx_v_callback = values[2];
    __pyx_v_query_class = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("search", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 699, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_18search(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_name, __pyx_v_query_type, __pyx_v_callback, __pyx_v_query_class);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_18search(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_query_type, PyObject *__pyx_v_callback, PyObject *__pyx_v_query_class) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("search", 0);

  /* "cyares/channel.pyx":700
 * 
 *     def search(self, object name, object query_type, object callback = None , object query_class = None):
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)             # <<<<<<<<<<<<<<
 * 
 *     def process_fd(self, int read_fd, int write_fd):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_query_class == Py_None);
  if (__pyx_t_2) {
    __pyx_t_1 = C_IN;
  } else {
    __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_v_query_class); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 700, __pyx_L1_error)
    __pyx_t_1 = ((int)__pyx_t_3);
  }
  __pyx_t_4 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_search(__pyx_v_self, __pyx_v_name, __pyx_v_query_type, __pyx_t_1, __pyx_v_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":699
 *         return fut
 * 
 *     def search(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyares.channel.Channel.search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":702
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
 *     def process_fd(self, int read_fd, int write_fd):             # <<<<<<<<<<<<<<
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, <ares_socket_t>write_fd)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_21process_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_21process_fd = {"process_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_21process_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_21process_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_read_fd;
  int __pyx_v_write_fd;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_fd (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_read_fd,&__pyx_mstate_global->__pyx_n_u_write_fd,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 702, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "process_fd", 0) < 0) __PYX_ERR(0, 702, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("process_fd", 1, 2, 2, i); __PYX_ERR(0, 702, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 702, __pyx_L3_error)
    }
    __pyx_v_read_fd = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_read_fd == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_write_fd = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_write_fd == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_fd", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 702, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.process_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_20process_fd(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_read_fd, __pyx_v_write_fd);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_20process_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_read_fd, int __pyx_v_write_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_fd", 0);

  /* "cyares/channel.pyx":703
 * 
 *     def process_fd(self, int read_fd, int write_fd):
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, <ares_socket_t>write_fd)             # <<<<<<<<<<<<<<
 * 
 *     # === Custom ===
*/
  ares_process_fd(__pyx_v_self->channel, ((ares_socket_t)__pyx_v_read_fd), ((ares_socket_t)__pyx_v_write_fd));

  /* "cyares/channel.pyx":702
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
 *     def process_fd(self, int read_fd, int write_fd):             # <<<<<<<<<<<<<<
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, <ares_socket_t>write_fd)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":708
 *     # NOTE: I'll make a pull-request to pycares to maybe implement this as well...
 * 
 *     def process_read_fd(self, int read_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes readable file-descriptor instead of needing to remember
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_23process_read_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6cyares_7channel_7Channel_22process_read_fd, "\n        processes readable file-descriptor instead of needing to remember \n        to set write-fd to CYARES_SOCKET_BAD\n\n        Parameters\n        ----------\n\n        :param read_fd: the readable file descriptor\n        ");
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_23process_read_fd = {"process_read_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_23process_read_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cyares_7channel_7Channel_22process_read_fd};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_23process_read_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_read_fd;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_read_fd (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_read_fd,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 708, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "process_read_fd", 0) < 0) __PYX_ERR(0, 708, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("process_read_fd", 1, 1, 1, i); __PYX_ERR(0, 708, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 708, __pyx_L3_error)
    }
    __pyx_v_read_fd = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_read_fd == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_read_fd", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 708, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.process_read_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_22process_read_fd(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_read_fd);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_22process_read_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_read_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_read_fd", 0);

  /* "cyares/channel.pyx":719
 *         """
 * 
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, ARES_SOCKET_BAD)             # <<<<<<<<<<<<<<
 * 
 *     def process_write_fd(self, int write_fd):
*/
  ares_process_fd(__pyx_v_self->channel, ((ares_socket_t)__pyx_v_read_fd), ARES_SOCKET_BAD);

  /* "cyares/channel.pyx":708
 *     # NOTE: I'll make a pull-request to pycares to maybe implement this as well...
 * 
 *     def process_read_fd(self, int read_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes readable file-descriptor instead of needing to remember
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":721
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, ARES_SOCKET_BAD)
 * 
 *     def process_write_fd(self, int write_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes writable file-descriptor instead of needing to remember
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_25process_write_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6cyares_7channel_7Channel_24process_write_fd, "\n        processes writable file-descriptor instead of needing to remember \n        to set write-fd to CYARES_SOCKET_BAD\n\n        Parameters\n        ----------\n\n        :param write_fd: the readable file descriptor\n        ");
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_25process_write_fd = {"process_write_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_25process_write_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cyares_7channel_7Channel_24process_write_fd};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_25process_write_fd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_write_fd;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_write_fd (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_write_fd,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 721, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 721, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "process_write_fd", 0) < 0) __PYX_ERR(0, 721, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("process_write_fd", 1, 1, 1, i); __PYX_ERR(0, 721, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 721, __pyx_L3_error)
    }
    __pyx_v_write_fd = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_write_fd == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_write_fd", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 721, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.process_write_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_24process_write_fd(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_write_fd);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_24process_write_fd(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_write_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_write_fd", 0);

  /* "cyares/channel.pyx":732
 *         """
 * 
 *         ares_process_fd(self.channel, ARES_SOCKET_BAD, <ares_socket_t>write_fd)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  ares_process_fd(__pyx_v_self->channel, ARES_SOCKET_BAD, ((ares_socket_t)__pyx_v_write_fd));

  /* "cyares/channel.pyx":721
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, ARES_SOCKET_BAD)
 * 
 *     def process_write_fd(self, int write_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes writable file-descriptor instead of needing to remember
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":735
 * 
 * 
 *     def timeout(self, double t = 0):             # <<<<<<<<<<<<<<
 *         cdef timeval maxtv
 *         cdef timeval tv
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_27timeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_27timeout = {"timeout", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_27timeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_27timeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_t;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("timeout (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_t,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 735, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 735, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "timeout", 0) < 0) __PYX_ERR(0, 735, __pyx_L3_error)
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 735, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_t = __Pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 735, __pyx_L3_error)
    } else {
      __pyx_v_t = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("timeout", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 735, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_26timeout(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_t);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_26timeout(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, double __pyx_v_t) {
  struct timeval __pyx_v_maxtv;
  struct timeval __pyx_v_tv;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timeout", 0);

  /* "cyares/channel.pyx":739
 *         cdef timeval tv
 * 
 *         if t:             # <<<<<<<<<<<<<<
 *             if t >= 0.0:
 *                 maxtv.tv_sec = <time_t>floor(t)
*/
  __pyx_t_1 = (__pyx_v_t != 0);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":740
 * 
 *         if t:
 *             if t >= 0.0:             # <<<<<<<<<<<<<<
 *                 maxtv.tv_sec = <time_t>floor(t)
 *                 maxtv.tv_usec = <long>(fmod(t, 1.0) * 1000000)
*/
    __pyx_t_1 = (__pyx_v_t >= 0.0);
    if (likely(__pyx_t_1)) {

      /* "cyares/channel.pyx":741
 *         if t:
 *             if t >= 0.0:
 *                 maxtv.tv_sec = <time_t>floor(t)             # <<<<<<<<<<<<<<
 *                 maxtv.tv_usec = <long>(fmod(t, 1.0) * 1000000)
 *             else:
*/
      __pyx_v_maxtv.tv_sec = ((time_t)floor(__pyx_v_t));

      /* "cyares/channel.pyx":742
 *             if t >= 0.0:
 *                 maxtv.tv_sec = <time_t>floor(t)
 *                 maxtv.tv_usec = <long>(fmod(t, 1.0) * 1000000)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError("timeout needs to be a positive number or 0.0")
*/
      __pyx_v_maxtv.tv_usec = ((long)(fmod(__pyx_v_t, 1.0) * 1000000.0));

      /* "cyares/channel.pyx":740
 * 
 *         if t:
 *             if t >= 0.0:             # <<<<<<<<<<<<<<
 *                 maxtv.tv_sec = <time_t>floor(t)
 *                 maxtv.tv_usec = <long>(fmod(t, 1.0) * 1000000)
*/
      goto __pyx_L4;
    }

    /* "cyares/channel.pyx":744
 *                 maxtv.tv_usec = <long>(fmod(t, 1.0) * 1000000)
 *             else:
 *                 raise ValueError("timeout needs to be a positive number or 0.0")             # <<<<<<<<<<<<<<
 *         else:
 *             # no segfaulting!
*/
    /*else*/ {
      __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_builtin_ValueError);
      __pyx_t_4 = __pyx_builtin_ValueError; 
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_timeout_needs_to_be_a_positive_n};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 744, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "cyares/channel.pyx":739
 *         cdef timeval tv
 * 
 *         if t:             # <<<<<<<<<<<<<<
 *             if t >= 0.0:
 *                 maxtv.tv_sec = <time_t>floor(t)
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":747
 *         else:
 *             # no segfaulting!
 *             maxtv.tv_sec = 0             # <<<<<<<<<<<<<<
 *             maxtv.tv_usec = 0
 * 
*/
  /*else*/ {
    __pyx_v_maxtv.tv_sec = 0;

    /* "cyares/channel.pyx":748
 *             # no segfaulting!
 *             maxtv.tv_sec = 0
 *             maxtv.tv_usec = 0             # <<<<<<<<<<<<<<
 * 
 *         ares_timeout(self.channel, &maxtv, &tv)
*/
    __pyx_v_maxtv.tv_usec = 0;
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":750
 *             maxtv.tv_usec = 0
 * 
 *         ares_timeout(self.channel, &maxtv, &tv)             # <<<<<<<<<<<<<<
 * 
 *         if not (tv.tv_sec and tv.tv_usec):
*/
  (void)(ares_timeout(__pyx_v_self->channel, (&__pyx_v_maxtv), (&__pyx_v_tv)));

  /* "cyares/channel.pyx":752
 *         ares_timeout(self.channel, &maxtv, &tv)
 * 
 *         if not (tv.tv_sec and tv.tv_usec):             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
*/
  __pyx_t_6 = (__pyx_v_tv.tv_sec != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_v_tv.tv_usec != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  __pyx_t_6 = (!__pyx_t_1);
  if (__pyx_t_6) {

    /* "cyares/channel.pyx":753
 * 
 *         if not (tv.tv_sec and tv.tv_usec):
 *             return 0.0             # <<<<<<<<<<<<<<
 * 
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_float_0_0);
    __pyx_r = __pyx_mstate_global->__pyx_float_0_0;
    goto __pyx_L0;

    /* "cyares/channel.pyx":752
 *         ares_timeout(self.channel, &maxtv, &tv)
 * 
 *         if not (tv.tv_sec and tv.tv_usec):             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
*/
  }

  /* "cyares/channel.pyx":755
 *             return 0.0
 * 
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)             # <<<<<<<<<<<<<<
 * 
 *     def getaddrinfo(
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_tv.tv_sec + (((double)__pyx_v_tv.tv_usec) / 1000000.0))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":735
 * 
 * 
 *     def timeout(self, double t = 0):             # <<<<<<<<<<<<<<
 *         cdef timeval maxtv
 *         cdef timeval tv
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyares.channel.Channel.timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":757
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 * 
 *     def getaddrinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         object host,
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_29getaddrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_29getaddrinfo = {"getaddrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_29getaddrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_29getaddrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  PyObject *__pyx_v_callback = 0;
  int __pyx_v_family;
  int __pyx_v_socktype;
  int __pyx_v_proto;
  int __pyx_v_flags;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getaddrinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_host,&__pyx_mstate_global->__pyx_n_u_port,&__pyx_mstate_global->__pyx_n_u_callback,&__pyx_mstate_global->__pyx_n_u_family,&__pyx_mstate_global->__pyx_n_u_socktype,&__pyx_mstate_global->__pyx_n_u_proto,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 757, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "getaddrinfo", 0) < 0) __PYX_ERR(0, 757, __pyx_L3_error)

      /* "cyares/channel.pyx":760
 *         self,
 *         object host,
 *         object port = None,             # <<<<<<<<<<<<<<
 *         object callback = None,
 *         int family = 0,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":761
 *         object host,
 *         object port = None,
 *         object callback = None,             # <<<<<<<<<<<<<<
 *         int family = 0,
 *         int socktype = 0,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("getaddrinfo", 0, 1, 7, i); __PYX_ERR(0, 757, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 757, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 757, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "cyares/channel.pyx":760
 *         self,
 *         object host,
 *         object port = None,             # <<<<<<<<<<<<<<
 *         object callback = None,
 *         int family = 0,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "cyares/channel.pyx":761
 *         object host,
 *         object port = None,
 *         object callback = None,             # <<<<<<<<<<<<<<
 *         int family = 0,
 *         int socktype = 0,
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_host = values[0];
    __pyx_v_port = values[1];
    __pyx_v_callback = values[2];
    if (values[3]) {
      __pyx_v_family = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    } else {
      __pyx_v_family = ((int)0);
    }
    if (values[4]) {
      __pyx_v_socktype = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_socktype == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L3_error)
    } else {
      __pyx_v_socktype = ((int)0);
    }
    if (values[5]) {
      __pyx_v_proto = __Pyx_PyLong_As_int(values[5]); if (unlikely((__pyx_v_proto == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 764, __pyx_L3_error)
    } else {
      __pyx_v_proto = ((int)0);
    }
    if (values[6]) {
      __pyx_v_flags = __Pyx_PyLong_As_int(values[6]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L3_error)
    } else {
      __pyx_v_flags = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getaddrinfo", 0, 1, 7, __pyx_nargs); __PYX_ERR(0, 757, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_28getaddrinfo(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_host, __pyx_v_port, __pyx_v_callback, __pyx_v_family, __pyx_v_socktype, __pyx_v_proto, __pyx_v_flags);

  /* "cyares/channel.pyx":757
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 * 
 *     def getaddrinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         object host,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_28getaddrinfo(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_callback, int __pyx_v_family, int __pyx_v_socktype, int __pyx_v_proto, int __pyx_v_flags) {
  char *__pyx_v_service;
  Py_buffer __pyx_v_view;
  Py_buffer __pyx_v_service_data;
  int __pyx_v_buffer_carried;
  struct ares_addrinfo_hints __pyx_v_hints;
  PyObject *__pyx_v_fut = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getaddrinfo", 0);
  __Pyx_INCREF(__pyx_v_port);

  /* "cyares/channel.pyx":767
 *         int flags = 0
 *     ):
 *         cdef char* service = NULL             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view, service_data
 *         cdef bint buffer_carried = 0
*/
  __pyx_v_service = NULL;

  /* "cyares/channel.pyx":769
 *         cdef char* service = NULL
 *         cdef Py_buffer view, service_data
 *         cdef bint buffer_carried = 0             # <<<<<<<<<<<<<<
 *         cdef ares_addrinfo_hints hints
 * 
*/
  __pyx_v_buffer_carried = 0;

  /* "cyares/channel.pyx":772
 *         cdef ares_addrinfo_hints hints
 * 
 *         cdef object fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 * 
 *         if port:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":774
 *         cdef object fut = self.__create_future(callback)
 * 
 *         if port:             # <<<<<<<<<<<<<<
 *             if isinstance(port, int):
 *                 port = bytes(port)
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_port); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 774, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":775
 * 
 *         if port:
 *             if isinstance(port, int):             # <<<<<<<<<<<<<<
 *                 port = bytes(port)
 *             cyares_get_buffer(port, &service_data)
*/
    __pyx_t_2 = PyLong_Check(__pyx_v_port); 
    if (__pyx_t_2) {

      /* "cyares/channel.pyx":776
 *         if port:
 *             if isinstance(port, int):
 *                 port = bytes(port)             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(port, &service_data)
 *             service = <char*>service_data.buf
*/
      __pyx_t_3 = NULL;
      __Pyx_INCREF((PyObject *)(&PyBytes_Type));
      __pyx_t_4 = ((PyObject *)(&PyBytes_Type)); 
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_port};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "cyares/channel.pyx":775
 * 
 *         if port:
 *             if isinstance(port, int):             # <<<<<<<<<<<<<<
 *                 port = bytes(port)
 *             cyares_get_buffer(port, &service_data)
*/
    }

    /* "cyares/channel.pyx":777
 *             if isinstance(port, int):
 *                 port = bytes(port)
 *             cyares_get_buffer(port, &service_data)             # <<<<<<<<<<<<<<
 *             service = <char*>service_data.buf
 *             buffer_carried = 1
*/
    __pyx_t_6 = cyares_get_buffer(__pyx_v_port, (&__pyx_v_service_data)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 777, __pyx_L1_error)

    /* "cyares/channel.pyx":778
 *                 port = bytes(port)
 *             cyares_get_buffer(port, &service_data)
 *             service = <char*>service_data.buf             # <<<<<<<<<<<<<<
 *             buffer_carried = 1
 * 
*/
    __pyx_v_service = ((char *)__pyx_v_service_data.buf);

    /* "cyares/channel.pyx":779
 *             cyares_get_buffer(port, &service_data)
 *             service = <char*>service_data.buf
 *             buffer_carried = 1             # <<<<<<<<<<<<<<
 * 
 *         cyares_get_domain_name_buffer(host, &view)
*/
    __pyx_v_buffer_carried = 1;

    /* "cyares/channel.pyx":774
 *         cdef object fut = self.__create_future(callback)
 * 
 *         if port:             # <<<<<<<<<<<<<<
 *             if isinstance(port, int):
 *                 port = bytes(port)
*/
  }

  /* "cyares/channel.pyx":781
 *             buffer_carried = 1
 * 
 *         cyares_get_domain_name_buffer(host, &view)             # <<<<<<<<<<<<<<
 * 
 *         hints.ai_flags = flags
*/
  __pyx_t_6 = __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(__pyx_v_host, (&__pyx_v_view)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 781, __pyx_L1_error)

  /* "cyares/channel.pyx":783
 *         cyares_get_domain_name_buffer(host, &view)
 * 
 *         hints.ai_flags = flags             # <<<<<<<<<<<<<<
 *         hints.ai_family = family
 *         hints.ai_socktype = socktype
*/
  __pyx_v_hints.ai_flags = __pyx_v_flags;

  /* "cyares/channel.pyx":784
 * 
 *         hints.ai_flags = flags
 *         hints.ai_family = family             # <<<<<<<<<<<<<<
 *         hints.ai_socktype = socktype
 *         hints.ai_protocol = proto
*/
  __pyx_v_hints.ai_family = __pyx_v_family;

  /* "cyares/channel.pyx":785
 *         hints.ai_flags = flags
 *         hints.ai_family = family
 *         hints.ai_socktype = socktype             # <<<<<<<<<<<<<<
 *         hints.ai_protocol = proto
 * 
*/
  __pyx_v_hints.ai_socktype = __pyx_v_socktype;

  /* "cyares/channel.pyx":786
 *         hints.ai_family = family
 *         hints.ai_socktype = socktype
 *         hints.ai_protocol = proto             # <<<<<<<<<<<<<<
 * 
 *         ares_getaddrinfo(
*/
  __pyx_v_hints.ai_protocol = __pyx_v_proto;

  /* "cyares/channel.pyx":788
 *         hints.ai_protocol = proto
 * 
 *         ares_getaddrinfo(             # <<<<<<<<<<<<<<
 *             self.channel,
 *             <char*>view.buf,
*/
  ares_getaddrinfo(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_service, (&__pyx_v_hints), __pyx_f_6cyares_9callbacks___callback_getaddrinfo, ((void *)__pyx_v_fut));

  /* "cyares/channel.pyx":797
 *         )
 * 
 *         cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         if buffer_carried:
 *             cyares_release_buffer(&service_data)
*/
  cyares_release_buffer((&__pyx_v_view));

  /* "cyares/channel.pyx":798
 * 
 *         cyares_release_buffer(&view)
 *         if buffer_carried:             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&service_data)
 *         if callback:
*/
  if (__pyx_v_buffer_carried) {

    /* "cyares/channel.pyx":799
 *         cyares_release_buffer(&view)
 *         if buffer_carried:
 *             cyares_release_buffer(&service_data)             # <<<<<<<<<<<<<<
 *         if callback:
 *             fut.add_done_callback(callback)
*/
    cyares_release_buffer((&__pyx_v_service_data));

    /* "cyares/channel.pyx":798
 * 
 *         cyares_release_buffer(&view)
 *         if buffer_carried:             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&service_data)
 *         if callback:
*/
  }

  /* "cyares/channel.pyx":800
 *         if buffer_carried:
 *             cyares_release_buffer(&service_data)
 *         if callback:             # <<<<<<<<<<<<<<
 *             fut.add_done_callback(callback)
 *         return fut
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_callback); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 800, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":801
 *             cyares_release_buffer(&service_data)
 *         if callback:
 *             fut.add_done_callback(callback)             # <<<<<<<<<<<<<<
 *         return fut
 * 
*/
    __pyx_t_4 = __pyx_v_fut;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_callback};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_add_done_callback, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "cyares/channel.pyx":800
 *         if buffer_carried:
 *             cyares_release_buffer(&service_data)
 *         if callback:             # <<<<<<<<<<<<<<
 *             fut.add_done_callback(callback)
 *         return fut
*/
  }

  /* "cyares/channel.pyx":802
 *         if callback:
 *             fut.add_done_callback(callback)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     def getnameinfo(
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":757
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 * 
 *     def getaddrinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         object host,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyares.channel.Channel.getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":804
 *         return fut
 * 
 *     def getnameinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         tuple address,
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_31getnameinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_31getnameinfo = {"getnameinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_31getnameinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_31getnameinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_address = 0;
  int __pyx_v_flags;
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getnameinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_address,&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_callback,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 804, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "getnameinfo", 0) < 0) __PYX_ERR(0, 804, __pyx_L3_error)

      /* "cyares/channel.pyx":808
 *         tuple address,
 *         int flags,
 *         object callback = None             # <<<<<<<<<<<<<<
 *     ):
 *         cdef sockaddr_in sa4
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("getnameinfo", 0, 2, 3, i); __PYX_ERR(0, 804, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_address = ((PyObject*)values[0]);
    __pyx_v_flags = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 807, __pyx_L3_error)
    __pyx_v_callback = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getnameinfo", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 804, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_address), (&PyTuple_Type), 1, "address", 1))) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_30getnameinfo(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_address, __pyx_v_flags, __pyx_v_callback);

  /* "cyares/channel.pyx":804
 *         return fut
 * 
 *     def getnameinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         tuple address,
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_30getnameinfo(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_address, int __pyx_v_flags, PyObject *__pyx_v_callback) {
  struct sockaddr_in __pyx_v_sa4;
  struct sockaddr_in6 __pyx_v_sa6;
  PyObject *__pyx_v_fut = 0;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_v_ip = NULL;
  PyObject *__pyx_v_port = NULL;
  PyObject *__pyx_v_flowinfo = NULL;
  PyObject *__pyx_v_scope_id = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  unsigned short __pyx_t_9;
  unsigned long __pyx_t_10;
  uint32_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getnameinfo", 0);

  /* "cyares/channel.pyx":814
 *         cdef object fut
 *         cdef Py_buffer view
 *         if len(address) == 2:             # <<<<<<<<<<<<<<
 *             ip, port = address
 *             cyares_get_buffer(ip, &view)
*/
  if (unlikely(__pyx_v_address == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 814, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_address); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 814, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 2);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":815
 *         cdef Py_buffer view
 *         if len(address) == 2:
 *             ip, port = address             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):
*/
    if (likely(__pyx_v_address != Py_None)) {
      PyObject* sequence = __pyx_v_address;
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 815, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 815, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 815, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 815, __pyx_L1_error)
    }
    __pyx_v_ip = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_port = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "cyares/channel.pyx":816
 *         if len(address) == 2:
 *             ip, port = address
 *             cyares_get_buffer(ip, &view)             # <<<<<<<<<<<<<<
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):
 *                 cyares_release_buffer(&view)
*/
    __pyx_t_5 = cyares_get_buffer(__pyx_v_ip, (&__pyx_v_view)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 816, __pyx_L1_error)

    /* "cyares/channel.pyx":817
 *             ip, port = address
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):             # <<<<<<<<<<<<<<
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv4 address %r" % ip)
*/
    __pyx_t_2 = (!(ares_inet_pton(AF_INET, ((char *)__pyx_v_view.buf), (&__pyx_v_sa4.sin_addr)) != 0));
    if (unlikely(__pyx_t_2)) {

      /* "cyares/channel.pyx":818
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):
 *                 cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid IPv4 address %r" % ip)
 *             sa4.sin_family = AF_INET
*/
      cyares_release_buffer((&__pyx_v_view));

      /* "cyares/channel.pyx":819
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv4 address %r" % ip)             # <<<<<<<<<<<<<<
 *             sa4.sin_family = AF_INET
 *             sa4.sin_port = cyares_htons(port)
*/
      __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_builtin_ValueError);
      __pyx_t_6 = __pyx_builtin_ValueError; 
      __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Invalid_IPv4_address_r, __pyx_v_ip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 819, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 819, __pyx_L1_error)

      /* "cyares/channel.pyx":817
 *             ip, port = address
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET, <char*>view.buf, &sa4.sin_addr):             # <<<<<<<<<<<<<<
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv4 address %r" % ip)
*/
    }

    /* "cyares/channel.pyx":820
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv4 address %r" % ip)
 *             sa4.sin_family = AF_INET             # <<<<<<<<<<<<<<
 *             sa4.sin_port = cyares_htons(port)
 *             cyares_release_buffer(&view)
*/
    __pyx_v_sa4.sin_family = AF_INET;

    /* "cyares/channel.pyx":821
 *                 raise ValueError("Invalid IPv4 address %r" % ip)
 *             sa4.sin_family = AF_INET
 *             sa4.sin_port = cyares_htons(port)             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)
*/
    __pyx_t_9 = __Pyx_PyLong_As_unsigned_short(__pyx_v_port); if (unlikely((__pyx_t_9 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L1_error)
    __pyx_v_sa4.sin_port = cyares_htons(__pyx_t_9);

    /* "cyares/channel.pyx":822
 *             sa4.sin_family = AF_INET
 *             sa4.sin_port = cyares_htons(port)
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_getnameinfo(
*/
    cyares_release_buffer((&__pyx_v_view));

    /* "cyares/channel.pyx":823
 *             sa4.sin_port = cyares_htons(port)
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *             ares_getnameinfo(
 *                 self.channel,
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_fut = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "cyares/channel.pyx":824
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)
 *             ares_getnameinfo(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <sockaddr*>&sa4,
*/
    ares_getnameinfo(__pyx_v_self->channel, ((struct sockaddr *)(&__pyx_v_sa4)), (sizeof(__pyx_v_sa4)), __pyx_v_flags, __pyx_f_6cyares_9callbacks___callback_nameinfo, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":814
 *         cdef object fut
 *         cdef Py_buffer view
 *         if len(address) == 2:             # <<<<<<<<<<<<<<
 *             ip, port = address
 *             cyares_get_buffer(ip, &view)
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":832
 *                 <void*>fut
 *             )
 *         elif len(address) == 4:             # <<<<<<<<<<<<<<
 *             (ip, port, flowinfo, scope_id) = address
 *             cyares_get_buffer(ip, &view)
*/
  if (unlikely(__pyx_v_address == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 832, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_address); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 832, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 4);
  if (likely(__pyx_t_2)) {

    /* "cyares/channel.pyx":833
 *             )
 *         elif len(address) == 4:
 *             (ip, port, flowinfo, scope_id) = address             # <<<<<<<<<<<<<<
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):
*/
    if (likely(__pyx_v_address != Py_None)) {
      PyObject* sequence = __pyx_v_address;
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(sequence);
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 833, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 3);
      __Pyx_INCREF(__pyx_t_3);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_4,&__pyx_t_6,&__pyx_t_7,&__pyx_t_3};
        for (i=0; i < 4; i++) {
          PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 833, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 833, __pyx_L1_error)
    }
    __pyx_v_ip = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_port = __pyx_t_6;
    __pyx_t_6 = 0;
    __pyx_v_flowinfo = __pyx_t_7;
    __pyx_t_7 = 0;
    __pyx_v_scope_id = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyares/channel.pyx":834
 *         elif len(address) == 4:
 *             (ip, port, flowinfo, scope_id) = address
 *             cyares_get_buffer(ip, &view)             # <<<<<<<<<<<<<<
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):
 *                 cyares_release_buffer(&view)
*/
    __pyx_t_5 = cyares_get_buffer(__pyx_v_ip, (&__pyx_v_view)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 834, __pyx_L1_error)

    /* "cyares/channel.pyx":835
 *             (ip, port, flowinfo, scope_id) = address
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):             # <<<<<<<<<<<<<<
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv6 address %r" % ip)
*/
    __pyx_t_2 = (!(ares_inet_pton(AF_INET6, ((char *)__pyx_v_view.buf), (&__pyx_v_sa6.sin6_addr)) != 0));
    if (unlikely(__pyx_t_2)) {

      /* "cyares/channel.pyx":836
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):
 *                 cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid IPv6 address %r" % ip)
 *             sa6.sin6_family = AF_INET6
*/
      cyares_release_buffer((&__pyx_v_view));

      /* "cyares/channel.pyx":837
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv6 address %r" % ip)             # <<<<<<<<<<<<<<
 *             sa6.sin6_family = AF_INET6
 *             sa6.sin6_port = cyares_htons(port)
*/
      __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_builtin_ValueError);
      __pyx_t_6 = __pyx_builtin_ValueError; 
      __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Invalid_IPv6_address_r, __pyx_v_ip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 837, __pyx_L1_error)

      /* "cyares/channel.pyx":835
 *             (ip, port, flowinfo, scope_id) = address
 *             cyares_get_buffer(ip, &view)
 *             if not ares_inet_pton(AF_INET6, <char*>view.buf, &sa6.sin6_addr):             # <<<<<<<<<<<<<<
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv6 address %r" % ip)
*/
    }

    /* "cyares/channel.pyx":838
 *                 cyares_release_buffer(&view)
 *                 raise ValueError("Invalid IPv6 address %r" % ip)
 *             sa6.sin6_family = AF_INET6             # <<<<<<<<<<<<<<
 *             sa6.sin6_port = cyares_htons(port)
 *             sa6.sin6_flowinfo = cyares_htonl(flowinfo) # Pycares Comment: I'm unsure about byteorder here.
*/
    __pyx_v_sa6.sin6_family = AF_INET6;

    /* "cyares/channel.pyx":839
 *                 raise ValueError("Invalid IPv6 address %r" % ip)
 *             sa6.sin6_family = AF_INET6
 *             sa6.sin6_port = cyares_htons(port)             # <<<<<<<<<<<<<<
 *             sa6.sin6_flowinfo = cyares_htonl(flowinfo) # Pycares Comment: I'm unsure about byteorder here.
 *             sa6.sin6_scope_id = scope_id # Pycares Comment: Yes, without htonl.
*/
    __pyx_t_9 = __Pyx_PyLong_As_unsigned_short(__pyx_v_port); if (unlikely((__pyx_t_9 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L1_error)
    __pyx_v_sa6.sin6_port = cyares_htons(__pyx_t_9);

    /* "cyares/channel.pyx":840
 *             sa6.sin6_family = AF_INET6
 *             sa6.sin6_port = cyares_htons(port)
 *             sa6.sin6_flowinfo = cyares_htonl(flowinfo) # Pycares Comment: I'm unsure about byteorder here.             # <<<<<<<<<<<<<<
 *             sa6.sin6_scope_id = scope_id # Pycares Comment: Yes, without htonl.
 *             cyares_release_buffer(&view)
*/
    __pyx_t_10 = __Pyx_PyLong_As_unsigned_long(__pyx_v_flowinfo); if (unlikely((__pyx_t_10 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L1_error)
    __pyx_v_sa6.sin6_flowinfo = cyares_htonl(__pyx_t_10);

    /* "cyares/channel.pyx":841
 *             sa6.sin6_port = cyares_htons(port)
 *             sa6.sin6_flowinfo = cyares_htonl(flowinfo) # Pycares Comment: I'm unsure about byteorder here.
 *             sa6.sin6_scope_id = scope_id # Pycares Comment: Yes, without htonl.             # <<<<<<<<<<<<<<
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)
*/
    __pyx_t_11 = __Pyx_PyLong_As_uint32_t(__pyx_v_scope_id); if (unlikely((__pyx_t_11 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L1_error)
    __pyx_v_sa6.sin6_scope_id = __pyx_t_11;

    /* "cyares/channel.pyx":842
 *             sa6.sin6_flowinfo = cyares_htonl(flowinfo) # Pycares Comment: I'm unsure about byteorder here.
 *             sa6.sin6_scope_id = scope_id # Pycares Comment: Yes, without htonl.
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_getnameinfo(
*/
    cyares_release_buffer((&__pyx_v_view));

    /* "cyares/channel.pyx":843
 *             sa6.sin6_scope_id = scope_id # Pycares Comment: Yes, without htonl.
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *             ares_getnameinfo(
 *                 self.channel,
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_fut = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyares/channel.pyx":844
 *             cyares_release_buffer(&view)
 *             fut = self.__create_future(callback)
 *             ares_getnameinfo(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 <sockaddr*>&sa6,
*/
    ares_getnameinfo(__pyx_v_self->channel, ((struct sockaddr *)(&__pyx_v_sa6)), (sizeof(__pyx_v_sa6)), __pyx_v_flags, __pyx_f_6cyares_9callbacks___callback_nameinfo, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":832
 *                 <void*>fut
 *             )
 *         elif len(address) == 4:             # <<<<<<<<<<<<<<
 *             (ip, port, flowinfo, scope_id) = address
 *             cyares_get_buffer(ip, &view)
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":853
 *             )
 *         else:
 *             raise ValueError("Invalid address argument")             # <<<<<<<<<<<<<<
 * 
 *         return fut
*/
  /*else*/ {
    __pyx_t_6 = NULL;
    __Pyx_INCREF(__pyx_builtin_ValueError);
    __pyx_t_4 = __pyx_builtin_ValueError; 
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_Invalid_address_argument};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 853, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":855
 *             raise ValueError("Invalid address argument")
 * 
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     def gethostbyname(
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":804
 *         return fut
 * 
 *     def getnameinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         tuple address,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyares.channel.Channel.getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XDECREF(__pyx_v_ip);
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_XDECREF(__pyx_v_flowinfo);
  __Pyx_XDECREF(__pyx_v_scope_id);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":857
 *         return fut
 * 
 *     def gethostbyname(             # <<<<<<<<<<<<<<
 *         self,
 *         object name,
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_33gethostbyname(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_33gethostbyname = {"gethostbyname", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_33gethostbyname, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_33gethostbyname(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  int __pyx_v_family;
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gethostbyname (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name_2,&__pyx_mstate_global->__pyx_n_u_family,&__pyx_mstate_global->__pyx_n_u_callback,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 857, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 857, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 857, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 857, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "gethostbyname", 0) < 0) __PYX_ERR(0, 857, __pyx_L3_error)

      /* "cyares/channel.pyx":861
 *         object name,
 *         int family,
 *         object callback = None             # <<<<<<<<<<<<<<
 *     ):
 * 
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("gethostbyname", 0, 2, 3, i); __PYX_ERR(0, 857, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 857, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 857, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 857, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_name = values[0];
    __pyx_v_family = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 860, __pyx_L3_error)
    __pyx_v_callback = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("gethostbyname", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 857, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.gethostbyname", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_32gethostbyname(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_name, __pyx_v_family, __pyx_v_callback);

  /* "cyares/channel.pyx":857
 *         return fut
 * 
 *     def gethostbyname(             # <<<<<<<<<<<<<<
 *         self,
 *         object name,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_32gethostbyname(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_v_family, PyObject *__pyx_v_callback) {
  Py_buffer __pyx_v_view;
  PyObject *__pyx_v_fut = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gethostbyname", 0);

  /* "cyares/channel.pyx":865
 * 
 *         cdef Py_buffer view
 *         cdef object fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 * 
 *         cyares_get_domain_name_buffer(name, &view)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":867
 *         cdef object fut = self.__create_future(callback)
 * 
 *         cyares_get_domain_name_buffer(name, &view)             # <<<<<<<<<<<<<<
 *         ares_gethostbyname(
 *             self.channel,
*/
  __pyx_t_2 = __pyx_f_6cyares_7channel_cyares_get_domain_name_buffer(__pyx_v_name, (&__pyx_v_view)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 867, __pyx_L1_error)

  /* "cyares/channel.pyx":868
 * 
 *         cyares_get_domain_name_buffer(name, &view)
 *         ares_gethostbyname(             # <<<<<<<<<<<<<<
 *             self.channel,
 *             <char*>view.buf,
*/
  ares_gethostbyname(__pyx_v_self->channel, ((char *)__pyx_v_view.buf), __pyx_v_family, __pyx_f_6cyares_9callbacks___callback_gethostbyname, ((void *)__pyx_v_fut));

  /* "cyares/channel.pyx":875
 *             <void*>fut
 *         )
 *         cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         return fut
 * 
*/
  cyares_release_buffer((&__pyx_v_view));

  /* "cyares/channel.pyx":876
 *         )
 *         cyares_release_buffer(&view)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":857
 *         return fut
 * 
 *     def gethostbyname(             # <<<<<<<<<<<<<<
 *         self,
 *         object name,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.gethostbyname", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":879
 * 
 * 
 *     def gethostbyaddr(self, object addr, object callback = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_35gethostbyaddr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_35gethostbyaddr = {"gethostbyaddr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_35gethostbyaddr, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_35gethostbyaddr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_addr = 0;
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gethostbyaddr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,&__pyx_mstate_global->__pyx_n_u_callback,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 879, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 879, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 879, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "gethostbyaddr", 0) < 0) __PYX_ERR(0, 879, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("gethostbyaddr", 0, 1, 2, i); __PYX_ERR(0, 879, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 879, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 879, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_addr = values[0];
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("gethostbyaddr", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 879, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.gethostbyaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_34gethostbyaddr(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_addr, __pyx_v_callback);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_34gethostbyaddr(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_callback) {
  struct in_addr __pyx_v_addr4;
  struct ares_in6_addr __pyx_v_addr6;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_v_fut = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gethostbyaddr", 0);

  /* "cyares/channel.pyx":885
 *         cdef object fut
 * 
 *         cyares_get_buffer(addr, &view)             # <<<<<<<<<<<<<<
 * 
 *         if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
*/
  __pyx_t_1 = cyares_get_buffer(__pyx_v_addr, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 885, __pyx_L1_error)

  /* "cyares/channel.pyx":887
 *         cyares_get_buffer(addr, &view)
 * 
 *         if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(
*/
  __pyx_t_2 = (ares_inet_pton(AF_INET, ((char *)__pyx_v_view.buf), (&__pyx_v_addr4)) != 0);
  if (__pyx_t_2) {

    /* "cyares/channel.pyx":888
 * 
 *         if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *             fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *             ares_gethostbyaddr(
 *                 self.channel,
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_fut = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyares/channel.pyx":889
 *         if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 &addr4,
*/
    ares_gethostbyaddr(__pyx_v_self->channel, (&__pyx_v_addr4), (sizeof(__pyx_v_addr4)), AF_INET, __pyx_f_6cyares_9callbacks___callback_gethostbyaddr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":887
 *         cyares_get_buffer(addr, &view)
 * 
 *         if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":899
 * 
 * 
 *         elif ares_inet_pton(AF_INET6, <char*>view.buf, &addr6):             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(
*/
  __pyx_t_2 = (ares_inet_pton(AF_INET6, ((char *)__pyx_v_view.buf), (&__pyx_v_addr6)) != 0);
  if (likely(__pyx_t_2)) {

    /* "cyares/channel.pyx":900
 * 
 *         elif ares_inet_pton(AF_INET6, <char*>view.buf, &addr6):
 *             fut = self.__create_future(callback)             # <<<<<<<<<<<<<<
 *             ares_gethostbyaddr(
 *                 self.channel,
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__create_future(__pyx_v_self, __pyx_v_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_fut = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyares/channel.pyx":901
 *         elif ares_inet_pton(AF_INET6, <char*>view.buf, &addr6):
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(             # <<<<<<<<<<<<<<
 *                 self.channel,
 *                 &addr6,
*/
    ares_gethostbyaddr(__pyx_v_self->channel, (&__pyx_v_addr6), (sizeof(__pyx_v_addr6)), AF_INET6, __pyx_f_6cyares_9callbacks___callback_gethostbyaddr, ((void *)__pyx_v_fut));

    /* "cyares/channel.pyx":899
 * 
 * 
 *         elif ares_inet_pton(AF_INET6, <char*>view.buf, &addr6):             # <<<<<<<<<<<<<<
 *             fut = self.__create_future(callback)
 *             ares_gethostbyaddr(
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":911
 * 
 *         else:
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid IP address")
 * 
*/
  /*else*/ {
    cyares_release_buffer((&__pyx_v_view));

    /* "cyares/channel.pyx":912
 *         else:
 *             cyares_release_buffer(&view)
 *             raise ValueError("invalid IP address")             # <<<<<<<<<<<<<<
 * 
 *         cyares_release_buffer(&view)
*/
    __pyx_t_4 = NULL;
    __Pyx_INCREF(__pyx_builtin_ValueError);
    __pyx_t_5 = __pyx_builtin_ValueError; 
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_invalid_IP_address};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 912, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":914
 *             raise ValueError("invalid IP address")
 * 
 *         cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 *         return fut
 * 
*/
  cyares_release_buffer((&__pyx_v_view));

  /* "cyares/channel.pyx":915
 * 
 *         cyares_release_buffer(&view)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "cyares/channel.pyx":879
 * 
 * 
 *     def gethostbyaddr(self, object addr, object callback = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyares.channel.Channel.gethostbyaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":920
 * 
 * 
 *     def set_local_dev(self, object dev):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         cyares_get_buffer(dev, &view)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_37set_local_dev(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_37set_local_dev = {"set_local_dev", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_37set_local_dev, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_37set_local_dev(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_dev = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_local_dev (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dev,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 920, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 920, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_local_dev", 0) < 0) __PYX_ERR(0, 920, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_local_dev", 1, 1, 1, i); __PYX_ERR(0, 920, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 920, __pyx_L3_error)
    }
    __pyx_v_dev = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_local_dev", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 920, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.set_local_dev", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_36set_local_dev(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_dev);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_36set_local_dev(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_dev) {
  Py_buffer __pyx_v_view;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_local_dev", 0);

  /* "cyares/channel.pyx":922
 *     def set_local_dev(self, object dev):
 *         cdef Py_buffer view
 *         cyares_get_buffer(dev, &view)             # <<<<<<<<<<<<<<
 *         ares_set_local_dev(self.channel, <char*>view.buf)
 *         cyares_release_buffer(&view)
*/
  __pyx_t_1 = cyares_get_buffer(__pyx_v_dev, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 922, __pyx_L1_error)

  /* "cyares/channel.pyx":923
 *         cdef Py_buffer view
 *         cyares_get_buffer(dev, &view)
 *         ares_set_local_dev(self.channel, <char*>view.buf)             # <<<<<<<<<<<<<<
 *         cyares_release_buffer(&view)
 * 
*/
  ares_set_local_dev(__pyx_v_self->channel, ((char *)__pyx_v_view.buf));

  /* "cyares/channel.pyx":924
 *         cyares_get_buffer(dev, &view)
 *         ares_set_local_dev(self.channel, <char*>view.buf)
 *         cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *     def set_local_ip(self, object ip):
*/
  cyares_release_buffer((&__pyx_v_view));

  /* "cyares/channel.pyx":920
 * 
 * 
 *     def set_local_dev(self, object dev):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         cyares_get_buffer(dev, &view)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyares.channel.Channel.set_local_dev", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":926
 *         cyares_release_buffer(&view)
 * 
 *     def set_local_ip(self, object ip):             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_39set_local_ip(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_39set_local_ip = {"set_local_ip", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_39set_local_ip, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_39set_local_ip(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_ip = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_local_ip (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ip,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 926, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 926, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_local_ip", 0) < 0) __PYX_ERR(0, 926, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_local_ip", 1, 1, 1, i); __PYX_ERR(0, 926, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 926, __pyx_L3_error)
    }
    __pyx_v_ip = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_local_ip", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 926, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.set_local_ip", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_38set_local_ip(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_ip);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_38set_local_ip(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_ip) {
  struct in_addr __pyx_v_addr4;
  struct ares_in6_addr __pyx_v_addr6;
  Py_buffer __pyx_v_view;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_local_ip", 0);

  /* "cyares/channel.pyx":930
 *         cdef ares_in6_addr addr6
 *         cdef Py_buffer view
 *         cyares_get_buffer(ip, &view)             # <<<<<<<<<<<<<<
 *         try:
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
*/
  __pyx_t_1 = cyares_get_buffer(__pyx_v_ip, (&__pyx_v_view)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 930, __pyx_L1_error)

  /* "cyares/channel.pyx":931
 *         cdef Py_buffer view
 *         cyares_get_buffer(ip, &view)
 *         try:             # <<<<<<<<<<<<<<
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
*/
  /*try:*/ {

    /* "cyares/channel.pyx":932
 *         cyares_get_buffer(ip, &view)
 *         try:
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):             # <<<<<<<<<<<<<<
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):
*/
    __pyx_t_2 = (ares_inet_pton(AF_INET, ((char *)__pyx_v_view.buf), (&__pyx_v_addr4)) != 0);
    if (__pyx_t_2) {

      /* "cyares/channel.pyx":933
 *         try:
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))             # <<<<<<<<<<<<<<
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):
 *                 ares_set_local_ip6(self.channel, <unsigned char*>view.buf)
*/
      ares_set_local_ip4(__pyx_v_self->channel, ((unsigned int)cyares_htonl(__pyx_v_addr4.s_addr)));

      /* "cyares/channel.pyx":932
 *         cyares_get_buffer(ip, &view)
 *         try:
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):             # <<<<<<<<<<<<<<
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):
*/
      goto __pyx_L6;
    }

    /* "cyares/channel.pyx":934
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):             # <<<<<<<<<<<<<<
 *                 ares_set_local_ip6(self.channel, <unsigned char*>view.buf)
 *             else:
*/
    __pyx_t_2 = (ares_inet_pton(AF_INET, ((char *)__pyx_v_view.buf), (&__pyx_v_addr6)) != 0);
    if (likely(__pyx_t_2)) {

      /* "cyares/channel.pyx":935
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):
 *                 ares_set_local_ip6(self.channel, <unsigned char*>view.buf)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError("invalid IP address")
*/
      ares_set_local_ip6(__pyx_v_self->channel, ((unsigned char *)__pyx_v_view.buf));

      /* "cyares/channel.pyx":934
 *             if ares_inet_pton(AF_INET, <char*>view.buf, &addr4):
 *                 ares_set_local_ip4(self.channel, <unsigned int>cyares_htonl(addr4.s_addr))
 *             elif ares_inet_pton(AF_INET, <char*>view.buf, &addr6):             # <<<<<<<<<<<<<<
 *                 ares_set_local_ip6(self.channel, <unsigned char*>view.buf)
 *             else:
*/
      goto __pyx_L6;
    }

    /* "cyares/channel.pyx":937
 *                 ares_set_local_ip6(self.channel, <unsigned char*>view.buf)
 *             else:
 *                 raise ValueError("invalid IP address")             # <<<<<<<<<<<<<<
 *         finally:
 *             cyares_release_buffer(&view)
*/
    /*else*/ {
      __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_builtin_ValueError);
      __pyx_t_5 = __pyx_builtin_ValueError; 
      __pyx_t_6 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_invalid_IP_address};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 937, __pyx_L4_error)
    }
    __pyx_L6:;
  }

  /* "cyares/channel.pyx":939
 *                 raise ValueError("invalid IP address")
 *         finally:
 *             cyares_release_buffer(&view)             # <<<<<<<<<<<<<<
 * 
 *     def getsock(self):
*/
  /*finally:*/ {
    /*normal exit:*/{
      cyares_release_buffer((&__pyx_v_view));
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ( unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_1 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        cyares_release_buffer((&__pyx_v_view));
      }
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "cyares/channel.pyx":926
 *         cyares_release_buffer(&view)
 * 
 *     def set_local_ip(self, object ip):             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cyares.channel.Channel.set_local_ip", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":941
 *             cyares_release_buffer(&view)
 * 
 *     def getsock(self):             # <<<<<<<<<<<<<<
 *         cdef list rfds = []
 *         cdef list wfds = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_41getsock(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_41getsock = {"getsock", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_41getsock, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_41getsock(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsock (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("getsock", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("getsock", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_40getsock(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_40getsock(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_v_rfds = 0;
  PyObject *__pyx_v_wfds = 0;
  ares_socket_t __pyx_v_socks[16];
  int __pyx_v_bitmask;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsock", 0);

  /* "cyares/channel.pyx":942
 * 
 *     def getsock(self):
 *         cdef list rfds = []             # <<<<<<<<<<<<<<
 *         cdef list wfds = []
 *         cdef ares_socket_t[16] socks
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rfds = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":943
 *     def getsock(self):
 *         cdef list rfds = []
 *         cdef list wfds = []             # <<<<<<<<<<<<<<
 *         cdef ares_socket_t[16] socks
 *         # NOTE: ARES_GETSOCK_MAXNUM SHOULD BE 16 reason for predefining
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wfds = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyares/channel.pyx":947
 *         # NOTE: ARES_GETSOCK_MAXNUM SHOULD BE 16 reason for predefining
 *         # above was so that cython wouldn't screw with me.
 *         cdef int bitmask = ares_getsock(self.channel, socks, ARES_GETSOCK_MAXNUM)             # <<<<<<<<<<<<<<
 *         cdef int i
 * 
*/
  __pyx_v_bitmask = ares_getsock(__pyx_v_self->channel, __pyx_v_socks, ARES_GETSOCK_MAXNUM);

  /* "cyares/channel.pyx":950
 *         cdef int i
 * 
 *         for i in range(ARES_GETSOCK_MAXNUM):             # <<<<<<<<<<<<<<
 *             if ARES_GETSOCK_READABLE(bitmask, i):
 *                 rfds.append(<object>socks[i])
*/
  __pyx_t_2 = ARES_GETSOCK_MAXNUM;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "cyares/channel.pyx":951
 * 
 *         for i in range(ARES_GETSOCK_MAXNUM):
 *             if ARES_GETSOCK_READABLE(bitmask, i):             # <<<<<<<<<<<<<<
 *                 rfds.append(<object>socks[i])
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):
*/
    __pyx_t_5 = (ARES_GETSOCK_READABLE(__pyx_v_bitmask, __pyx_v_i) != 0);
    if (__pyx_t_5) {

      /* "cyares/channel.pyx":952
 *         for i in range(ARES_GETSOCK_MAXNUM):
 *             if ARES_GETSOCK_READABLE(bitmask, i):
 *                 rfds.append(<object>socks[i])             # <<<<<<<<<<<<<<
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):
 *                 wfds.append(socks[i])
*/
      __pyx_t_1 = __Pyx_PyLong_From_ares_socket_t((__pyx_v_socks[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_rfds, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyares/channel.pyx":951
 * 
 *         for i in range(ARES_GETSOCK_MAXNUM):
 *             if ARES_GETSOCK_READABLE(bitmask, i):             # <<<<<<<<<<<<<<
 *                 rfds.append(<object>socks[i])
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):
*/
    }

    /* "cyares/channel.pyx":953
 *             if ARES_GETSOCK_READABLE(bitmask, i):
 *                 rfds.append(<object>socks[i])
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):             # <<<<<<<<<<<<<<
 *                 wfds.append(socks[i])
 * 
*/
    __pyx_t_5 = (ARES_GETSOCK_WRITABLE(__pyx_v_bitmask, __pyx_v_i) != 0);
    if (__pyx_t_5) {

      /* "cyares/channel.pyx":954
 *                 rfds.append(<object>socks[i])
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):
 *                 wfds.append(socks[i])             # <<<<<<<<<<<<<<
 * 
 *         return rfds, wfds
*/
      __pyx_t_1 = __Pyx_PyLong_From_ares_socket_t((__pyx_v_socks[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_wfds, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 954, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyares/channel.pyx":953
 *             if ARES_GETSOCK_READABLE(bitmask, i):
 *                 rfds.append(<object>socks[i])
 *             if ARES_GETSOCK_WRITABLE(bitmask, i):             # <<<<<<<<<<<<<<
 *                 wfds.append(socks[i])
 * 
*/
    }
  }

  /* "cyares/channel.pyx":956
 *                 wfds.append(socks[i])
 * 
 *         return rfds, wfds             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_rfds);
  __Pyx_GIVEREF(__pyx_v_rfds);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_rfds) != (0)) __PYX_ERR(0, 956, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_wfds);
  __Pyx_GIVEREF(__pyx_v_wfds);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_wfds) != (0)) __PYX_ERR(0, 956, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":941
 *             cyares_release_buffer(&view)
 * 
 *     def getsock(self):             # <<<<<<<<<<<<<<
 *         cdef list rfds = []
 *         cdef list wfds = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.getsock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rfds);
  __Pyx_XDECREF(__pyx_v_wfds);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":959
 * 
 * 
 *     cdef ares_status_t __wait(self, int timeout_ms):             # <<<<<<<<<<<<<<
 *         return ares_queue_wait_empty(self.channel, timeout_ms)
 * 
*/

static ares_status_t __pyx_f_6cyares_7channel_7Channel__Channel__wait(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, int __pyx_v_timeout_ms) {
  ares_status_t __pyx_r;

  /* "cyares/channel.pyx":960
 * 
 *     cdef ares_status_t __wait(self, int timeout_ms):
 *         return ares_queue_wait_empty(self.channel, timeout_ms)             # <<<<<<<<<<<<<<
 * 
 *     def wait(self, object timeout = None):
*/
  __pyx_r = ares_queue_wait_empty(__pyx_v_self->channel, __pyx_v_timeout_ms);
  goto __pyx_L0;

  /* "cyares/channel.pyx":959
 * 
 * 
 *     cdef ares_status_t __wait(self, int timeout_ms):             # <<<<<<<<<<<<<<
 *         return ares_queue_wait_empty(self.channel, timeout_ms)
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cyares/channel.pyx":962
 *         return ares_queue_wait_empty(self.channel, timeout_ms)
 * 
 *     def wait(self, object timeout = None):             # <<<<<<<<<<<<<<
 *         cdef ares_status_t status
 *         cdef int ms
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_43wait(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_43wait = {"wait", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_43wait, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_43wait(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wait (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 962, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 962, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wait", 0) < 0) __PYX_ERR(0, 962, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 962, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_timeout = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wait", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 962, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.wait", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_42wait(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_42wait(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, PyObject *__pyx_v_timeout) {
  ares_status_t __pyx_v_status;
  int __pyx_v_ms;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  ares_status_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wait", 0);

  /* "cyares/channel.pyx":966
 *         cdef int ms
 * 
 *         if timeout is None:             # <<<<<<<<<<<<<<
 *             status = self.__wait(-1)
 *         else:
*/
  __pyx_t_1 = (__pyx_v_timeout == Py_None);
  if (__pyx_t_1) {

    /* "cyares/channel.pyx":967
 * 
 *         if timeout is None:
 *             status = self.__wait(-1)             # <<<<<<<<<<<<<<
 *         else:
 *             ms = cyares_seconds_to_milliseconds(timeout)
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__wait(__pyx_v_self, -1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 967, __pyx_L1_error)
    __pyx_v_status = __pyx_t_2;

    /* "cyares/channel.pyx":966
 *         cdef int ms
 * 
 *         if timeout is None:             # <<<<<<<<<<<<<<
 *             status = self.__wait(-1)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "cyares/channel.pyx":969
 *             status = self.__wait(-1)
 *         else:
 *             ms = cyares_seconds_to_milliseconds(timeout)             # <<<<<<<<<<<<<<
 *             if ms < 0:
 *                 raise
*/
  /*else*/ {
    __pyx_t_3 = __pyx_f_6cyares_7channel_cyares_seconds_to_milliseconds(__pyx_v_timeout); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 969, __pyx_L1_error)
    __pyx_v_ms = __pyx_t_3;

    /* "cyares/channel.pyx":970
 *         else:
 *             ms = cyares_seconds_to_milliseconds(timeout)
 *             if ms < 0:             # <<<<<<<<<<<<<<
 *                 raise
 * 
*/
    __pyx_t_1 = (__pyx_v_ms < 0);
    if (unlikely(__pyx_t_1)) {

      /* "cyares/channel.pyx":971
 *             ms = cyares_seconds_to_milliseconds(timeout)
 *             if ms < 0:
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *             status = self.__wait(ms)
*/
      __Pyx_ReraiseException();
      __PYX_ERR(0, 971, __pyx_L1_error)

      /* "cyares/channel.pyx":970
 *         else:
 *             ms = cyares_seconds_to_milliseconds(timeout)
 *             if ms < 0:             # <<<<<<<<<<<<<<
 *                 raise
 * 
*/
    }

    /* "cyares/channel.pyx":973
 *                 raise
 * 
 *             status = self.__wait(ms)             # <<<<<<<<<<<<<<
 * 
 *         if status == ARES_SUCCESS:
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_6cyares_7channel_Channel *)__pyx_v_self->__pyx_vtab)->_Channel__wait(__pyx_v_self, __pyx_v_ms); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L1_error)
    __pyx_v_status = __pyx_t_2;
  }
  __pyx_L3:;

  /* "cyares/channel.pyx":975
 *             status = self.__wait(ms)
 * 
 *         if status == ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             return True
 *         elif status == ARES_ETIMEOUT:
*/
  switch (__pyx_v_status) {
    case ARES_SUCCESS:

    /* "cyares/channel.pyx":976
 * 
 *         if status == ARES_SUCCESS:
 *             return True             # <<<<<<<<<<<<<<
 *         elif status == ARES_ETIMEOUT:
 *             return False
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "cyares/channel.pyx":975
 *             status = self.__wait(ms)
 * 
 *         if status == ARES_SUCCESS:             # <<<<<<<<<<<<<<
 *             return True
 *         elif status == ARES_ETIMEOUT:
*/
    break;
    case ARES_ETIMEOUT:

    /* "cyares/channel.pyx":978
 *             return True
 *         elif status == ARES_ETIMEOUT:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             raise AresError(status)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "cyares/channel.pyx":977
 *         if status == ARES_SUCCESS:
 *             return True
 *         elif status == ARES_ETIMEOUT:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
*/
    break;
    default:

    /* "cyares/channel.pyx":980
 *             return False
 *         else:
 *             raise AresError(status)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_t_5 = NULL;
    __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError);
    __pyx_t_6 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError); 
    __pyx_t_7 = __Pyx_PyLong_From_ares_status_t(__pyx_v_status); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 980, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_4);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_4), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 980, __pyx_L1_error)
    break;
  }

  /* "cyares/channel.pyx":962
 *         return ares_queue_wait_empty(self.channel, timeout_ms)
 * 
 *     def wait(self, object timeout = None):             # <<<<<<<<<<<<<<
 *         cdef ares_status_t status
 *         cdef int ms
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cyares.channel.Channel.wait", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":982
 *             raise AresError(status)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def running_queries(self):
 *         """obtains a number of actively running queries
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_15running_queries_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6cyares_7channel_7Channel_15running_queries_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_15running_queries___get__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_15running_queries___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "cyares/channel.pyx":986
 *         """obtains a number of actively running queries
 *         NOTE: this property is immutable by the end user"""
 *         return ares_queue_active_queries(self.channel)             # <<<<<<<<<<<<<<
 * 
 *     @running_queries.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(ares_queue_active_queries(__pyx_v_self->channel)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":982
 *             raise AresError(status)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def running_queries(self):
 *         """obtains a number of actively running queries
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.running_queries.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":988
 *         return ares_queue_active_queries(self.channel)
 * 
 *     @running_queries.setter             # <<<<<<<<<<<<<<
 *     def running_queries(self, object ignore):
 *         raise ValueError("running_queries is an immutable property")
*/

/* Python wrapper */
static int __pyx_pw_6cyares_7channel_7Channel_15running_queries_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_ignore); /*proto*/
static int __pyx_pw_6cyares_7channel_7Channel_15running_queries_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_ignore) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_15running_queries_2__set__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), ((PyObject *)__pyx_v_ignore));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6cyares_7channel_7Channel_15running_queries_2__set__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ignore) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "cyares/channel.pyx":990
 *     @running_queries.setter
 *     def running_queries(self, object ignore):
 *         raise ValueError("running_queries is an immutable property")             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__pyx_builtin_ValueError);
  __pyx_t_3 = __pyx_builtin_ValueError; 
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_running_queries_is_an_immutable};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 990, __pyx_L1_error)

  /* "cyares/channel.pyx":988
 *         return ares_queue_active_queries(self.channel)
 * 
 *     @running_queries.setter             # <<<<<<<<<<<<<<
 *     def running_queries(self, object ignore):
 *         raise ValueError("running_queries is an immutable property")
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cyares.channel.Channel.running_queries.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pxd":28
 *         # allow programmers to access this event_thread flag
 *         # if checking before wait(...) is required...
 *         readonly bint event_thread             # <<<<<<<<<<<<<<
 *         SocketHandle socket_handle # if we have one
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_12event_thread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6cyares_7channel_7Channel_12event_thread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_12event_thread___get__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_12event_thread___get__(struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->event_thread); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.Channel.event_thread.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_45__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_45__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_45__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_45__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_44__reduce_cython__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_44__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_mstate_global->__pyx_kp_u_self_channel_self_options_cannot, 0, 0);
  __PYX_ERR(3, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyares.channel.Channel.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_7Channel_47__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6cyares_7channel_7Channel_47__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_47__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6cyares_7channel_7Channel_47__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(3, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(3, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(3, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(3, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(3, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("cyares.channel.Channel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6cyares_7channel_7Channel_46__setstate_cython__(((struct __pyx_obj_6cyares_7channel_Channel *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_7Channel_46__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6cyares_7channel_Channel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_mstate_global->__pyx_kp_u_self_channel_self_options_cannot, 0, 0);
  __PYX_ERR(3, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyares.channel.Channel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyares/channel.pyx":995
 * 
 * 
 * def cyares_threadsafety():             # <<<<<<<<<<<<<<
 *     """
 *     pycares documentation says:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6cyares_7channel_5cyares_threadsafety(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_6cyares_7channel_4cyares_threadsafety, "\n    pycares documentation says:\n    Check if c-ares was compiled with thread safety support.\n\n    :return: True if thread-safe, False otherwise.\n    :rtype: bool\n    ");
static PyMethodDef __pyx_mdef_6cyares_7channel_5cyares_threadsafety = {"cyares_threadsafety", (PyCFunction)__pyx_pw_6cyares_7channel_5cyares_threadsafety, METH_NOARGS, __pyx_doc_6cyares_7channel_4cyares_threadsafety};
static PyObject *__pyx_pw_6cyares_7channel_5cyares_threadsafety(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cyares_threadsafety (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6cyares_7channel_4cyares_threadsafety(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cyares_7channel_4cyares_threadsafety(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cyares_threadsafety", 0);

  /* "cyares/channel.pyx":1003
 *     :rtype: bool
 *     """
 *     return ares_threadsafety() == ARES_TRUE             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((ares_threadsafety() == ARES_TRUE)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyares/channel.pyx":995
 * 
 * 
 * def cyares_threadsafety():             # <<<<<<<<<<<<<<
 *     """
 *     pycares documentation says:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyares.channel.cyares_threadsafety", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_6cyares_7channel_Channel __pyx_vtable_6cyares_7channel_Channel;

static PyObject *__pyx_tp_new_6cyares_7channel_Channel(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6cyares_7channel_Channel *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6cyares_7channel_Channel *)o);
  p->__pyx_vtab = __pyx_vtabptr_6cyares_7channel_Channel;
  p->handles = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_query_lookups = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->socket_handle = ((struct __pyx_obj_6cyares_13socket_handle_SocketHandle *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6cyares_7channel_Channel(PyObject *o) {
  struct __pyx_obj_6cyares_7channel_Channel *p = (struct __pyx_obj_6cyares_7channel_Channel *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6cyares_7channel_Channel) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6cyares_7channel_7Channel_7__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->handles);
  Py_CLEAR(p->_query_lookups);
  Py_CLEAR(p->socket_handle);
  #if CYTHON_USE_TYPE_SLOTS
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6cyares_7channel_Channel(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6cyares_7channel_Channel *p = (struct __pyx_obj_6cyares_7channel_Channel *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->handles) {
    e = (*v)(p->handles, a); if (e) return e;
  }
  if (p->_query_lookups) {
    e = (*v)(p->_query_lookups, a); if (e) return e;
  }
  if (p->socket_handle) {
    e = (*v)(((PyObject *)p->socket_handle), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6cyares_7channel_Channel(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6cyares_7channel_Channel *p = (struct __pyx_obj_6cyares_7channel_Channel *)o;
  tmp = ((PyObject*)p->handles);
  p->handles = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_query_lookups);
  p->_query_lookups = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->socket_handle);
  p->socket_handle = ((struct __pyx_obj_6cyares_13socket_handle_SocketHandle *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6cyares_7channel_7Channel_servers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6cyares_7channel_7Channel_7servers_1__get__(o);
}

static int __pyx_setprop_6cyares_7channel_7Channel_servers(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6cyares_7channel_7Channel_7servers_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6cyares_7channel_7Channel_running_queries(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6cyares_7channel_7Channel_15running_queries_1__get__(o);
}

static int __pyx_setprop_6cyares_7channel_7Channel_running_queries(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6cyares_7channel_7Channel_15running_queries_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6cyares_7channel_7Channel_event_thread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6cyares_7channel_7Channel_12event_thread_1__get__(o);
}

static PyMethodDef __pyx_methods_6cyares_7channel_Channel[] = {
  {"reinit", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_5reinit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__enter__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_9__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__exit__", (PyCFunction)(void(*)(void))(PyCFunctionWithKeywords)__pyx_pw_6cyares_7channel_7Channel_11__exit__, METH_VARARGS|METH_KEYWORDS, 0},
  {"_Channel__remove_future", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_13_Channel__remove_future, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"debug", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_15debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_17query, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"search", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_19search, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"process_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_21process_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"process_read_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_23process_read_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cyares_7channel_7Channel_22process_read_fd},
  {"process_write_fd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_25process_write_fd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cyares_7channel_7Channel_24process_write_fd},
  {"timeout", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_27timeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"getaddrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_29getaddrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"getnameinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_31getnameinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"gethostbyname", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_33gethostbyname, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"gethostbyaddr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_35gethostbyaddr, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_local_dev", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_37set_local_dev, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_local_ip", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_39set_local_ip, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"getsock", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_41getsock, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wait", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_43wait, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_45__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cyares_7channel_7Channel_47__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6cyares_7channel_Channel[] = {
  {"servers", __pyx_getprop_6cyares_7channel_7Channel_servers, __pyx_setprop_6cyares_7channel_7Channel_servers, 0, 0},
  {"running_queries", __pyx_getprop_6cyares_7channel_7Channel_running_queries, __pyx_setprop_6cyares_7channel_7Channel_running_queries, PyDoc_STR("obtains a number of actively running queries \n        NOTE: this property is immutable by the end user"), 0},
  {"event_thread", __pyx_getprop_6cyares_7channel_7Channel_event_thread, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6cyares_7channel_Channel_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6cyares_7channel_Channel},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6cyares_7channel_Channel},
  {Py_tp_clear, (void *)__pyx_tp_clear_6cyares_7channel_Channel},
  {Py_tp_methods, (void *)__pyx_methods_6cyares_7channel_Channel},
  {Py_tp_getset, (void *)__pyx_getsets_6cyares_7channel_Channel},
  {Py_tp_init, (void *)__pyx_pw_6cyares_7channel_7Channel_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6cyares_7channel_Channel},
  {0, 0},
};
static PyType_Spec __pyx_type_6cyares_7channel_Channel_spec = {
  "cyares.channel.Channel",
  sizeof(struct __pyx_obj_6cyares_7channel_Channel),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6cyares_7channel_Channel_slots,
};
#else

static PyTypeObject __pyx_type_6cyares_7channel_Channel = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyares.channel.""Channel", /*tp_name*/
  sizeof(struct __pyx_obj_6cyares_7channel_Channel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6cyares_7channel_Channel, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6cyares_7channel_Channel, /*tp_traverse*/
  __pyx_tp_clear_6cyares_7channel_Channel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6cyares_7channel_Channel, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6cyares_7channel_Channel, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6cyares_7channel_7Channel_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6cyares_7channel_Channel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6cyares_7channel_Future = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_mstate_global->__pyx_n_u_Future, (void *)&__pyx_v_6cyares_7channel_Future, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6cyares_7channel_Channel = &__pyx_vtable_6cyares_7channel_Channel;
  __pyx_vtable_6cyares_7channel_Channel.cancel = (void (*)(struct __pyx_obj_6cyares_7channel_Channel *, int __pyx_skip_dispatch))__pyx_f_6cyares_7channel_7Channel_cancel;
  __pyx_vtable_6cyares_7channel_Channel._malloc = (void *(*)(struct __pyx_obj_6cyares_7channel_Channel *, size_t))__pyx_f_6cyares_7channel_7Channel__malloc;
  __pyx_vtable_6cyares_7channel_Channel._query = (PyObject *(*)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *, PyObject *, int, PyObject *))__pyx_f_6cyares_7channel_7Channel__query;
  __pyx_vtable_6cyares_7channel_Channel._search = (PyObject *(*)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *, PyObject *, int, PyObject *))__pyx_f_6cyares_7channel_7Channel__search;
  __pyx_vtable_6cyares_7channel_Channel._Channel__create_future = (PyObject *(*)(struct __pyx_obj_6cyares_7channel_Channel *, PyObject *))__pyx_f_6cyares_7channel_7Channel__Channel__create_future;
  __pyx_vtable_6cyares_7channel_Channel._Channel__wait = (ares_status_t (*)(struct __pyx_obj_6cyares_7channel_Channel *, int))__pyx_f_6cyares_7channel_7Channel__Channel__wait;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6cyares_7channel_Channel_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6cyares_7channel_Channel)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6cyares_7channel_Channel_spec, __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel = &__pyx_type_6cyares_7channel_Channel;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6cyares_7channel_Channel->tp_dictoffset && __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6cyares_7channel_Channel, __pyx_vtabptr_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Channel, (PyObject *) __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6cyares_7channel_Channel) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cyares.exception"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_6cyares_9exception_AresError = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.exception", "AresError",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_9exception_AresError), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_9exception_AresError),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_9exception_AresError), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_9exception_AresError),
  #else
  sizeof(struct __pyx_obj_6cyares_9exception_AresError), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_9exception_AresError),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_9exception_AresError) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_1_2(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cyares.resulttypes"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_AresResult = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "AresResult",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_AresResult), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_AresResult),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_AresResult), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_AresResult),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_AresResult), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_AresResult),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_AresResult) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_optval_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_optval_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_optval_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_optval_result) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_optval_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_optval_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_optval_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_optval_result)) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_opt_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_opt_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result) __PYX_ERR(1, 58, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_opt_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_opt_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_opt_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_opt_result)) __PYX_ERR(1, 58, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_hinfo_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_hinfo_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result) __PYX_ERR(1, 79, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_hinfo_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_hinfo_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_hinfo_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_hinfo_result)) __PYX_ERR(1, 79, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_sig_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_sig_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result) __PYX_ERR(1, 97, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_sig_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_sig_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_sig_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_sig_result)) __PYX_ERR(1, 97, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_tlsa_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_tlsa_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result) __PYX_ERR(1, 132, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_tlsa_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_tlsa_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_tlsa_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_tlsa_result)) __PYX_ERR(1, 132, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_svcb_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_svcb_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result) __PYX_ERR(1, 158, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_svcb_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_svcb_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_svcb_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_svcb_result)) __PYX_ERR(1, 158, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_https_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_https_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result) __PYX_ERR(1, 180, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_https_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_https_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_https_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_https_result)) __PYX_ERR(1, 180, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_uri_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_uri_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_uri_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_uri_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_uri_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_uri_result)) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_raw_rr_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_raw_rr_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result) __PYX_ERR(1, 221, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_raw_rr_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_raw_rr_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_raw_rr_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_raw_rr_result)) __PYX_ERR(1, 221, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_a_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_a_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_a_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_a_result) __PYX_ERR(1, 232, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_a_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_a_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_a_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_a_result)) __PYX_ERR(1, 232, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_aaaa_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_aaaa_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result) __PYX_ERR(1, 246, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_aaaa_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_aaaa_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_aaaa_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_aaaa_result)) __PYX_ERR(1, 246, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_caa_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_caa_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result) __PYX_ERR(1, 258, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_caa_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_caa_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_caa_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_caa_result)) __PYX_ERR(1, 258, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_cname_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_cname_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result) __PYX_ERR(1, 272, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_cname_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_cname_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_cname_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_cname_result)) __PYX_ERR(1, 272, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_mx_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_mx_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result) __PYX_ERR(1, 283, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_mx_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_mx_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_mx_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_mx_result)) __PYX_ERR(1, 283, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_naptr_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_naptr_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result) __PYX_ERR(1, 296, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_naptr_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_naptr_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_naptr_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_naptr_result)) __PYX_ERR(1, 296, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_ns_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ns_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result) __PYX_ERR(1, 313, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_ns_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ns_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ns_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_ns_result)) __PYX_ERR(1, 313, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_ptr_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_ptr_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result) __PYX_ERR(1, 323, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_ptr_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_ptr_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_ptr_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_ptr_result)) __PYX_ERR(1, 323, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_soa_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_soa_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result) __PYX_ERR(1, 341, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_soa_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_soa_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_soa_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_soa_result)) __PYX_ERR(1, 341, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_srv_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_srv_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result) __PYX_ERR(1, 361, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_srv_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_srv_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_srv_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_srv_result)) __PYX_ERR(1, 361, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_query_txt_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_query_txt_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result) __PYX_ERR(1, 376, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_query_txt_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_query_txt_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_query_txt_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_query_txt_result)) __PYX_ERR(1, 376, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_host_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_host_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_host_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_host_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_host_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_host_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_host_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_host_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_host_result) __PYX_ERR(1, 396, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_host_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_host_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_host_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_host_result)) __PYX_ERR(1, 396, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_nameinfo_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_nameinfo_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result) __PYX_ERR(1, 406, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_nameinfo_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_nameinfo_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_nameinfo_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_nameinfo_result)) __PYX_ERR(1, 406, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_addrinfo_node_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_node_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result) __PYX_ERR(1, 415, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_node_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_node_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_node_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_node_result)) __PYX_ERR(1, 415, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_addrinfo_cname_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_cname_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result) __PYX_ERR(1, 427, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_cname_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_cname_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_cname_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_cname_result)) __PYX_ERR(1, 427, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.resulttypes", "ares_addrinfo_result",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result),
  #else
  sizeof(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_11resulttypes_ares_addrinfo_result),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result) __PYX_ERR(1, 437, __pyx_L1_error)
  __pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_result = (struct __pyx_vtabstruct_6cyares_11resulttypes_ares_addrinfo_result*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_11resulttypes_ares_addrinfo_result); if (unlikely(!__pyx_vtabptr_6cyares_11resulttypes_ares_addrinfo_result)) __PYX_ERR(1, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cyares.socket_handle"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_6cyares_13socket_handle_SocketHandle = __Pyx_ImportType_3_1_2(__pyx_t_1, "cyares.socket_handle", "SocketHandle",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_6cyares_13socket_handle_SocketHandle), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_13socket_handle_SocketHandle),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_6cyares_13socket_handle_SocketHandle), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_13socket_handle_SocketHandle),
  #else
  sizeof(struct __pyx_obj_6cyares_13socket_handle_SocketHandle), __PYX_GET_STRUCT_ALIGNMENT_3_1_2(struct __pyx_obj_6cyares_13socket_handle_SocketHandle),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_2); if (!__pyx_mstate->__pyx_ptype_6cyares_13socket_handle_SocketHandle) __PYX_ERR(6, 4, __pyx_L1_error)
  __pyx_vtabptr_6cyares_13socket_handle_SocketHandle = (struct __pyx_vtabstruct_6cyares_13socket_handle_SocketHandle*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_6cyares_13socket_handle_SocketHandle); if (unlikely(!__pyx_vtabptr_6cyares_13socket_handle_SocketHandle)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cyares.socket_handle"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__socket_state_callback", (void (**)(void))&__pyx_f_6cyares_13socket_handle___socket_state_callback, "void (void *, ares_socket_t, int, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cyares.callbacks"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_a", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_a, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_aaaa", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_aaaa, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_caa", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_caa, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_cname", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_cname, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_mx", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_mx, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_naptr", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_naptr, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_ns", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_ns, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_ptr", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_ptr, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_soa", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_soa, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_srv", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_srv, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_query_on_txt", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_query_on_txt, "void (void *, int, int, unsigned char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_getaddrinfo", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_getaddrinfo, "void (void *, int, int, struct ares_addrinfo *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_gethostbyname", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_gethostbyname, "void (void *, int, int, struct hostent *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_nameinfo", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_nameinfo, "void (void *, int, int, char *, char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_gethostbyaddr", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_gethostbyaddr, "void (void *, int, int, struct hostent *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_3_1_2(__pyx_t_1, "__callback_dns_rec__any", (void (**)(void))&__pyx_f_6cyares_9callbacks___callback_dns_rec__any, "void (void *, ares_status_t, size_t, ares_dns_record_t const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_channel(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_channel},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, Py_MOD_GIL_USED},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "channel",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_channel(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_channel(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x03090000
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_channel(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'channel' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "channel" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif

__Pyx_RefNannySetupContext("PyInit_channel", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if 0 || defined(__Pyx_CyFunction_USED) || defined(__Pyx_FusedFunction_USED) || defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED) || defined(__Pyx_AsyncGen_USED)
  if (__pyx_CommonTypesMetaclass_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  if (__pyx_module_is_main_cyares__channel) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cyares.channel")) {
      if (unlikely((PyDict_SetItemString(modules, "cyares.channel", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_variable_export_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "cyares/handles.pxi":1
 * from concurrent.futures import Future as _Future             # <<<<<<<<<<<<<<
 * 
 * # XXX: Planning to move this all to Cython in the future but for now were stuck with this...
*/
  __pyx_t_2 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_Future); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_concurrent_futures, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Future); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Future_2, __pyx_t_2) < 0) __PYX_ERR(7, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/handles.pxi":5
 * # XXX: Planning to move this all to Cython in the future but for now were stuck with this...
 * 
 * cdef Future = _Future             # <<<<<<<<<<<<<<
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Future_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_6cyares_7channel_Future);
  __Pyx_DECREF_SET(__pyx_v_6cyares_7channel_Future, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cyares/channel.pyx":23
 * 
 * 
 * cdef bint HAS_IDNA2008 = False             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "Python.h":
*/
  __pyx_v_6cyares_7channel_HAS_IDNA2008 = 0;

  /* "cyares/channel.pyx":29
 * 
 * 
 * try:             # <<<<<<<<<<<<<<
 *     # if someone would like to optimize or write a cython / c module that does this,
 *     # be my guest because I'll happily take it because otherwise we need to find a
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "cyares/channel.pyx":35
 *     #  - Vizonex
 * 
 *     from idna import decode as idna_decode  # type: ignore (Silly cyright)             # <<<<<<<<<<<<<<
 *     HAS_IDNA2008 = True
 * except ModuleNotFoundError:
*/
      __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_idna, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_idna_decode, __pyx_t_3) < 0) __PYX_ERR(0, 35, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "cyares/channel.pyx":36
 * 
 *     from idna import decode as idna_decode  # type: ignore (Silly cyright)
 *     HAS_IDNA2008 = True             # <<<<<<<<<<<<<<
 * except ModuleNotFoundError:
 *     HAS_IDNA2008 = False
*/
      __pyx_v_6cyares_7channel_HAS_IDNA2008 = 1;

      /* "cyares/channel.pyx":29
 * 
 * 
 * try:             # <<<<<<<<<<<<<<
 *     # if someone would like to optimize or write a cython / c module that does this,
 *     # be my guest because I'll happily take it because otherwise we need to find a
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyares/channel.pyx":37
 *     from idna import decode as idna_decode  # type: ignore (Silly cyright)
 *     HAS_IDNA2008 = True
 * except ModuleNotFoundError:             # <<<<<<<<<<<<<<
 *     HAS_IDNA2008 = False
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ModuleNotFoundError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("cyares.channel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 37, __pyx_L4_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "cyares/channel.pyx":38
 *     HAS_IDNA2008 = True
 * except ModuleNotFoundError:
 *     HAS_IDNA2008 = False             # <<<<<<<<<<<<<<
 * 
 * # PYTEST PURPOSES ONLY, USING THESE IS DISCOURAGED!!!
*/
      __pyx_v_6cyares_7channel_HAS_IDNA2008 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;

    /* "cyares/channel.pyx":29
 * 
 * 
 * try:             # <<<<<<<<<<<<<<
 *     # if someone would like to optimize or write a cython / c module that does this,
 *     # be my guest because I'll happily take it because otherwise we need to find a
*/
    __pyx_L4_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_4, __pyx_t_5);
    __pyx_L7_try_end:;
  }

  /* "cyares/channel.pyx":41
 * 
 * # PYTEST PURPOSES ONLY, USING THESE IS DISCOURAGED!!!
 * def __htons(unsigned short s):             # <<<<<<<<<<<<<<
 *     return cyares_htons(s)
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_1__htons, 0, __pyx_mstate_global->__pyx_n_u_htons, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_htons, __pyx_t_7) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":44
 *     return cyares_htons(s)
 * 
 * def __htonl(unsigned long l):             # <<<<<<<<<<<<<<
 *     return cyares_htonl(l)
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_3__htonl, 0, __pyx_mstate_global->__pyx_n_u_htonl, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_htonl, __pyx_t_7) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":111
 * 
 * 
 * CYARES_SOCKET_BAD = ARES_SOCKET_BAD             # <<<<<<<<<<<<<<
 * 
 * # From pycares
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(ARES_SOCKET_BAD); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CYARES_SOCKET_BAD, __pyx_t_7) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":116
 * 
 * # Query types
 * QUERY_TYPE_A = T_A             # <<<<<<<<<<<<<<
 * QUERY_TYPE_AAAA = T_AAAA
 * QUERY_TYPE_ANY = T_ANY
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_A); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_A, __pyx_t_7) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":117
 * # Query types
 * QUERY_TYPE_A = T_A
 * QUERY_TYPE_AAAA = T_AAAA             # <<<<<<<<<<<<<<
 * QUERY_TYPE_ANY = T_ANY
 * QUERY_TYPE_CAA = T_CAA
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_AAAA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_AAAA, __pyx_t_7) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":118
 * QUERY_TYPE_A = T_A
 * QUERY_TYPE_AAAA = T_AAAA
 * QUERY_TYPE_ANY = T_ANY             # <<<<<<<<<<<<<<
 * QUERY_TYPE_CAA = T_CAA
 * QUERY_TYPE_CNAME = T_CNAME
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_ANY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_ANY, __pyx_t_7) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":119
 * QUERY_TYPE_AAAA = T_AAAA
 * QUERY_TYPE_ANY = T_ANY
 * QUERY_TYPE_CAA = T_CAA             # <<<<<<<<<<<<<<
 * QUERY_TYPE_CNAME = T_CNAME
 * QUERY_TYPE_MX = T_MX
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_CAA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_CAA, __pyx_t_7) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":120
 * QUERY_TYPE_ANY = T_ANY
 * QUERY_TYPE_CAA = T_CAA
 * QUERY_TYPE_CNAME = T_CNAME             # <<<<<<<<<<<<<<
 * QUERY_TYPE_MX = T_MX
 * QUERY_TYPE_NAPTR = T_NAPTR
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_CNAME); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_CNAME, __pyx_t_7) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":121
 * QUERY_TYPE_CAA = T_CAA
 * QUERY_TYPE_CNAME = T_CNAME
 * QUERY_TYPE_MX = T_MX             # <<<<<<<<<<<<<<
 * QUERY_TYPE_NAPTR = T_NAPTR
 * QUERY_TYPE_NS = T_NS
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_MX); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_MX, __pyx_t_7) < 0) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":122
 * QUERY_TYPE_CNAME = T_CNAME
 * QUERY_TYPE_MX = T_MX
 * QUERY_TYPE_NAPTR = T_NAPTR             # <<<<<<<<<<<<<<
 * QUERY_TYPE_NS = T_NS
 * QUERY_TYPE_PTR = T_PTR
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_NAPTR); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_NAPTR, __pyx_t_7) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":123
 * QUERY_TYPE_MX = T_MX
 * QUERY_TYPE_NAPTR = T_NAPTR
 * QUERY_TYPE_NS = T_NS             # <<<<<<<<<<<<<<
 * QUERY_TYPE_PTR = T_PTR
 * QUERY_TYPE_SOA = T_SOA
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_NS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_NS, __pyx_t_7) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":124
 * QUERY_TYPE_NAPTR = T_NAPTR
 * QUERY_TYPE_NS = T_NS
 * QUERY_TYPE_PTR = T_PTR             # <<<<<<<<<<<<<<
 * QUERY_TYPE_SOA = T_SOA
 * QUERY_TYPE_SRV = T_SRV
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_PTR); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_PTR, __pyx_t_7) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":125
 * QUERY_TYPE_NS = T_NS
 * QUERY_TYPE_PTR = T_PTR
 * QUERY_TYPE_SOA = T_SOA             # <<<<<<<<<<<<<<
 * QUERY_TYPE_SRV = T_SRV
 * QUERY_TYPE_TXT = T_TXT
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_SOA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_SOA, __pyx_t_7) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":126
 * QUERY_TYPE_PTR = T_PTR
 * QUERY_TYPE_SOA = T_SOA
 * QUERY_TYPE_SRV = T_SRV             # <<<<<<<<<<<<<<
 * QUERY_TYPE_TXT = T_TXT
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_SRV); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_SRV, __pyx_t_7) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":127
 * QUERY_TYPE_SOA = T_SOA
 * QUERY_TYPE_SRV = T_SRV
 * QUERY_TYPE_TXT = T_TXT             # <<<<<<<<<<<<<<
 * 
 * # Query classes
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(T_TXT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_TYPE_TXT, __pyx_t_7) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":130
 * 
 * # Query classes
 * QUERY_CLASS_IN = C_IN             # <<<<<<<<<<<<<<
 * QUERY_CLASS_CHAOS = C_CHAOS
 * QUERY_CLASS_HS = C_HS
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(C_IN); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_CLASS_IN, __pyx_t_7) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":131
 * # Query classes
 * QUERY_CLASS_IN = C_IN
 * QUERY_CLASS_CHAOS = C_CHAOS             # <<<<<<<<<<<<<<
 * QUERY_CLASS_HS = C_HS
 * QUERY_CLASS_NONE = C_NONE
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(C_CHAOS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_CLASS_CHAOS, __pyx_t_7) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":132
 * QUERY_CLASS_IN = C_IN
 * QUERY_CLASS_CHAOS = C_CHAOS
 * QUERY_CLASS_HS = C_HS             # <<<<<<<<<<<<<<
 * QUERY_CLASS_NONE = C_NONE
 * QUERY_CLASS_ANY = C_ANY
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(C_HS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_CLASS_HS, __pyx_t_7) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":133
 * QUERY_CLASS_CHAOS = C_CHAOS
 * QUERY_CLASS_HS = C_HS
 * QUERY_CLASS_NONE = C_NONE             # <<<<<<<<<<<<<<
 * QUERY_CLASS_ANY = C_ANY
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(C_NONE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_CLASS_NONE, __pyx_t_7) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":134
 * QUERY_CLASS_HS = C_HS
 * QUERY_CLASS_NONE = C_NONE
 * QUERY_CLASS_ANY = C_ANY             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_int(C_ANY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_QUERY_CLASS_ANY, __pyx_t_7) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":325
 * 
 * 
 *     cpdef void cancel(self) noexcept:             # <<<<<<<<<<<<<<
 *         ares_cancel(self.channel)
 *         self._cancelled = True
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_3cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_cancel, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_cancel, __pyx_t_7) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":329
 *         self._cancelled = True
 * 
 *     def reinit(self):             # <<<<<<<<<<<<<<
 *         cdef int r = ares_reinit(self.channel)
 *         if r != ARES_SUCCESS:
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_5reinit, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_reinit, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_reinit, __pyx_t_7) < 0) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":352
 *         ares_destroy(self.channel)
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_9__enter__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel___enter, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_enter, __pyx_t_7) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":355
 *         return self
 * 
 *     def __exit__(self, *args):             # <<<<<<<<<<<<<<
 *         # TODO: Handle collection & Closure
 *         self.cancel()
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_11__exit__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel___exit, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_exit, __pyx_t_7) < 0) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":370
 * 
 *     # Not public to .pyi, please do not use... - Vizonex
 *     def __remove_future(self, object fut):             # <<<<<<<<<<<<<<
 *         self.handles.remove(fut)
 *         self._closed += 1
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_13_Channel__remove_future, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel___remove_future, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_Channel__remove_future, __pyx_t_7) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":379
 *     # using such functions for non-development purposes is discouraged.
 * 
 *     def debug(self):             # <<<<<<<<<<<<<<
 *         print('Running Handles: {}'.format(self._running))
 *         print('Running Queries: {}'.format(self.running_queries))
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_15debug, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_debug, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_debug, __pyx_t_7) < 0) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":553
 *         return fut
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_17query, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_query, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[7]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_query, __pyx_t_7) < 0) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":699
 *         return fut
 * 
 *     def search(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_19search, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_search, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[7]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_search, __pyx_t_7) < 0) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":702
 *         return self._search(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
 *     def process_fd(self, int read_fd, int write_fd):             # <<<<<<<<<<<<<<
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, <ares_socket_t>write_fd)
 * 
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_21process_fd, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_process_fd, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_process_fd, __pyx_t_7) < 0) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":708
 *     # NOTE: I'll make a pull-request to pycares to maybe implement this as well...
 * 
 *     def process_read_fd(self, int read_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes readable file-descriptor instead of needing to remember
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_23process_read_fd, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_process_read_fd, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_process_read_fd, __pyx_t_7) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":721
 *         ares_process_fd(self.channel, <ares_socket_t>read_fd, ARES_SOCKET_BAD)
 * 
 *     def process_write_fd(self, int write_fd):             # <<<<<<<<<<<<<<
 *         """
 *         processes writable file-descriptor instead of needing to remember
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_25process_write_fd, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_process_write_fd, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_process_write_fd, __pyx_t_7) < 0) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":735
 * 
 * 
 *     def timeout(self, double t = 0):             # <<<<<<<<<<<<<<
 *         cdef timeval maxtv
 *         cdef timeval tv
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_27timeout, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_timeout, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[8]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_timeout, __pyx_t_7) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":757
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 * 
 *     def getaddrinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         object host,
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_29getaddrinfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_getaddrinfo, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[9]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_getaddrinfo, __pyx_t_7) < 0) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":804
 *         return fut
 * 
 *     def getnameinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         tuple address,
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_31getnameinfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_getnameinfo, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[10]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_getnameinfo, __pyx_t_7) < 0) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":857
 *         return fut
 * 
 *     def gethostbyname(             # <<<<<<<<<<<<<<
 *         self,
 *         object name,
*/
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_33gethostbyname, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_gethostbyname, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[10]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_gethostbyname, __pyx_t_7) < 0) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyares/channel.pyx":879
 * 
 * 
 *     def gethostbyaddr(self, object addr, object callback = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_None) < 0) __PYX_ERR(0, 879, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_35gethostbyaddr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_gethostbyaddr, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_gethostbyaddr, __pyx_t_3) < 0) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":920
 * 
 * 
 *     def set_local_dev(self, object dev):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         cyares_get_buffer(dev, &view)
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_37set_local_dev, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_set_local_dev, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_set_local_dev, __pyx_t_3) < 0) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":926
 *         cyares_release_buffer(&view)
 * 
 *     def set_local_ip(self, object ip):             # <<<<<<<<<<<<<<
 *         cdef in_addr addr4
 *         cdef ares_in6_addr addr6
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_39set_local_ip, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_set_local_ip, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_set_local_ip, __pyx_t_3) < 0) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":941
 *             cyares_release_buffer(&view)
 * 
 *     def getsock(self):             # <<<<<<<<<<<<<<
 *         cdef list rfds = []
 *         cdef list wfds = []
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_41getsock, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_getsock, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_getsock, __pyx_t_3) < 0) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":962
 *         return ares_queue_wait_empty(self.channel, timeout_ms)
 * 
 *     def wait(self, object timeout = None):             # <<<<<<<<<<<<<<
 *         cdef ares_status_t status
 *         cdef int ms
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_43wait, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel_wait, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[10]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6cyares_7channel_Channel, __pyx_mstate_global->__pyx_n_u_wait, __pyx_t_3) < 0) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_45__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.channel,self.options cannot be converted to a Python object for pickling"
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_7Channel_47__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Channel___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":995
 * 
 * 
 * def cyares_threadsafety():             # <<<<<<<<<<<<<<
 *     """
 *     pycares documentation says:
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cyares_7channel_5cyares_threadsafety, 0, __pyx_mstate_global->__pyx_n_u_cyares_threadsafety, NULL, __pyx_mstate_global->__pyx_n_u_cyares_channel, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cyares_threadsafety, __pyx_t_3) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cyares/channel.pyx":1008
 * 
 * 
 * cdef int init_status = ares_library_init(ARES_LIB_INIT_ALL)             # <<<<<<<<<<<<<<
 * if ARES_SUCCESS != init_status:
 *     raise AresError(init_status)
*/
  __pyx_v_6cyares_7channel_init_status = ares_library_init(ARES_LIB_INIT_ALL);

  /* "cyares/channel.pyx":1009
 * 
 * cdef int init_status = ares_library_init(ARES_LIB_INIT_ALL)
 * if ARES_SUCCESS != init_status:             # <<<<<<<<<<<<<<
 *     raise AresError(init_status)
 * 
*/
  __pyx_t_8 = (ARES_SUCCESS != __pyx_v_6cyares_7channel_init_status);
  if (unlikely(__pyx_t_8)) {

    /* "cyares/channel.pyx":1010
 * cdef int init_status = ares_library_init(ARES_LIB_INIT_ALL)
 * if ARES_SUCCESS != init_status:
 *     raise AresError(init_status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_7 = NULL;
    __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError);
    __pyx_t_2 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6cyares_9exception_AresError); 
    __pyx_t_9 = __Pyx_PyLong_From_int(__pyx_v_6cyares_7channel_init_status); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_9};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_3);
    }
    __Pyx_Raise(((PyObject *)__pyx_t_3), 0, 0, 0);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1010, __pyx_L1_error)

    /* "cyares/channel.pyx":1009
 * 
 * cdef int init_status = ares_library_init(ARES_LIB_INIT_ALL)
 * if ARES_SUCCESS != init_status:             # <<<<<<<<<<<<<<
 *     raise AresError(init_status)
 * 
*/
  }

  /* "cyares/channel.pyx":1
 * # cython: embed_signature=Trie             # <<<<<<<<<<<<<<
 * cimport cython
 * from cpython.exc cimport PyErr_NoMemory, PyErr_SetObject
*/
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init cyares.channel", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cyares.channel");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */

typedef struct {
    const char *s;
#if 179 <= 65535
    const unsigned short n;
#elif 179 / 2 < INT_MAX
    const unsigned int n;
#elif 179 / 2 < LONG_MAX
    const unsigned long n;
#else
    const Py_ssize_t n;
#endif
#if 1 <= 31
    const unsigned int encoding : 5;
#elif 1 <= 255
    const unsigned char encoding;
#elif 1 <= 65535
    const unsigned short encoding;
#else
    const Py_ssize_t encoding;
#endif
    const unsigned int is_unicode : 1;
    const unsigned int intern : 1;
} __Pyx_StringTabEntry;
static const char * const __pyx_string_tab_encodings[] = { 0 };
static const __Pyx_StringTabEntry __pyx_string_tab[] = {
  {__pyx_k_, sizeof(__pyx_k_), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_ */
  {__pyx_k_A, sizeof(__pyx_k_A), 0, 1, 1}, /* PyObject cname: __pyx_n_u_A */
  {__pyx_k_AAAA, sizeof(__pyx_k_AAAA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_AAAA */
  {__pyx_k_ANY, sizeof(__pyx_k_ANY), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ANY */
  {__pyx_k_CAA, sizeof(__pyx_k_CAA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CAA */
  {__pyx_k_CNAME, sizeof(__pyx_k_CNAME), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CNAME */
  {__pyx_k_CYARES_SOCKET_BAD, sizeof(__pyx_k_CYARES_SOCKET_BAD), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CYARES_SOCKET_BAD */
  {__pyx_k_Channel, sizeof(__pyx_k_Channel), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel */
  {__pyx_k_Channel___enter, sizeof(__pyx_k_Channel___enter), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel___enter */
  {__pyx_k_Channel___exit, sizeof(__pyx_k_Channel___exit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel___exit */
  {__pyx_k_Channel___reduce_cython, sizeof(__pyx_k_Channel___reduce_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel___reduce_cython */
  {__pyx_k_Channel___remove_future, sizeof(__pyx_k_Channel___remove_future), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel___remove_future */
  {__pyx_k_Channel___setstate_cython, sizeof(__pyx_k_Channel___setstate_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel___setstate_cython */
  {__pyx_k_Channel__remove_future, sizeof(__pyx_k_Channel__remove_future), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel__remove_future */
  {__pyx_k_Channel_cancel, sizeof(__pyx_k_Channel_cancel), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_cancel */
  {__pyx_k_Channel_debug, sizeof(__pyx_k_Channel_debug), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_debug */
  {__pyx_k_Channel_getaddrinfo, sizeof(__pyx_k_Channel_getaddrinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_getaddrinfo */
  {__pyx_k_Channel_gethostbyaddr, sizeof(__pyx_k_Channel_gethostbyaddr), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_gethostbyaddr */
  {__pyx_k_Channel_gethostbyname, sizeof(__pyx_k_Channel_gethostbyname), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_gethostbyname */
  {__pyx_k_Channel_getnameinfo, sizeof(__pyx_k_Channel_getnameinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_getnameinfo */
  {__pyx_k_Channel_getsock, sizeof(__pyx_k_Channel_getsock), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_getsock */
  {__pyx_k_Channel_process_fd, sizeof(__pyx_k_Channel_process_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_process_fd */
  {__pyx_k_Channel_process_read_fd, sizeof(__pyx_k_Channel_process_read_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_process_read_fd */
  {__pyx_k_Channel_process_write_fd, sizeof(__pyx_k_Channel_process_write_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_process_write_fd */
  {__pyx_k_Channel_query, sizeof(__pyx_k_Channel_query), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_query */
  {__pyx_k_Channel_reinit, sizeof(__pyx_k_Channel_reinit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_reinit */
  {__pyx_k_Channel_search, sizeof(__pyx_k_Channel_search), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_search */
  {__pyx_k_Channel_set_local_dev, sizeof(__pyx_k_Channel_set_local_dev), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_set_local_dev */
  {__pyx_k_Channel_set_local_ip, sizeof(__pyx_k_Channel_set_local_ip), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_set_local_ip */
  {__pyx_k_Channel_timeout, sizeof(__pyx_k_Channel_timeout), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_timeout */
  {__pyx_k_Channel_wait, sizeof(__pyx_k_Channel_wait), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Channel_wait */
  {__pyx_k_Closed_Handles, sizeof(__pyx_k_Closed_Handles), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Closed_Handles */
  {__pyx_k_Domain_names_being_idna_decoded, sizeof(__pyx_k_Domain_names_being_idna_decoded), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Domain_names_being_idna_decoded */
  {__pyx_k_Future, sizeof(__pyx_k_Future), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Future */
  {__pyx_k_Future_2, sizeof(__pyx_k_Future_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Future_2 */
  {__pyx_k_HINFO, sizeof(__pyx_k_HINFO), 0, 1, 1}, /* PyObject cname: __pyx_n_u_HINFO */
  {__pyx_k_HTTPS, sizeof(__pyx_k_HTTPS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_HTTPS */
  {__pyx_k_Invalid_IPv4_address_r, sizeof(__pyx_k_Invalid_IPv4_address_r), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Invalid_IPv4_address_r */
  {__pyx_k_Invalid_IPv6_address_r, sizeof(__pyx_k_Invalid_IPv6_address_r), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Invalid_IPv6_address_r */
  {__pyx_k_Invalid_address_argument, sizeof(__pyx_k_Invalid_address_argument), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Invalid_address_argument */
  {__pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_KeyError */
  {__pyx_k_MX, sizeof(__pyx_k_MX), 0, 1, 1}, /* PyObject cname: __pyx_n_u_MX */
  {__pyx_k_ModuleNotFoundError, sizeof(__pyx_k_ModuleNotFoundError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ModuleNotFoundError */
  {__pyx_k_NAPTR, sizeof(__pyx_k_NAPTR), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NAPTR */
  {__pyx_k_NS, sizeof(__pyx_k_NS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NS */
  {__pyx_k_None, sizeof(__pyx_k_None), 0, 1, 1}, /* PyObject cname: __pyx_n_u_None */
  {__pyx_k_Note_that_Cython_is_deliberately, sizeof(__pyx_k_Note_that_Cython_is_deliberately), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Note_that_Cython_is_deliberately */
  {__pyx_k_OPT, sizeof(__pyx_k_OPT), 0, 1, 1}, /* PyObject cname: __pyx_n_u_OPT */
  {__pyx_k_PTR, sizeof(__pyx_k_PTR), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PTR */
  {__pyx_k_Provided_callbacks_must_be_calla, sizeof(__pyx_k_Provided_callbacks_must_be_calla), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Provided_callbacks_must_be_calla */
  {__pyx_k_QUERY_CLASS_ANY, sizeof(__pyx_k_QUERY_CLASS_ANY), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_CLASS_ANY */
  {__pyx_k_QUERY_CLASS_CHAOS, sizeof(__pyx_k_QUERY_CLASS_CHAOS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_CLASS_CHAOS */
  {__pyx_k_QUERY_CLASS_HS, sizeof(__pyx_k_QUERY_CLASS_HS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_CLASS_HS */
  {__pyx_k_QUERY_CLASS_IN, sizeof(__pyx_k_QUERY_CLASS_IN), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_CLASS_IN */
  {__pyx_k_QUERY_CLASS_NONE, sizeof(__pyx_k_QUERY_CLASS_NONE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_CLASS_NONE */
  {__pyx_k_QUERY_TYPE_A, sizeof(__pyx_k_QUERY_TYPE_A), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_A */
  {__pyx_k_QUERY_TYPE_AAAA, sizeof(__pyx_k_QUERY_TYPE_AAAA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_AAAA */
  {__pyx_k_QUERY_TYPE_ANY, sizeof(__pyx_k_QUERY_TYPE_ANY), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_ANY */
  {__pyx_k_QUERY_TYPE_CAA, sizeof(__pyx_k_QUERY_TYPE_CAA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_CAA */
  {__pyx_k_QUERY_TYPE_CNAME, sizeof(__pyx_k_QUERY_TYPE_CNAME), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_CNAME */
  {__pyx_k_QUERY_TYPE_MX, sizeof(__pyx_k_QUERY_TYPE_MX), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_MX */
  {__pyx_k_QUERY_TYPE_NAPTR, sizeof(__pyx_k_QUERY_TYPE_NAPTR), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_NAPTR */
  {__pyx_k_QUERY_TYPE_NS, sizeof(__pyx_k_QUERY_TYPE_NS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_NS */
  {__pyx_k_QUERY_TYPE_PTR, sizeof(__pyx_k_QUERY_TYPE_PTR), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_PTR */
  {__pyx_k_QUERY_TYPE_SOA, sizeof(__pyx_k_QUERY_TYPE_SOA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_SOA */
  {__pyx_k_QUERY_TYPE_SRV, sizeof(__pyx_k_QUERY_TYPE_SRV), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_SRV */
  {__pyx_k_QUERY_TYPE_TXT, sizeof(__pyx_k_QUERY_TYPE_TXT), 0, 1, 1}, /* PyObject cname: __pyx_n_u_QUERY_TYPE_TXT */
  {__pyx_k_Running_Handles, sizeof(__pyx_k_Running_Handles), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Running_Handles */
  {__pyx_k_Running_Queries, sizeof(__pyx_k_Running_Queries), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Running_Queries */
  {__pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_RuntimeError */
  {__pyx_k_SIG, sizeof(__pyx_k_SIG), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SIG */
  {__pyx_k_SOA, sizeof(__pyx_k_SOA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SOA */
  {__pyx_k_SRV, sizeof(__pyx_k_SRV), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SRV */
  {__pyx_k_SVCB, sizeof(__pyx_k_SVCB), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SVCB */
  {__pyx_k_TLSA, sizeof(__pyx_k_TLSA), 0, 1, 1}, /* PyObject cname: __pyx_n_u_TLSA */
  {__pyx_k_TXT, sizeof(__pyx_k_TXT), 0, 1, 1}, /* PyObject cname: __pyx_n_u_TXT */
  {__pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_TypeError */
  {__pyx_k_URI, sizeof(__pyx_k_URI), 0, 1, 1}, /* PyObject cname: __pyx_n_u_URI */
  {__pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ValueError */
  {__pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0}, /* PyObject cname: __pyx_kp_u__2 */
  {__pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0}, /* PyObject cname: __pyx_kp_u__3 */
  {__pyx_k_add_done_callback, sizeof(__pyx_k_add_done_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_add_done_callback */
  {__pyx_k_add_note, sizeof(__pyx_k_add_note), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_add_note */
  {__pyx_k_addr, sizeof(__pyx_k_addr), 0, 1, 1}, /* PyObject cname: __pyx_n_u_addr */
  {__pyx_k_addr4, sizeof(__pyx_k_addr4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_addr4 */
  {__pyx_k_addr6, sizeof(__pyx_k_addr6), 0, 1, 1}, /* PyObject cname: __pyx_n_u_addr6 */
  {__pyx_k_address, sizeof(__pyx_k_address), 0, 1, 1}, /* PyObject cname: __pyx_n_u_address */
  {__pyx_k_algorithm, sizeof(__pyx_k_algorithm), 0, 1, 1}, /* PyObject cname: __pyx_n_u_algorithm */
  {__pyx_k_args, sizeof(__pyx_k_args), 0, 1, 1}, /* PyObject cname: __pyx_n_u_args */
  {__pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 1, 1}, /* PyObject cname: __pyx_n_u_asyncio_coroutines */
  {__pyx_k_bitmask, sizeof(__pyx_k_bitmask), 0, 1, 1}, /* PyObject cname: __pyx_n_u_bitmask */
  {__pyx_k_buffer_carried, sizeof(__pyx_k_buffer_carried), 0, 1, 1}, /* PyObject cname: __pyx_n_u_buffer_carried */
  {__pyx_k_c_ares_is_not_built_with_thread, sizeof(__pyx_k_c_ares_is_not_built_with_thread), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_c_ares_is_not_built_with_thread */
  {__pyx_k_callback, sizeof(__pyx_k_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_callback */
  {__pyx_k_cancel, sizeof(__pyx_k_cancel), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cancel */
  {__pyx_k_cert_usage, sizeof(__pyx_k_cert_usage), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cert_usage */
  {__pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cline_in_traceback */
  {__pyx_k_concurrent_futures, sizeof(__pyx_k_concurrent_futures), 0, 1, 1}, /* PyObject cname: __pyx_n_u_concurrent_futures */
  {__pyx_k_cpu, sizeof(__pyx_k_cpu), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cpu */
  {__pyx_k_cyares_channel, sizeof(__pyx_k_cyares_channel), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cyares_channel */
  {__pyx_k_cyares_channel_pyx, sizeof(__pyx_k_cyares_channel_pyx), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_cyares_channel_pyx */
  {__pyx_k_cyares_threadsafety, sizeof(__pyx_k_cyares_threadsafety), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cyares_threadsafety */
  {__pyx_k_data, sizeof(__pyx_k_data), 0, 1, 1}, /* PyObject cname: __pyx_n_u_data */
  {__pyx_k_debug, sizeof(__pyx_k_debug), 0, 1, 1}, /* PyObject cname: __pyx_n_u_debug */
  {__pyx_k_decode, sizeof(__pyx_k_decode), 0, 1, 1}, /* PyObject cname: __pyx_n_u_decode */
  {__pyx_k_dev, sizeof(__pyx_k_dev), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dev */
  {__pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_disable */
  {__pyx_k_domains, sizeof(__pyx_k_domains), 0, 1, 1}, /* PyObject cname: __pyx_n_u_domains */
  {__pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_enable */
  {__pyx_k_encode, sizeof(__pyx_k_encode), 0, 1, 1}, /* PyObject cname: __pyx_n_u_encode */
  {__pyx_k_enter, sizeof(__pyx_k_enter), 0, 1, 1}, /* PyObject cname: __pyx_n_u_enter */
  {__pyx_k_event_thread, sizeof(__pyx_k_event_thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_event_thread */
  {__pyx_k_exit, sizeof(__pyx_k_exit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_exit */
  {__pyx_k_expiration, sizeof(__pyx_k_expiration), 0, 1, 1}, /* PyObject cname: __pyx_n_u_expiration */
  {__pyx_k_family, sizeof(__pyx_k_family), 0, 1, 1}, /* PyObject cname: __pyx_n_u_family */
  {__pyx_k_flags, sizeof(__pyx_k_flags), 0, 1, 1}, /* PyObject cname: __pyx_n_u_flags */
  {__pyx_k_flowinfo, sizeof(__pyx_k_flowinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_flowinfo */
  {__pyx_k_format, sizeof(__pyx_k_format), 0, 1, 1}, /* PyObject cname: __pyx_n_u_format */
  {__pyx_k_func, sizeof(__pyx_k_func), 0, 1, 1}, /* PyObject cname: __pyx_n_u_func */
  {__pyx_k_fut, sizeof(__pyx_k_fut), 0, 1, 1}, /* PyObject cname: __pyx_n_u_fut */
  {__pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_gc */
  {__pyx_k_getaddrinfo, sizeof(__pyx_k_getaddrinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getaddrinfo */
  {__pyx_k_gethostbyaddr, sizeof(__pyx_k_gethostbyaddr), 0, 1, 1}, /* PyObject cname: __pyx_n_u_gethostbyaddr */
  {__pyx_k_gethostbyname, sizeof(__pyx_k_gethostbyname), 0, 1, 1}, /* PyObject cname: __pyx_n_u_gethostbyname */
  {__pyx_k_getnameinfo, sizeof(__pyx_k_getnameinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getnameinfo */
  {__pyx_k_getsock, sizeof(__pyx_k_getsock), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getsock */
  {__pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getstate */
  {__pyx_k_hints, sizeof(__pyx_k_hints), 0, 1, 1}, /* PyObject cname: __pyx_n_u_hints */
  {__pyx_k_host, sizeof(__pyx_k_host), 0, 1, 1}, /* PyObject cname: __pyx_n_u_host */
  {__pyx_k_htonl, sizeof(__pyx_k_htonl), 0, 1, 1}, /* PyObject cname: __pyx_n_u_htonl */
  {__pyx_k_htons, sizeof(__pyx_k_htons), 0, 1, 1}, /* PyObject cname: __pyx_n_u_htons */
  {__pyx_k_i, sizeof(__pyx_k_i), 0, 1, 1}, /* PyObject cname: __pyx_n_u_i */
  {__pyx_k_id, sizeof(__pyx_k_id), 0, 1, 1}, /* PyObject cname: __pyx_n_u_id */
  {__pyx_k_idna, sizeof(__pyx_k_idna), 0, 1, 1}, /* PyObject cname: __pyx_n_u_idna */
  {__pyx_k_idna_decode, sizeof(__pyx_k_idna_decode), 0, 1, 1}, /* PyObject cname: __pyx_n_u_idna_decode */
  {__pyx_k_inception, sizeof(__pyx_k_inception), 0, 1, 1}, /* PyObject cname: __pyx_n_u_inception */
  {__pyx_k_invalid_IP_address, sizeof(__pyx_k_invalid_IP_address), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_invalid_IP_address */
  {__pyx_k_invalid_query_type_specified, sizeof(__pyx_k_invalid_query_type_specified), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_invalid_query_type_specified */
  {__pyx_k_ip, sizeof(__pyx_k_ip), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ip */
  {__pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_coroutine */
  {__pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_isenabled */
  {__pyx_k_key_tag, sizeof(__pyx_k_key_tag), 0, 1, 1}, /* PyObject cname: __pyx_n_u_key_tag */
  {__pyx_k_l, sizeof(__pyx_k_l), 0, 1, 1}, /* PyObject cname: __pyx_n_u_l */
  {__pyx_k_labels, sizeof(__pyx_k_labels), 0, 1, 1}, /* PyObject cname: __pyx_n_u_labels */
  {__pyx_k_local_dev, sizeof(__pyx_k_local_dev), 0, 1, 1}, /* PyObject cname: __pyx_n_u_local_dev */
  {__pyx_k_local_ip, sizeof(__pyx_k_local_ip), 0, 1, 1}, /* PyObject cname: __pyx_n_u_local_ip */
  {__pyx_k_lookups, sizeof(__pyx_k_lookups), 0, 1, 1}, /* PyObject cname: __pyx_n_u_lookups */
  {__pyx_k_main, sizeof(__pyx_k_main), 0, 1, 1}, /* PyObject cname: __pyx_n_u_main */
  {__pyx_k_match, sizeof(__pyx_k_match), 0, 1, 1}, /* PyObject cname: __pyx_n_u_match */
  {__pyx_k_maxtv, sizeof(__pyx_k_maxtv), 0, 1, 1}, /* PyObject cname: __pyx_n_u_maxtv */
  {__pyx_k_module, sizeof(__pyx_k_module), 0, 1, 1}, /* PyObject cname: __pyx_n_u_module */
  {__pyx_k_ms, sizeof(__pyx_k_ms), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ms */
  {__pyx_k_name, sizeof(__pyx_k_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_name */
  {__pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_name_2 */
  {__pyx_k_ndots, sizeof(__pyx_k_ndots), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ndots */
  {__pyx_k_options, sizeof(__pyx_k_options), 0, 1, 1}, /* PyObject cname: __pyx_n_u_options */
  {__pyx_k_original_ttl, sizeof(__pyx_k_original_ttl), 0, 1, 1}, /* PyObject cname: __pyx_n_u_original_ttl */
  {__pyx_k_os, sizeof(__pyx_k_os), 0, 1, 1}, /* PyObject cname: __pyx_n_u_os */
  {__pyx_k_params, sizeof(__pyx_k_params), 0, 1, 1}, /* PyObject cname: __pyx_n_u_params */
  {__pyx_k_pop, sizeof(__pyx_k_pop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pop */
  {__pyx_k_port, sizeof(__pyx_k_port), 0, 1, 1}, /* PyObject cname: __pyx_n_u_port */
  {__pyx_k_print, sizeof(__pyx_k_print), 0, 1, 1}, /* PyObject cname: __pyx_n_u_print */
  {__pyx_k_priority, sizeof(__pyx_k_priority), 0, 1, 1}, /* PyObject cname: __pyx_n_u_priority */
  {__pyx_k_process_fd, sizeof(__pyx_k_process_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_process_fd */
  {__pyx_k_process_read_fd, sizeof(__pyx_k_process_read_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_process_read_fd */
  {__pyx_k_process_write_fd, sizeof(__pyx_k_process_write_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_process_write_fd */
  {__pyx_k_proto, sizeof(__pyx_k_proto), 0, 1, 1}, /* PyObject cname: __pyx_n_u_proto */
  {__pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_capi */
  {__pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_state */
  {__pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_vtable */
  {__pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 1, 1}, /* PyObject cname: __pyx_n_u_qualname */
  {__pyx_k_query, sizeof(__pyx_k_query), 0, 1, 1}, /* PyObject cname: __pyx_n_u_query */
  {__pyx_k_query_class, sizeof(__pyx_k_query_class), 0, 1, 1}, /* PyObject cname: __pyx_n_u_query_class */
  {__pyx_k_query_type, sizeof(__pyx_k_query_type), 0, 1, 1}, /* PyObject cname: __pyx_n_u_query_type */
  {__pyx_k_r, sizeof(__pyx_k_r), 0, 1, 1}, /* PyObject cname: __pyx_n_u_r */
  {__pyx_k_range, sizeof(__pyx_k_range), 0, 1, 1}, /* PyObject cname: __pyx_n_u_range */
  {__pyx_k_read_fd, sizeof(__pyx_k_read_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_read_fd */
  {__pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce */
  {__pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce_cython */
  {__pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce_ex */
  {__pyx_k_reinit, sizeof(__pyx_k_reinit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reinit */
  {__pyx_k_remove_future, sizeof(__pyx_k_remove_future), 0, 1, 1}, /* PyObject cname: __pyx_n_u_remove_future */
  {__pyx_k_resolvconf_path, sizeof(__pyx_k_resolvconf_path), 0, 1, 1}, /* PyObject cname: __pyx_n_u_resolvconf_path */
  {__pyx_k_return, sizeof(__pyx_k_return), 0, 1, 1}, /* PyObject cname: __pyx_n_u_return */
  {__pyx_k_rfds, sizeof(__pyx_k_rfds), 0, 1, 1}, /* PyObject cname: __pyx_n_u_rfds */
  {__pyx_k_rotate, sizeof(__pyx_k_rotate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_rotate */
  {__pyx_k_running_queries, sizeof(__pyx_k_running_queries), 0, 1, 1}, /* PyObject cname: __pyx_n_u_running_queries */
  {__pyx_k_running_queries_is_an_immutable, sizeof(__pyx_k_running_queries_is_an_immutable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_running_queries_is_an_immutable */
  {__pyx_k_s, sizeof(__pyx_k_s), 0, 1, 1}, /* PyObject cname: __pyx_n_u_s */
  {__pyx_k_sa4, sizeof(__pyx_k_sa4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_sa4 */
  {__pyx_k_sa6, sizeof(__pyx_k_sa6), 0, 1, 1}, /* PyObject cname: __pyx_n_u_sa6 */
  {__pyx_k_scope_id, sizeof(__pyx_k_scope_id), 0, 1, 1}, /* PyObject cname: __pyx_n_u_scope_id */
  {__pyx_k_search, sizeof(__pyx_k_search), 0, 1, 1}, /* PyObject cname: __pyx_n_u_search */
  {__pyx_k_selector, sizeof(__pyx_k_selector), 0, 1, 1}, /* PyObject cname: __pyx_n_u_selector */
  {__pyx_k_self, sizeof(__pyx_k_self), 0, 1, 1}, /* PyObject cname: __pyx_n_u_self */
  {__pyx_k_self_channel_self_options_cannot, sizeof(__pyx_k_self_channel_self_options_cannot), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_self_channel_self_options_cannot */
  {__pyx_k_servers, sizeof(__pyx_k_servers), 0, 1, 1}, /* PyObject cname: __pyx_n_u_servers */
  {__pyx_k_service, sizeof(__pyx_k_service), 0, 1, 1}, /* PyObject cname: __pyx_n_u_service */
  {__pyx_k_service_data, sizeof(__pyx_k_service_data), 0, 1, 1}, /* PyObject cname: __pyx_n_u_service_data */
  {__pyx_k_set_local_dev, sizeof(__pyx_k_set_local_dev), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_local_dev */
  {__pyx_k_set_local_ip, sizeof(__pyx_k_set_local_ip), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_local_ip */
  {__pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_name */
  {__pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_setstate */
  {__pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_setstate_cython */
  {__pyx_k_signature, sizeof(__pyx_k_signature), 0, 1, 1}, /* PyObject cname: __pyx_n_u_signature */
  {__pyx_k_signers_name, sizeof(__pyx_k_signers_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_signers_name */
  {__pyx_k_sock_state_cb, sizeof(__pyx_k_sock_state_cb), 0, 1, 1}, /* PyObject cname: __pyx_n_u_sock_state_cb */
  {__pyx_k_sock_state_cb_and_event_thread_c, sizeof(__pyx_k_sock_state_cb_and_event_thread_c), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_sock_state_cb_and_event_thread_c */
  {__pyx_k_sock_state_cb_must_be_callable, sizeof(__pyx_k_sock_state_cb_must_be_callable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_sock_state_cb_must_be_callable */
  {__pyx_k_socket_receive_buffer_size, sizeof(__pyx_k_socket_receive_buffer_size), 0, 1, 1}, /* PyObject cname: __pyx_n_u_socket_receive_buffer_size */
  {__pyx_k_socket_send_buffer_size, sizeof(__pyx_k_socket_send_buffer_size), 0, 1, 1}, /* PyObject cname: __pyx_n_u_socket_send_buffer_size */
  {__pyx_k_socks, sizeof(__pyx_k_socks), 0, 1, 1}, /* PyObject cname: __pyx_n_u_socks */
  {__pyx_k_socktype, sizeof(__pyx_k_socktype), 0, 1, 1}, /* PyObject cname: __pyx_n_u_socktype */
  {__pyx_k_status, sizeof(__pyx_k_status), 0, 1, 1}, /* PyObject cname: __pyx_n_u_status */
  {__pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_stringsource */
  {__pyx_k_t, sizeof(__pyx_k_t), 0, 1, 1}, /* PyObject cname: __pyx_n_u_t */
  {__pyx_k_target, sizeof(__pyx_k_target), 0, 1, 1}, /* PyObject cname: __pyx_n_u_target */
  {__pyx_k_tcp_port, sizeof(__pyx_k_tcp_port), 0, 1, 1}, /* PyObject cname: __pyx_n_u_tcp_port */
  {__pyx_k_test, sizeof(__pyx_k_test), 0, 1, 1}, /* PyObject cname: __pyx_n_u_test */
  {__pyx_k_timeout, sizeof(__pyx_k_timeout), 0, 1, 1}, /* PyObject cname: __pyx_n_u_timeout */
  {__pyx_k_timeout_must_be_an_integer_or_fl, sizeof(__pyx_k_timeout_must_be_an_integer_or_fl), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_timeout_must_be_an_integer_or_fl */
  {__pyx_k_timeout_needs_to_be_a_positive_n, sizeof(__pyx_k_timeout_needs_to_be_a_positive_n), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_timeout_needs_to_be_a_positive_n */
  {__pyx_k_timeout_should_not_be_below_0_go, sizeof(__pyx_k_timeout_should_not_be_below_0_go), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_timeout_should_not_be_below_0_go */
  {__pyx_k_tries, sizeof(__pyx_k_tries), 0, 1, 1}, /* PyObject cname: __pyx_n_u_tries */
  {__pyx_k_tv, sizeof(__pyx_k_tv), 0, 1, 1}, /* PyObject cname: __pyx_n_u_tv */
  {__pyx_k_type, sizeof(__pyx_k_type), 0, 1, 1}, /* PyObject cname: __pyx_n_u_type */
  {__pyx_k_type_covered, sizeof(__pyx_k_type_covered), 0, 1, 1}, /* PyObject cname: __pyx_n_u_type_covered */
  {__pyx_k_udp_port, sizeof(__pyx_k_udp_port), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udp_port */
  {__pyx_k_udp_size, sizeof(__pyx_k_udp_size), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udp_size */
  {__pyx_k_val, sizeof(__pyx_k_val), 0, 1, 1}, /* PyObject cname: __pyx_n_u_val */
  {__pyx_k_version, sizeof(__pyx_k_version), 0, 1, 1}, /* PyObject cname: __pyx_n_u_version */
  {__pyx_k_view, sizeof(__pyx_k_view), 0, 1, 1}, /* PyObject cname: __pyx_n_u_view */
  {__pyx_k_wait, sizeof(__pyx_k_wait), 0, 1, 1}, /* PyObject cname: __pyx_n_u_wait */
  {__pyx_k_weight, sizeof(__pyx_k_weight), 0, 1, 1}, /* PyObject cname: __pyx_n_u_weight */
  {__pyx_k_wfds, sizeof(__pyx_k_wfds), 0, 1, 1}, /* PyObject cname: __pyx_n_u_wfds */
  {__pyx_k_write_fd, sizeof(__pyx_k_write_fd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_write_fd */
  {0, 0, 0, 0, 0}
};
/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names);

/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_ModuleNotFoundError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_ModuleNotFoundError); if (!__pyx_builtin_ModuleNotFoundError) __PYX_ERR(0, 37, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_print); if (!__pyx_builtin_print) __PYX_ERR(0, 380, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 414, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_range); if (!__pyx_builtin_range) __PYX_ERR(0, 950, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "resulttypes.pxd":45
 *         r.id = ares_dns_rr_get_opt(dns_rr, key, idx, &cstr, &len)
 *         r.val = cyares_unicode_from_uchar_and_size(cstr, <Py_ssize_t>len)
 *         r._attrs = ("id", "val")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_id, __pyx_mstate_global->__pyx_n_u_val); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "resulttypes.pxd":75
 *         r.options = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_OPT_OPTIONS)
 *         r.type = "OPT"
 *         r._attrs = ("udp_size", "version", "flags", "options", "type")             # <<<<<<<<<<<<<<
 *         return r
 *         #
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(5, __pyx_mstate_global->__pyx_n_u_udp_size, __pyx_mstate_global->__pyx_n_u_version, __pyx_mstate_global->__pyx_n_u_flags, __pyx_mstate_global->__pyx_n_u_options, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "resulttypes.pxd":93
 *         r.os = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_HINFO_OS)
 *         r.type = "HINFO"
 *         r._attrs = ("cpu", "os", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(3, __pyx_mstate_global->__pyx_n_u_cpu, __pyx_mstate_global->__pyx_n_u_os, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "resulttypes.pxd":128
 *         r.signers_name = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_SIG_SIGNERS_NAME)
 *         r.type = "SIG"
 *         r._attrs = ("type_covered", "algorithm", "labels", "original_ttl", "expiration", "inception", "key_tag", "signature", "signers_name", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(10, __pyx_mstate_global->__pyx_n_u_type_covered, __pyx_mstate_global->__pyx_n_u_algorithm, __pyx_mstate_global->__pyx_n_u_labels, __pyx_mstate_global->__pyx_n_u_original_ttl, __pyx_mstate_global->__pyx_n_u_expiration, __pyx_mstate_global->__pyx_n_u_inception, __pyx_mstate_global->__pyx_n_u_key_tag, __pyx_mstate_global->__pyx_n_u_signature, __pyx_mstate_global->__pyx_n_u_signers_name, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);

  /* "resulttypes.pxd":153
 *         r.data = cyares_unicode_from_uchar_and_size(cstr, l)
 *         r.type = "TLSA"
 *         r._attrs = ("cert_usage", "selector", "match", "data", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[4] = PyTuple_Pack(5, __pyx_mstate_global->__pyx_n_u_cert_usage, __pyx_mstate_global->__pyx_n_u_selector, __pyx_mstate_global->__pyx_n_u_match, __pyx_mstate_global->__pyx_n_u_data, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[4])) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);

  /* "resulttypes.pxd":174
 *         r.params = cyares_dns_rr_get_optresults(dns_rr, ARES_RR_SVCB_PARAMS)
 *         r.type = "SVCB"
 *         r._attrs = ("priority", "target", "params", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[5] = PyTuple_Pack(4, __pyx_mstate_global->__pyx_n_u_priority, __pyx_mstate_global->__pyx_n_u_target, __pyx_mstate_global->__pyx_n_u_params, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[5])) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);

  /* "resulttypes.pxd":216
 *         r.target = cyares_dns_rr_get_bytes(dns_rr, ARES_RR_URI_TARGET)
 *         r.type = "URI"
 *         r._attrs =  ("priority", "target", "weight", "type")             # <<<<<<<<<<<<<<
 *         return r
 * 
*/
  __pyx_mstate_global->__pyx_tuple[6] = PyTuple_Pack(4, __pyx_mstate_global->__pyx_n_u_priority, __pyx_mstate_global->__pyx_n_u_target, __pyx_mstate_global->__pyx_n_u_weight, __pyx_mstate_global->__pyx_n_u_type); if (unlikely(!__pyx_mstate_global->__pyx_tuple[6])) __PYX_ERR(1, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[6]);

  /* "cyares/channel.pyx":553
 *         return fut
 * 
 *     def query(self, object name, object query_type, object callback = None , object query_class = None):             # <<<<<<<<<<<<<<
 *         return self._query(name, query_type, C_IN if query_class is None else <int>query_class, callback)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[7] = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[7])) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[7]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[7]);

  /* "cyares/channel.pyx":735
 * 
 * 
 *     def timeout(self, double t = 0):             # <<<<<<<<<<<<<<
 *         cdef timeval maxtv
 *         cdef timeval tv
*/
  __pyx_mstate_global->__pyx_tuple[8] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_float_0_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[8])) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[8]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[8]);

  /* "cyares/channel.pyx":757
 *         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 * 
 *     def getaddrinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         object host,
*/
  __pyx_mstate_global->__pyx_tuple[9] = PyTuple_Pack(6, Py_None, Py_None, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[9])) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[9]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[9]);

  /* "cyares/channel.pyx":804
 *         return fut
 * 
 *     def getnameinfo(             # <<<<<<<<<<<<<<
 *         self,
 *         tuple address,
*/
  __pyx_mstate_global->__pyx_tuple[10] = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[10])) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  if (__Pyx_InitStrings(__pyx_string_tab, __pyx_mstate->__pyx_string_tab, __pyx_string_tab_encodings) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_mstate->__pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_mstate->__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_0 = PyLong_FromLong(0); if (unlikely(!__pyx_mstate->__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_1000 = PyLong_FromLong(1000); if (unlikely(!__pyx_mstate->__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
\
        typedef struct {
            unsigned int argcount : 4;
            unsigned int num_posonly_args : 1;
            unsigned int num_kwonly_args : 1;
            unsigned int nlocals : 4;
            unsigned int flags : 10;
            unsigned int first_line : 10;
            unsigned int line_table_length : 13;
        } __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 41, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_s};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_htons, __pyx_k_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 44, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_l};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_htonl, __pyx_k_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 325, 18};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_cancel, __pyx_k_A_1D_N, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 329, 31};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_r};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_reinit, __pyx_k_A_Q_2S_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 352, 7};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_enter, __pyx_k_A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS), 355, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_args};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_exit, __pyx_k_A_G1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 370, 20};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_fut};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_remove_future, __pyx_k_A_HG1A_L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 379, 50};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_debug, __pyx_k_A_Q_7_4q_Q_7_4q_Q_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 553, 38};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_query_type, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_query_class};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_query, __pyx_k_FFbbc_t7_6_X_JeS_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 699, 38};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_query_type, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_query_class};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_search, __pyx_k_GGccd_t81F_hl_ZuTaab, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 702, 19};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_read_fd, __pyx_mstate->__pyx_n_u_write_fd};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_process_fd, __pyx_k_A_q_JoYoQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 708, 19};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_read_fd};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_process_read_fd, __pyx_k_A_q_JoYa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 721, 20};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_write_fd};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_process_write_fd, __pyx_k_A_q_J_7_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 735, 132};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_t, __pyx_mstate->__pyx_n_u_maxtv, __pyx_mstate->__pyx_n_u_tv};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_timeout, __pyx_k_1_r_A_ZxuAQ_t1CuBa_j_1_A_AT_1G1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {8, 0, 0, 14, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 757, 213};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_host, __pyx_mstate->__pyx_n_u_port, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_family, __pyx_mstate->__pyx_n_u_socktype, __pyx_mstate->__pyx_n_u_proto, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_service, __pyx_mstate->__pyx_n_u_view, __pyx_mstate->__pyx_n_u_service_data, __pyx_mstate->__pyx_n_u_buffer_carried, __pyx_mstate->__pyx_n_u_hints, __pyx_mstate->__pyx_n_u_fut};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_getaddrinfo, __pyx_k_A_Q_1_aq_1_z_uAQ_QfAQ_g_Q_QfAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 804, 345};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_address, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_sa4, __pyx_mstate->__pyx_n_u_sa6, __pyx_mstate->__pyx_n_u_fut, __pyx_mstate->__pyx_n_u_view, __pyx_mstate->__pyx_n_u_ip, __pyx_mstate->__pyx_n_u_port, __pyx_mstate->__pyx_n_u_flowinfo, __pyx_mstate->__pyx_n_u_scope_id};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_getnameinfo, __pyx_k_A_3ay_1_q_Qd_1_t_7_fAS_Qaq_j_2Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 857, 74};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_family, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_view, __pyx_mstate->__pyx_n_u_fut};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_gethostbyname, __pyx_k_A_aq_QfAQ_4q_1_Qaq_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 879, 178};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_addr, __pyx_mstate->__pyx_n_u_callback, __pyx_mstate->__pyx_n_u_addr4, __pyx_mstate->__pyx_n_u_addr6, __pyx_mstate->__pyx_n_u_view, __pyx_mstate->__pyx_n_u_fut};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_gethostbyaddr, __pyx_k_9A_7_fAQ_aq_a_A_q_q_1JgT_q_aq_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 920, 37};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_dev, __pyx_mstate->__pyx_n_u_view};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_set_local_dev, __pyx_k_A_q_4z_A_Qaq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 926, 108};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ip, __pyx_mstate->__pyx_n_u_addr4, __pyx_mstate->__pyx_n_u_addr6, __pyx_mstate->__pyx_n_u_view};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_set_local_ip, __pyx_k_A_aq_Qiwd_4z_1EQR_q_F_1_4z1A_Q_j, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 941, 101};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rfds, __pyx_mstate->__pyx_n_u_wfds, __pyx_mstate->__pyx_n_u_socks, __pyx_mstate->__pyx_n_u_bitmask, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_getsock, __pyx_k_A_q_JgQ_E_aq_AYa_G1HE_AYa_G1E_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 962, 95};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_ms};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_wait, __pyx_k_A_83a_T_q_s_A_T_7_Q_1_Cq_1_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 995, 16};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_cyares_channel_pyx, __pyx_mstate->__pyx_n_u_cyares_threadsafety, __pyx_k_Cs, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
{
    PyObject *typesModule=NULL;
    typesModule = PyImport_ImportModule("types");
    if (typesModule) {
        __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
        Py_DECREF(typesModule);
    }
} // error handling follows
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* GetRuntimeVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    static unsigned long __Pyx_cached_runtime_version = 0;
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
    return __Pyx_cached_runtime_version;
#endif
}

/* tp_new */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __Pyx_tp_new_fallback(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    PyObject *new_func = NULL, *new_args = NULL, *obj;
    Py_ssize_t i, nargs = PyTuple_Size(args);
    if (unlikely(nargs < 0)) goto bad;
    new_args = PyTuple_New(nargs + 1);
    if (unlikely(!new_args)) goto bad;
    for (i = 0; i < nargs + 1; i++) {
        PyObject *item = (i == 0) ? type_obj : PyTuple_GetItem(args, i - 1);
        if (unlikely(!item)) goto bad;
        Py_INCREF(item);
        if (unlikely(PyTuple_SetItem(new_args, i, item)) < 0) goto bad;
    }
    new_func = PyObject_GetAttrString(type_obj, "__new__");
    if (unlikely(!new_func)) goto bad;
    obj = PyObject_Call(new_func, new_args, kwargs);
    Py_DECREF(new_func);
    Py_DECREF(new_args);
    return obj;
 bad:
    Py_XDECREF(new_func);
    Py_XDECREF(new_args);
    return NULL;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* TupleAndListFromArray */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < 0)) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject *const *args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
    PyObject *kwdefs;
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    if (
            co->co_kwonlyargcount == 0 &&
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
        vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
        if (f) {
            return f(func, args, _nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, _nargs, NULL);
        }
        #elif CYTHON_COMPILING_IN_LIMITED_API && CYTHON_VECTORCALL
        return PyObject_Vectorcall(func, args, _nargs, NULL);
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* UnpackUnboundCMethod */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywords */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    Py_ssize_t pos = 0;
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (PyDict_Next(kwds, &pos, &key, NULL)) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            #if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
            #endif
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
            #if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(key);
            #endif
        }
    }
    __Pyx_END_CRITICAL_SECTION();
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_MultiplyObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a, x;
#ifdef HAVE_LONG_LONG
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla, llx;
#endif
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op1);
    }
    if (likely(__Pyx_PyLong_IsCompact(op1))) {
        a = __Pyx_PyLong_CompactValue(op1);
    } else {
        const digit* digits = __Pyx_PyLong_Digits(op1);
        const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
        switch (size) {
            case -2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                    a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                    lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                    a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                    lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                    a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                    lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
        }
    }
            CYTHON_UNUSED_VAR(a);
            CYTHON_UNUSED_VAR(b);
            #ifdef HAVE_LONG_LONG
            lla = a;
            goto long_long;
            #else
            return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            #endif
        return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
    long_long:
            llx = lla * llb;
        return PyLong_FromLongLong(llx);
#endif
    return __Pyx_Fallback___Pyx_PyLong_MultiplyObjC(op1, op2, inplace);
    
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_MultiplyObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) * (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_MultiplyObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_MultiplyObjC(op1, op2, inplace);
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
        Py_ssize_t pos = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        PyDict_Next(kwds, &pos, &key, NULL);
        Py_INCREF(key);
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
#if CYTHON_COMPILING_IN_LIMITED_API
            || __Pyx_get_runtime_version() >= 0x030A0000
#endif
            )
            return result;
        {
            Py_ssize_t size = __Pyx_PySet_GET_SIZE(result);
            if (likely(size > 0))
                return result;
#if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(size < 0)) {
                Py_DECREF(result);
                return NULL;
            }
#endif
        }
        Py_DECREF(result);
#endif
    }
    return __Pyx_PyObject_CallNoArg((PyObject*) &PyFrozenSet_Type);
}

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_remove */
static int __Pyx_PySet_RemoveNotFound(PyObject *set, PyObject *key, int found) {
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    if (likely(found == 0)) {
        PyObject *tup;
        tup = PyTuple_Pack(1, key);
        if (!tup)
            return -1;
        PyErr_SetObject(PyExc_KeyError, tup);
        Py_DECREF(tup);
        return -1;
    }
    return found;
}
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found != 1)) {
        return __Pyx_PySet_RemoveNotFound(set, key, found);
    }
    return 0;
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* ReRaiseException */
static CYTHON_INLINE void __Pyx_ReraiseException(void) {
    PyObject *type = NULL, *value = NULL, *tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
  #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    value = exc_info->exc_value;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (unlikely(value == Py_None)) {
        value = NULL;
    } else if (value) {
        Py_INCREF(value);
        type = (PyObject*) Py_TYPE(value);
        Py_INCREF(type);
        tb = PyException_GetTraceback(value);
    }
    #else
    type = exc_info->exc_type;
    tb = exc_info->exc_traceback;
    Py_XINCREF(type);
    Py_XINCREF(value);
    Py_XINCREF(tb);
    #endif
  #else
    type = tstate->exc_type;
    value = tstate->exc_value;
    tb = tstate->exc_traceback;
    Py_XINCREF(type);
    Py_XINCREF(value);
    Py_XINCREF(tb);
  #endif
#else
    PyErr_GetExcInfo(&type, &value, &tb);
#endif
    if (unlikely(!type || type == Py_None)) {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(tb);
        PyErr_SetString(PyExc_RuntimeError,
            "No active exception to reraise");
    } else {
        PyErr_Restore(type, value, tb);
    }
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* LimitedApiGetTypeDict */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_1_2
#define __PYX_HAVE_RT_ImportType_3_1_2
static PyTypeObject *__Pyx_ImportType_3_1_2(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_1_2 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_1_2 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_1_2 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__2);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       int is_safe_type) {
    PyObject* value;
    CYTHON_MAYBE_UNUSED_VAR(is_safe_type);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethod(d, "setdefault", "OO", key, default_value);
#elif PY_VERSION_HEX >= 0x030d0000
    PyDict_SetDefaultRef(d, key, default_value, &value);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* FetchCommonType */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromMetaclass(metaclass, abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type, 1);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass */
PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
#endif
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        Py_INCREF(key);
        Py_INCREF(value);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_dict_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __Pyx_BEGIN_CRITICAL_SECTION(*cython_runtime_dict);
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback))
        Py_XINCREF(use_cline);
        __Pyx_END_CRITICAL_SECTION();
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_INCREF(use_cline);
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        py_funcname = PyUnicode_FromString(funcname);
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE unsigned short __Pyx_PyLong_As_unsigned_short(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned short val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned short) -1;
        val = __Pyx_PyLong_As_unsigned_short(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned short, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned short) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 2 * PyLong_SHIFT)) {
                            return (unsigned short) (((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned short) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 3 * PyLong_SHIFT)) {
                            return (unsigned short) (((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned short) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 4 * PyLong_SHIFT)) {
                            return (unsigned short) (((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned short) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned short) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned short, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned short) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned short) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned short) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned short) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned short) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned short) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned short val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned short) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned short) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned short) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned short) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned short) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned short) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned short) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned short) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned short) 1) << (sizeof(unsigned short) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned short) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned short");
    return (unsigned short) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned short");
    return (unsigned short) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyLong_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned short) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned short) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned short),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned short));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ares_socket_t(ares_socket_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const ares_socket_t neg_one = (ares_socket_t) -1, const_zero = (ares_socket_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ares_socket_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ares_socket_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(ares_socket_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ares_socket_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(ares_socket_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(ares_socket_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(ares_socket_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ares_status_t(ares_status_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const ares_status_t neg_one = (ares_status_t) -1, const_zero = (ares_status_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ares_status_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ares_status_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(ares_status_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ares_status_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(ares_status_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(ares_status_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(ares_status_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__3);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    if (__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_capi, &d) == -1)
        goto bad;
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_mstate_global->__pyx_n_u_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_1_2
#define __PYX_HAVE_RT_ImportFunction_3_1_2
static int __Pyx_ImportFunction_3_1_2(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, funcname, &cobj);
#else
    cobj = PyDict_GetItemString(d, funcname);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL, *line_table_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL
        && !CYTHON_COMPILING_IN_GRAAL) {
        line_table_bytes = PyBytes_FromStringAndSize(line_table, descr.line_table_length);
        if (unlikely(!line_table_bytes)) goto done;
        Py_ssize_t code_len = (descr.line_table_length * 2 + 4) & ~3;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table_bytes) ? line_table_bytes : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    Py_XDECREF(line_table_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names) {
    while (t->s) {
        PyObject *str;
        if (t->is_unicode) {
            if (t->intern) {
                str = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                str = PyUnicode_Decode(t->s, t->n - 1, encoding_names[t->encoding], NULL);
            } else {
                str = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            str = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        if (!str)
            return -1;
        *target = str;
        if (PyObject_Hash(str) == -1)
            return -1;
        ++t;
        ++target;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
